/*
 * Software Suspend.
 *
 * Module level definitions.
 *
 */

#ifndef SWSUSP_COMMON_H
#define SWSUSP_COMMON_H

#include <linux/version.h>
#include <linux/mm.h>
#include <linux/suspend.h>
#include <linux/proc_fs.h>
#include <linux/smp_lock.h>
#include <linux/file.h>
#include <linux/utsname.h>
#include <linux/init.h>
#include <linux/compiler.h>
#include <linux/delay.h>
#include <linux/reboot.h>
#include <linux/vt_kern.h>
#include <linux/bitops.h>
#include <linux/interrupt.h>
#include <linux/kbd_kern.h>
#include <linux/keyboard.h>
#include <linux/spinlock.h>
#include <linux/genhd.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/swap.h>
#include <linux/pm.h>
#include <linux/raid/md.h>
#include <linux/selection.h>
#include <linux/console.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/suspend-debug.h>
#include <linux/zlib.h>
#include <linux/suspend-version-specific.h>

#include <asm/uaccess.h>
#include <asm/mmu_context.h>
#include <asm/pgtable.h>
#include <asm/io.h>

/* -------------- Version specific defines --------------------- */

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,99)
#define NUM_CPUS num_online_cpus()
#define MIN min
#define MAX max
#define PROCESS_SIG_MASK(p) (&(p)->sighand->siglock)
#define FREE_AREA_TYPE struct free_area
#define ZONE_TYPE struct zone
#define IDE_UNSUSPEND
#define IDE_SUSPEND
#define EAT_MEMORY_FLAGS GFP_ATOMIC | __GFP_NOWARN
#define CPU0_MASK (cpumask_of_cpu(0))

#define INACTIVE_LIST_FOR_EACH(procedure) \
do { \
 if (zone->nr_inactive) { \
  struct list_head * entry = zone->inactive_list.prev; \
  while (entry != &zone->inactive_list) { \
   struct page * page = list_entry(entry, struct page, lru); \
   { \
    procedure(page); \
   } \
   entry = entry->prev; \
  } \
 } \
} while(0)

#define ACTIVE_LIST_FOR_EACH(procedure) \
do { \
 if (zone->nr_active) { \
  struct list_head * entry = zone->active_list.prev; \
  while (entry != &zone->active_list) { \
   struct page * page; \
   page = list_entry(entry, struct page, lru); \
   procedure(page); \
   entry = entry->prev; \
  } \
 } \
} while(0)

#else /* Kernel version 2.4 series */

#if defined(CONFIG_FBCON_SPLASHSCREEN)
#include <linux/fb.h>
#include <video/fbcon.h>
#else
#include <linux/console_struct.h>
#endif

#define NUM_CPUS smp_num_cpus
#define MAX(a,b)	((signed long)(a)>(signed long)(b)?(a):(b))
#define MIN(a,b)	((signed long)(a)<(signed long)(b)?(a):(b))
#define PROCESS_SIG_MASK(p) (&(p)->sighand->siglock)
#define FREE_AREA_TYPE free_area_t
#define ZONE_TYPE zone_t
#ifdef CONFIG_SOFTWARE_SUSPEND2
#define IDE_UNSUSPEND ide_unsuspend();
#else
#define IDE_UNSUSPEND
#endif
#define IDE_SUSPEND ide_suspend();
#define EAT_MEMORY_FLAGS GFP_ATOMIC
extern void ide_unsuspend(void);
#define CPU0_MASK (1 << cpu_logical_map(0))

#define INACTIVE_LIST_FOR_EACH(procedure) \
do { \
 if (nr_inactive_pages) { \
  struct list_head * entry = inactive_list.prev; \
   while (entry != &inactive_list) { \
    struct page * page; \
    page = list_entry(entry, struct page, lru); \
    procedure(page); \
    entry = entry->prev; \
   } \
 } \
} while(0)

#define ACTIVE_LIST_FOR_EACH(procedure) \
do { \
 if (nr_active_pages) { \
  struct list_head * entry = active_list.prev; \
   while (entry != &active_list) { \
    struct page * page; \
    page = list_entry(entry, struct page, lru); \
    procedure(page); \
    entry = entry->prev; \
   } \
 } \
} while(0)

#endif /* Version specific defines */
#define WAKE_UP(p) signal_wake_up(p, 0)
#define RECALC_SIGPENDING recalc_sigpending()
#define FOR_EACH_THREAD_TASK_STRUCTS *p, *g
#define FOR_EACH_THREAD_START do_each_thread(g, p)
#define FOR_EACH_THREAD_END while_each_thread(g, p);
#define THREAD_PREEMPT_COUNT preempt_count()

struct rangechain {
	struct range * first;
	struct range * last;
	int size; /* size of the range ie sum (max-min+1) */
	int allocs;
	int frees;
	int debug;
	int timesusedoptimisation;
	char * name;
	struct range * lastaccessed, *prevtolastaccessed, *prevtoprev;
};

/*
 * We rely on ranges not fitting evenly into a page.
 * The last four bytes are used to store the number
 * of the page, to make saving & reloading pages simpler.
 */
struct range {
	unsigned long minimum;
	unsigned long maximum;
	struct range * next;
};

/* page backup entry */
struct pbe {
	struct page * origaddress;	/* Original address of page */
	struct page * address;		/* Address of copy of page */
	struct range * currentorigrange;
	struct range * currentdestrange;

	struct pagedir * pagedir;
};

struct pagedir {
	int pagedir_num;
	int pageset_size;
	int lastpageset_size;
	struct rangechain origranges;
	struct rangechain destranges;
	struct rangechain allocdranges;
};

struct pbelink {
	char dummy[sizeof(struct pbe) - sizeof(struct pbe *)];
	struct pbe * next;
};

#define SWAP_FILENAME_MAXLENGTH	32

struct suspend_header {
	__u32 version_code;
	unsigned long num_physpages;
	char machine[65];
	char version[65];
	int num_cpus;
	int page_size;
	unsigned long orig_mem_free;
	int num_range_pages;
	struct pagedir pagedir;
	struct range * unused_ranges;
	int pageset_2_size;
	int param0;
	int param1;
	int param2;
	int param3;
	int param4;
	int progress0;
	int progress1;
	int progress2;
	int progress3;
};

extern struct tq_struct suspend_tq;
extern int swsusp_default_console_level;
extern spinlock_t suspend_irq_lock __nosavedata;
extern unsigned long swsuspirqflags __nosavedata;
extern int max_async_ios;
extern int image_size_limit;
extern int now_resuming;
extern struct pagedir pagedir1, pagedir2;

struct pageset_sizes_result {
	int size1; /* Can't be unsigned - breaks MAX function */
	int size2;
	int size2low;
	int needmorespace;
};

/* */
extern int C_A_D;

#if defined(CONFIG_VT) && defined(CONFIG_VT_CONSOLE)
#define DEFAULT_SUSPEND_CONSOLE	(MAX_NR_CONSOLES-1)
#endif

#define TIMEOUT	(6 * HZ)	/* Timeout for stopping processes */
#define __ADDRESS(x)  ((unsigned long) phys_to_virt(x))
#define ADDRESS(x) __ADDRESS((x) << PAGE_SHIFT)

#define MB(x) ((x) >> (20 - PAGE_SHIFT))

/* References to section boundaries */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,99)
extern char _text[], _etext[], _edata[], __bss_start[], _end[];
extern char __nosave_begin[], __nosave_end[];

#else /* 2.4 version */

extern char _text, _etext, _edata, __bss_start, _end;
extern char __nosave_begin, __nosave_end;

#endif /* Version specific */

extern void ide_disk_unsuspend(int);
extern void ide_disk_suspend(void);
extern void vt_console_print(struct console *co, const char * b, unsigned count);
extern int try_to_free_pages_swsusp(int amount_needed);

#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
extern unsigned long swsusp_debug_state;
extern int swsusp_pool_level(void);
#define PRINTFREEMEM(desn) printlog(SUSPEND_MEMORY, SUSPEND_MEDIUM, \
		"Free memory %s: %d+%d.\n", desn, nr_free_pages(), swsusp_pool_level())
#else /* CONFIG_SOFTWARE_SUSPEND_DEBUG */
#define PRINTFREEMEM(desn) do { } while(0)
#endif /* CONFIG_SOFTWARE_SUSPEND_DEBUG */

#ifdef CONFIG_PREEMPT
#define PRINTPREEMPTCOUNT(desn) printlog(SUSPEND_FREEZER, SUSPEND_MEDIUM, \
		"Preempt count (CPU %d): %s: %d.\n", \
		smp_processor_id(), desn, THREAD_PREEMPT_COUNT)
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
#define swsusp_spin_lock_irq(lock) { \
	int count = THREAD_PREEMPT_COUNT; \
	printlog(SUSPEND_SPINLOCKS, SUSPEND_VERBOSE, "Spinlock %p. %d->%d.\n", lock, count, count+1); \
	spin_lock_irq(lock); \
}

#define swsusp_spin_lock_irqsave(lock, flags) { \
	int count = THREAD_PREEMPT_COUNT; \
	printlog(SUSPEND_SPINLOCKS, SUSPEND_VERBOSE, "Spinlock IRQSave %p. %d->%d.\n", lock, count, count+1); \
	spin_lock_irqsave(lock, flags); \
}

#define swsusp_spin_unlock_irq(lock) { \
	int count = THREAD_PREEMPT_COUNT; \
	printlog(SUSPEND_SPINLOCKS, SUSPEND_VERBOSE, "Spinunlock %p. %d->%d.\n", lock, count, count-1); \
	spin_unlock_irq(lock); \
}

#define swsusp_spin_unlock_irqrestore(lock, flags) { \
	int count = THREAD_PREEMPT_COUNT; \
	printlog(SUSPEND_SPINLOCKS, SUSPEND_VERBOSE, "Spinunlock IRQRestore %p. %d->%d\n", lock, count, count-1); \
	spin_unlock_irqrestore(lock, flags); \
}
#else /* No debug */
#define swsusp_spin_lock_irq(lock) spin_lock_irq(lock)
#define swsusp_spin_lock_irqsave(lock, flags) spin_lock_irqsave(lock, flags)
#define swsusp_spin_unlock_irq(lock) spin_unlock_irq(lock)
#define swsusp_spin_unlock_irqrestore(lock, flags) spin_unlock_irqrestore(lock, flags)
#endif
#else /* No preempt */
#define PRINTPREEMPTCOUNT(desn) do { } while(0)
#define swsusp_spin_lock_irq(lock) { \
	printlog(SUSPEND_SPINLOCKS, SUSPEND_VERBOSE, "Spinlock %p.\n", lock); \
	spin_lock_irq(lock); \
}

#define swsusp_spin_lock_irqsave(lock, flags) { \
	printlog(SUSPEND_SPINLOCKS, SUSPEND_VERBOSE, "Spinlock IRQSave %p.\n", lock); \
	spin_lock_irqsave(lock, flags); \
}

#define swsusp_spin_unlock_irq(lock) { \
	printlog(SUSPEND_SPINLOCKS, SUSPEND_VERBOSE, "Spinunlock %p.\n", lock); \
	spin_unlock_irq(lock); \
}

#define swsusp_spin_unlock_irqrestore(lock, flags) { \
	printlog(SUSPEND_SPINLOCKS, SUSPEND_VERBOSE, "Spinunlock IRQRestore %p.\n", lock); \
	spin_unlock_irqrestore(lock, flags); \
}
#endif

/* Local variables that should not be affected by save */
#define pageset1_size (pagedir1.pageset_size)
#define pageset2_size (pagedir2.pageset_size)

#define BITS_PER_PAGE (PAGE_SIZE * 8)
#define PAGES_PER_BITMAP ((max_mapnr + BITS_PER_PAGE - 1) / BITS_PER_PAGE)
#define BITMAP_ORDER (get_bitmask_order((PAGES_PER_BITMAP) - 1))

/*
 * XXX: We try to keep some more pages free so that I/O operations succeed
 * without paging. Might this be more?
 */
#define MIN_FREE_RAM (max_mapnr >> 7)

void prepare_status(int printalways, int clearbar, const char *fmt, ...);
void abort_suspend(const char *fmt, ...);

void thaw_processes(void);
extern int suspend_snprintf(char * buffer, int buffer_size, const char *fmt, ...);

/* ------ prepare_image.c ------ */
extern unsigned long get_grabbed_pages(int order);

/* ------ io.c ------ */
void do_suspend_sync(void);
int sanity_check_failed(char *reason);

/* ------ console.c ------ */
void check_shift_keys(int pause, char * message);
unsigned long update_status(unsigned long value, unsigned long maximum, const char *fmt, ...);

#define PBEPAGE_START(pbe) (((unsigned long) (pbe)) & PAGE_MASK)
extern unsigned long memory_for_plugins(void);
extern int expected_compression_ratio(void);

#define MAIN_STORAGE_NEEDED \
	((pageset1_size + pageset2_size) * expected_compression_ratio() / 100)

#define HEADER_STORAGE_NEEDED \
	(num_range_pages + 1 + \
	 header_storage_for_plugins())
	
#define STORAGE_NEEDED \
	(MAIN_STORAGE_NEEDED + HEADER_STORAGE_NEEDED)

#define RAM_TO_SUSPEND (1 + MAX((pageset1_size - pageset2_sizelow), 0) + MIN_FREE_RAM + \
	memory_for_plugins())

/* --------- Range chains ----------- */

#define RANGES_PER_PAGE (PAGE_SIZE / (sizeof(struct range)))
#define RANGE_PAGES_NEEDED(x) (((x)+RANGES_PER_PAGE-1)/RANGES_PER_PAGE)
#define RANGEPAGELINK(x) ((unsigned long *) ((((unsigned long) x) & PAGE_MASK) + PAGE_SIZE - sizeof(unsigned long)))

#define range_for_each(rangechain, rangepointer, value) \
if ((rangechain)->first) \
	for ((rangepointer) = (rangechain)->first, (value) = (rangepointer)->minimum; \
	     ((rangepointer) && ((rangepointer)->next || (value) <= (rangepointer)->maximum)); \
	     (((value) == (rangepointer)->maximum) ? \
		((rangepointer) = (rangepointer)->next, (value) = ((rangepointer) ? (rangepointer)->minimum : 0)) : \
	     	(value)++))

/*
 * When using compression and expected_compression > 0,
 * we allocate less swap entries, so GET_RANGE_NEXT can
 * validly run out of data to return.
 */
#define GET_RANGE_NEXT(currentrange, currentval) \
{ \
	if (currentrange) { \
		if ((currentval) == (currentrange)->maximum) { \
			if ((currentrange)->next) { \
				(currentrange) = (currentrange)->next; \
				(currentval) = (currentrange)->minimum; \
			} else { \
				(currentrange) = NULL; \
				(currentval) = 0; \
			} \
		} else \
			currentval++; \
	} \
}

struct range_entry {
	unsigned long thisvalue;
	struct range * rangeptr;
};

extern int max_ranges_used;
extern int num_range_pages;
int add_to_range_chain(struct rangechain * chain, unsigned long value);
void put_range_chain(struct rangechain * chain);
void print_chain(int debuglevel, struct rangechain * chain, int printasswap);
void set_chain_names(struct pagedir * p);
int free_ranges(void);
int append_to_range_chain(int chain, unsigned long min, unsigned long max);
void relativise_ranges(void);
void relativise_chain(struct rangechain * chain);
void absolutise_ranges(unsigned long * range_pages);
void absolutise_chain(struct rangechain * chain, unsigned long * range_pages);
unsigned long * get_rangepages_list(void);
void put_rangepages_list(unsigned long * range_pages);
int relocate_rangepages(unsigned long * rangepages);

extern struct range * first_range_page, * last_range_page;

#define RANGE_RELATIVE(x) (struct range *) ((((unsigned long) x) & (PAGE_SIZE - 1)) | \
				((*RANGEPAGELINK(x) & (PAGE_SIZE - 1)) << PAGE_SHIFT))
#define RANGE_ABSOLUTE(entry, list) (struct range *) ((((unsigned long) (entry)) & (PAGE_SIZE - 1)) | \
				list[((unsigned long) (entry)) >> PAGE_SHIFT])

extern unsigned long * inusemap;
extern unsigned long * pageset2map;

#define PAGENUMBER(page) (page-mem_map)
#define PAGEINDEX(page) ((PAGENUMBER(page))/(8*sizeof(unsigned long)))
#define PAGEBIT(page) ((int) ((PAGENUMBER(page))%(8 * sizeof(unsigned long))))

/* 
 * freepagesmap is used in two ways: 
 * - During suspend, to tag pages which are not used (to speed up count_data_pages);
 * - During resume, to tag pages which are is pagedir1. This does not tag pagedir2
 *   pages, so !== first use.
 */
#define PageInUse(page)		test_bit(PAGEBIT(page), &inusemap[PAGEINDEX(page)])
#define SetPageInUse(page)	set_bit(PAGEBIT(page), &inusemap[PAGEINDEX(page)])
#define ClearPageInUse(page)	clear_bit(PAGEBIT(page), &inusemap[PAGEINDEX(page)])

#define PagePageset2(page)	test_bit(PAGEBIT(page), &pageset2map[PAGEINDEX(page)])
#define SetPagePageset2(page)	set_bit(PAGEBIT(page), &pageset2map[PAGEINDEX(page)])
#define TestAndSetPagePageset2(page)	test_and_set_bit(PAGEBIT(page), &pageset2map[PAGEINDEX(page)])
#define TestAndClearPagePageset2(page)	test_and_clear_bit(PAGEBIT(page), &pageset2map[PAGEINDEX(page)])
#define ClearPagePageset2(page)	clear_bit(PAGEBIT(page), &pageset2map[PAGEINDEX(page)])

/* 
 *   #defs that are dependant on CONFIG_SOFTWARE_SUSPEND_DEBUG
 *
 */
#if defined(CONFIG_SOFTWARE_SUSPEND_DEBUG)
#define MDELAY(a)	if (TEST_ACTION_STATE(SUSPEND_SLOW)) { mdelay(a); } else { do { } while (0); }
#define beepOK \
	if (TEST_ACTION_STATE(SUSPEND_BEEP)) { \
		currentbeep += 200; \
		kd_mksound(currentbeep,HZ/8); \
		mdelay(150); \
	} else { \
		do { } while (0); \
	}
#define beepERR \
	if (TEST_ACTION_STATE(SUSPEND_BEEP)) { \
		kd_mksound(300,HZ/4); \
		mdelay(300); \
	} else { \
		do { } while (0); \
	} 

#else // #ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
#define MDELAY(a) do { } while (0)
#define beepOK do { } while (0)
#define beepERR do { } while (0)
#endif // #ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG

#define FILTER_PLUGIN 1
#define WRITER_PLUGIN 2

struct swsusp_filter_ops {
	int (*expected_compression) (void);
	struct list_head filter_list;
};

struct swsusp_writer_ops {

	/* Calls for allocating storage */

	long (*storage_available) (void); // Maximum size of image we can save
					  // (incl. space already allocated).
	
	unsigned long (*storage_allocated) (void);
					// Amount of storage already allocated
	int (*release_storage) (void);
	
	/* 
	 * Header space is allocated separately. Note that allocation
	 * of space for the header might result in allocated space 
	 * being stolen from the main pool if there is no unallocated
	 * space. We have to be able to allocate enough space for
	 * the header. We can eat memory to ensure there is enough
	 * for the main pool.
	 */
	long (*allocate_header_space) (unsigned long space_requested);
	int (*allocate_storage) (unsigned long space_requested);
	
	/* Read and write the metadata */	
	int (*write_header_init) (void);
	int (*write_header_chunk) (char * buffer_start, int buffer_size);
	int (*write_header_cleanup) (void);

	int (*read_header_init) (void);
	int (*read_header_chunk) (char * buffer_start, int buffer_size);
	int (*read_header_cleanup) (void);

	/* Prepare metadata to be saved (relativise/absolutise ranges) */
	int (*prepare_save_ranges) (void);
	int (*post_load_ranges) (void);
	
	/* Attempt to parse an image location */
	int (*parse_image_location) (char * buffer, int boot_time);

	/* Determine whether image exists that we can restore */
	int (*image_exists) (void);
	
	/* Destroy image if one exists */
	int (*invalidate_image) (void);
	
	/* Wait on I/O */
	int (*wait_on_io) (int flush_all);

	struct list_head writer_list;
};

#define SWSUSP_ASYNC 0
#define SWSUSP_SYNC  1

struct swsusp_plugin_ops {
	/* Functions common to filters and writers */
	int type;
	char * name;
	struct list_head plugin_list;
	unsigned long (*memory_needed) (void);
	unsigned long (*storage_needed) (void);
	int (*print_debug_info) (char * buffer, int size);
	int (*save_config_info) (char * buffer);
	void (*load_config_info) (char * buffer, int len);
	
	/* Writing the image proper */
	int (*write_init) (int stream_number);
	int (*write_chunk) (char * buffer_start);
	int (*write_cleanup) (void);

	/* Reading the image proper */
	int (*read_init) (int stream_number);
	int (*read_chunk) (char * buffer_start, int sync);
	int (*read_cleanup) (void);

	/* Reset plugin if image exists but reading aborted */
	void (*noresume_reset) (void);

	union {
		struct swsusp_filter_ops filter;
		struct swsusp_writer_ops writer;
	} ops;
};

extern struct swsusp_plugin_ops * active_writer;
extern struct list_head swsusp_filters, swsusp_writers, swsusp_plugins;
extern struct swsusp_plugin_ops * first_filter;
extern struct swsusp_plugin_ops * get_next_filter(struct swsusp_plugin_ops *);
extern int swsusp_register_plugin(struct swsusp_plugin_ops * plugin);

struct swsusp_proc_data {
	char * filename;
	int permissions;
	int type;
	union {
		struct {
			unsigned long * bit_vector;
			int bit;
		} bit;
		struct {
			int * variable;
			int minimum;
			int maximum;
		} integer;
		struct {
			unsigned long * variable;
			unsigned long minimum;
			unsigned long maximum;
		} ul;
		struct {
			void * read_proc;
			void * write_proc;
			void * data;
		} special;
	} data;
	struct list_head proc_data_list;
};

#define SWSUSP_PROC_DATA_CUSTOM		0
#define SWSUSP_PROC_DATA_BIT		1
#define SWSUSP_PROC_DATA_INTEGER	2
#define SWSUSP_PROC_DATA_UL		3

#ifdef CONFIG_SOFTWARE_SUSPEND_RELAXED_PROC
#define PROC_WRITEONLY 0222
#define PROC_READONLY 0444
#define PROC_RW 0666
#else
#define PROC_WRITEONLY 0200
#define PROC_READONLY 0400
#define PROC_RW 0600
#endif

struct proc_dir_entry * swsusp_register_procfile(struct swsusp_proc_data * swsusp_proc_data);
void swsusp_unregister_procfile(struct swsusp_proc_data * swsusp_proc_data);

struct debug_info_data
{
	int action;
	char * buffer;
	int buffer_size;
	int bytes_used;
};

#endif /* #ifndef SWSUSP_COMMON_H */
