/*
 * kernel/power/io.c
 *
 * Copyright (C) 1998-2001 Gabor Kuti <seasons@fornax.hu>
 * Copyright (C) 1998,2001,2002 Pavel Machek <pavel@suse.cz>
 * Copyright (C) 2002-2003 Florent Chabaud <fchabaud@free.fr>
 * Copyright (C) 2002-2004 Nigel Cunningham <ncunningham@users.sourceforge.net>
 *
 * This file is released under the GPLv2.
 *
 * This file contains data IO routines for swsusp.
 *
 */

#define SWSUSP_IO_C

#include <linux/suspend-common.h>

extern unsigned long orig_mem_free;
extern int swsusp_act_used;
extern int swsusp_lvl_used;
extern int swsusp_dbg_used;
extern void warmup_collision_cache(void);
extern int get_pageset1_load_addresses(void);
extern struct pagedir __nosavedata pagedir_resume;
extern struct range * unused_ranges;
extern int pm_prepare_console(void);

extern void get_next_pbe(struct pbe * pbe);
extern void get_first_pbe(struct pbe * pbe, struct pagedir * pagedir);

/* cleanup_finished_swsusp_io
 *
 * Description:	Very simple helper function to save #including all the
 * 		suspend-common code in fs/buffer.c and anywhere else we might
 * 		want to wait on suspend I/O in future.
 */

inline void cleanup_finished_swsusp_io(void)
{
	active_writer->ops.writer.wait_on_io(0);
}

/* fill_suspend_header()
 * 
 * Description:	Fill the suspend header structure.
 * Arguments:	struct suspend_header: Header data structure to be filled.
 */

static __inline__ void fill_suspend_header(struct suspend_header *sh)
{
	memset((char *)sh, 0, sizeof(*sh));

	sh->version_code = LINUX_VERSION_CODE;
	sh->num_physpages = num_physpages;
	sh->orig_mem_free = orig_mem_free;
	strncpy(sh->machine, system_utsname.machine, 65);
	strncpy(sh->version, system_utsname.version, 65);
	sh->num_cpus = NUM_CPUS;
	sh->page_size = PAGE_SIZE;
	sh->pagedir = pagedir1;
	sh->pagedir.origranges.first = pagedir1.origranges.first;
	sh->pagedir.destranges.first = pagedir1.destranges.first;
	sh->pagedir.allocdranges.first = pagedir1.allocdranges.first;
	sh->unused_ranges = unused_ranges;
	sh->num_range_pages = num_range_pages;
	sh->pageset_2_size = pagedir2.pageset_size;
	sh->param0 = swsusp_result;
	sh->param1 = swsusp_action;
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
	sh->param2 = swsusp_debug_state;
#endif
	sh->param3 = console_loglevel;

}

/* TODO: Handle page protection when saving and loading pages. */
#if 0
static void store_page_protection(void)
{
	pte_t * pte = NULL;
	int pageprot = 0;

	clear_bit(IO_RESTORE_PAGE_PROT, &io_info->flags);
	
	// Remove any page protection while we restore contents
	if (test_bit(IO_HANDLE_PAGE_PROT, &io_info->flags)) {
		pte = lookup_address((unsigned long) data_address);
		printlog(SUSPEND_IO, SUSPEND_VERBOSE, "pte %p", pte);
		if (pte) {
			pageprot = pte->pte_low & ~_PAGE_CHG_MASK;
			if (pageprot != pgprot_val(PAGE_KERNEL_NOCACHE)) {
				set_pte(pte, pte_modify(*pte, PAGE_KERNEL_NOCACHE));	
				set_bit(IO_RESTORE_PAGE_PROT, &io_info->flags);
			}
		}
	}
	io_info->pte = pte;
	io_info->pageprot = pageprot;
	
}

static void restore_page_protection(void)
{
		if (test_bit(IO_RESTORE_PAGE_PROT, &io_info->flags))
			set_pte(io_info->pte, pte_modify(*(io_info->pte), __pgprot(io_info->pageprot)));
}
#endif

/* write_pageset()
 *
 * Description:	Write a pageset to disk.
 * Arguments:	pagedir:	Pointer to the pagedir to be saved.
 * 		whichtowrite:	Controls what debugging output is printed.
 * Returns:	Zero on success or -1 on failure.
 */

int write_pageset(struct pagedir * pagedir, int whichtowrite)
{
	int nextupdate = 0, size, ret = 0, i, base = 0;
	int barmax = pagedir1.pageset_size + pagedir2.pageset_size;
	int start_time, end_time;
	long error = 0;
	struct pbe pbe;
	unsigned int origfree = nr_free_pages();
	struct list_head *filter;
	struct swsusp_plugin_ops * this_filter;

	PRINTFREEMEM("at start of write pageset");

	size = pagedir->pageset_size;
	if (!size)
		return 0;

	if (whichtowrite == 1) {
		prepare_status(1, 0, "Writing kernel & process data...");
		base = pagedir2.pageset_size;
	} else {
		prepare_status(1, 1, "Writing caches...");
	}	
	
	start_time = jiffies;

	print_chain(SUSPEND_HIGH, &pagedir->origranges, 0);
	print_chain(SUSPEND_HIGH, &pagedir->destranges, 0);

	/* Initialise page transformers */
	list_for_each(filter, &swsusp_filters) {
		this_filter = list_entry(filter, struct swsusp_plugin_ops, ops.filter.filter_list);
		if (this_filter->write_init)
			this_filter->write_init(whichtowrite);
	}

	/* Initialise writer */
	active_writer->write_init(whichtowrite);

	get_first_pbe(&pbe, pagedir);

	/* Write the data */
	for (i=0; i<size; i++) {
		char * virt = (char *) kmap(pbe.address);

		/* Status update */
		if (!(i&0x1FF)) 
			printnolog(SUSPEND_IO, SUSPEND_LOW, 0, ".");
		if (((i+base) >= nextupdate) || (!(i%(1 << (20 - PAGE_SHIFT)))))
			nextupdate = update_status(i + base, barmax, " %d/%d MB ", MB(base+i+1), MB(barmax));
		printnolog(SUSPEND_IO, SUSPEND_MEDIUM, 1, name_suspend "Writing pageset: %d/%d ", i + 1, size);

		/* Write */
		if ((ret = first_filter->write_chunk(virt))) {
			printk("Write chunk returned %d.\n", ret);
			abort_suspend("Failed to write a chunk of the image.\n");
			error = -1;
			goto write_pageset_free_buffers;
		}

		kunmap(pbe.address);

		/* Interactivity */
		check_shift_keys(0, NULL);

		if (TEST_RESULT_STATE(SUSPEND_ABORTED)) {
			abort_suspend("Aborting as requested.\n");
			error = -1;
			goto write_pageset_free_buffers;
		}

		/* Prepare next */
		get_next_pbe(&pbe);
	}

	printlog(SUSPEND_IO, SUSPEND_LOW, "|\n");

write_pageset_free_buffers:
	
	/* Flush data and cleanup */
	list_for_each(filter, &swsusp_filters) {
		this_filter = list_entry(filter, struct swsusp_plugin_ops, ops.filter.filter_list);
		if (this_filter->write_cleanup)
			this_filter->write_cleanup();
	}
	active_writer->write_cleanup();

	/* Statistics */
	end_time = jiffies;
	
	if ((end_time - start_time) && (!TEST_RESULT_STATE(SUSPEND_ABORTED)))
		printlog(SUSPEND_IO, SUSPEND_LOW, 
			"Time to write data: %d pages in %d jiffies => MB written per second: %lu.\n", 
			size,
			(end_time - start_time), 
			(MB((unsigned long) size) * HZ / (end_time - start_time)));

	PRINTFREEMEM("at end of write pageset");

	/* Sanity checking */
	if (nr_free_pages() != origfree) {
		abort_suspend("Number of free pages at start and end of write pageset don't match! (%d != %d)", origfree, nr_free_pages());
	}

	return error;
}

/* read_pageset()
 *
 * Description:	Read a pageset from disk.
 * Arguments:	pagedir:	Pointer to the pagedir to be saved.
 * 		whichtowrite:	Controls what debugging output is printed.
 * 		overwrittenpagesonly: Whether to read the whole pageset or only part.
 * Returns:	Zero on success or -1 on failure.
 */

int read_pageset(struct pagedir * pagedir, int whichtoread, int overwrittenpagesonly)
{
	int nextupdate = 0, result = 0, base = 0;
	int start_time, end_time, finish_at = pagedir->pageset_size;
	int barmax = pagedir1.pageset_size + pagedir2.pageset_size;
	long i;
	struct pbe pbe;
	struct list_head *filter;
	struct swsusp_plugin_ops * this_filter;

	PRINTFREEMEM("at start of read pageset");

	if (whichtoread == 1) {
		prepare_status(1, 1, "Reading kernel & process data...");
	} else {
		prepare_status(1, 0, "Reading caches...");
		if (overwrittenpagesonly)
			barmax = finish_at = MIN(pageset1_size, pageset2_size);
		else {
			base = pagedir1.pageset_size;
		}
	}	
	beepOK;
	
	start_time=jiffies;

	/* Initialise page transformers */
	list_for_each(filter, &swsusp_filters) {
		this_filter = list_entry(filter, struct swsusp_plugin_ops, ops.filter.filter_list);
		if ((this_filter->read_init) && (result = this_filter->read_init(whichtoread))) {
			abort_suspend("Failed to initialise a filter.\n");
			result = -1;
			goto read_pageset_free_buffers;
		}
	}

	/* Initialise writer */
	if ((result = active_writer->read_init(whichtoread))) {
		abort_suspend("Failed to initialise the writer.\n"); 
		result = -1;
		goto read_pageset_free_buffers;
	}

	get_first_pbe(&pbe, pagedir);

	/* Read the pages */
	for (i=0; i< finish_at; i++) {
		char * virt = (char *) kmap(pbe.address);

		/* Status */
		if (!(i&0x1FF)) 
			printnolog(SUSPEND_IO, SUSPEND_LOW, 0, ".");
		if (((i+base) >= nextupdate) || (!(i%(1 << (20 - PAGE_SHIFT)))))
			nextupdate = update_status(i+base, barmax, " %d/%d MB ", MB(base+i+1), MB(barmax));
		printnolog(SUSPEND_IO, SUSPEND_MEDIUM, 1, name_suspend "Reading pageset: %d/%d ", i + 1, finish_at);

		/* Read */
		if ((result = first_filter->read_chunk(virt, SWSUSP_ASYNC))) {
			abort_suspend("Failed to read a chunk of the image.\n");
			result = -1;
			goto read_pageset_free_buffers;
		}

		kunmap(pbe.address);

		/* Interactivity*/
		check_shift_keys(0, NULL);

		/* Prepare next */
		get_next_pbe(&pbe);
	}

	printlog(SUSPEND_IO, SUSPEND_LOW, "|\n");

read_pageset_free_buffers:

	/* Finish I/O, flush data and cleanup reads. */
	list_for_each(filter, &swsusp_filters) {
		this_filter = list_entry(filter, struct swsusp_plugin_ops, ops.filter.filter_list);
		if ((this_filter->read_cleanup) && (result = this_filter->read_cleanup()))
			abort_suspend("Failed to cleanup a filter.\n");
	}

	if ((result = active_writer->read_cleanup()))
		abort_suspend("Failed to cleanup the writer.\n");

	/* Statistics */
	end_time=jiffies;
	if ((end_time - start_time) && (!TEST_RESULT_STATE(SUSPEND_ABORTED)))
		printlog(SUSPEND_IO, SUSPEND_LOW, 
			"Time to read data: %d pages in %d jiffies => MB read per second: %lu.\n", 
			finish_at,
			(end_time - start_time), 
			(MB((unsigned long) finish_at) * HZ / (end_time - start_time)));

	PRINTFREEMEM("at end of read pageset");

	return result;
}

/* write_plugin_configs()
 *
 * Description:	Store the configuration for each plugin in the image header.
 * Returns:	Int: Zero on success, Error value otherwise.
 */
static int write_plugin_configs(void)
{
	struct list_head *plugin;
	struct swsusp_plugin_ops * this_plugin;
	char * buffer = (char *) get_zeroed_page(GFP_ATOMIC);
	int len;

	if (!buffer) {
		printk("Failed to allocate a buffer for saving plugin configuration info.\n");
		return -ENOMEM;
	}
		
	/* 
	 * We have to know which data goes with which plugin, so we at
	 * least write a length of zero for a plugin. Note that we are
	 * also assuming every plugin's config data takes <= PAGE_SIZE.
	 */

	/* For each plugin (in registration order) */
	list_for_each(plugin, &swsusp_plugins) {
		this_plugin = list_entry(plugin, struct swsusp_plugin_ops, plugin_list);

		/* Get the data from the plugin */
		len = 0;
		if (this_plugin->save_config_info)
			len = this_plugin->save_config_info(buffer);

		/* Save the size of the data and any data returned */
		active_writer->ops.writer.write_header_chunk((char *) &len, sizeof(int));
		if (len)
			active_writer->ops.writer.write_header_chunk(buffer, len);
	}

	free_pages((unsigned long) buffer, 0);
	return 0;
}

/* read_plugin_configs()
 *
 * Description:	Reload plugin configurations from the image header.
 * Returns:	Int. Zero on success, error value otherwise.
 */

static int read_plugin_configs(void)
{
	struct list_head *plugin;
	struct swsusp_plugin_ops * this_plugin;
	char * buffer = (char *) get_zeroed_page(GFP_ATOMIC);
	int len, result = 0;

	if (!buffer) {
		printk("Failed to allocate a buffer for reloading plugin configuration info.\n");
		return -ENOMEM;
	}
		
	/* For each plugin (in registration order) */
	list_for_each(plugin, &swsusp_plugins) {
		this_plugin = list_entry(plugin, struct swsusp_plugin_ops, plugin_list);

		/* Get the length of the data (if any) */
		result = active_writer->ops.writer.read_header_chunk((char *) &len, sizeof(int));
		if (!result) {
			printk("Failed to read the length of the next plugin's configuration data.\n");
			free_pages((unsigned long) buffer, 0);
			return -EINVAL;
		}

		/* Read any data and pass to the plugin */
		if (len) {
			active_writer->ops.writer.read_header_chunk(buffer, len);
			if (!this_plugin->save_config_info) {
				printk("Huh? Plugin %s appears to have a save_config_info, but not a load_config_info function!\n",
					this_plugin->name);
			} else
				this_plugin->load_config_info(buffer, len);
		}
	}

	free_pages((unsigned long) buffer, 0);
	return 0;
}

/* write_image_header()
 *
 * Description:	Write the image header after write the image proper.
 * Returns:	Int. Zero on success or -1 on failure.
 */

int write_image_header(void)
{
	int i, nextupdate = 0, ret;
	int total = pagedir1.pageset_size+pagedir2.pageset_size+2;
	int progress = total-1;
	unsigned long * rangepageslist = NULL;
	char * header_buffer = NULL;

	/* First, relativise all range information */
	rangepageslist = get_rangepages_list();
	if (!rangepageslist)
		return -1;

	print_chain(SUSPEND_HIGH, &pagedir1.origranges, 0);
	print_chain(SUSPEND_HIGH, &pagedir1.destranges, 0);
	print_chain(SUSPEND_HIGH, &pagedir1.allocdranges, 0);

	if (unused_ranges)
		unused_ranges = RANGE_RELATIVE(unused_ranges);
	
	relativise_chain(&pagedir1.origranges);
	relativise_chain(&pagedir1.destranges);
	relativise_chain(&pagedir1.allocdranges);

	if ((ret = active_writer->ops.writer.prepare_save_ranges())) {
		abort_suspend("Active writer's prepare_save_ranges function failed.\n");
		goto write_image_header_abort1;
	}

	relativise_ranges();

	/* Now prepare to write the header */
	if ((ret = active_writer->ops.writer.write_header_init())) {
		abort_suspend("Active writer's write_header_init function failed.\n");
		goto write_image_header_abort2;
	}

	/* Get a buffer */
	header_buffer = (char *) get_zeroed_page(GFP_ATOMIC);
	if (!header_buffer) {
		abort_suspend("Out of memory when trying to get page for header!\n");
		goto write_image_header_abort3;
	}

	/* Write the meta data */
	fill_suspend_header((struct suspend_header *) header_buffer);
	active_writer->ops.writer.write_header_chunk(header_buffer, sizeof(struct suspend_header));

	/* Write plugin configurations */
	if ((ret = write_plugin_configs())) {
		abort_suspend("Failed to write plugin configs.\n");
		goto write_image_header_abort3;
	}

	/* Write range pages */
	printlog(SUSPEND_IO, SUSPEND_LOW, name_suspend "Writing %d range pages.\n", num_range_pages);

	for (i=1; i<=num_range_pages; i++) {
		/* Status update */
		printnolog(SUSPEND_IO, SUSPEND_VERBOSE, 1, "%d/%d: %p.\n",
			i, num_range_pages, rangepageslist[i]);
		printlog(SUSPEND_IO, SUSPEND_LOW, "."); 

		if (i >= nextupdate)
			nextupdate = update_status(progress + i, total, NULL);

		/* Check for aborting/pausing */
		check_shift_keys(0, NULL);

		if (TEST_RESULT_STATE(SUSPEND_ABORTED)) {
			abort_suspend("Aborting as requested.\n");
			goto write_image_header_abort3;
		}
		
		/* Write one range page */
		active_writer->ops.writer.write_header_chunk((char *) rangepageslist[i], PAGE_SIZE);
		
		if (ret) {
			abort_suspend("Failed writing a page. Error number was %d.\n", ret);
			goto write_image_header_abort3;
		}
	}

	update_status(total - 1, total, NULL);

	/* Flush data and let writer cleanup */
	if (active_writer->ops.writer.write_header_cleanup()) {
		abort_suspend("Failed to cleanup writing header.\n");
		goto write_image_header_abort2;
	}

	if (TEST_RESULT_STATE(SUSPEND_ABORTED))
		goto write_image_header_abort2;

	printlog(SUSPEND_IO, SUSPEND_VERBOSE, "|\n");
	update_status(total, total, NULL);

	MDELAY(1000);
	free_pages((unsigned long) header_buffer, 0);

	return 0;

	/* 
	 * Aborting. We need to...
	 * - let the writer cleanup (if necessary)
	 * - revert ranges to absolute values
	 */
write_image_header_abort3:
	active_writer->ops.writer.write_header_cleanup();
	
write_image_header_abort2:
	absolutise_ranges(rangepageslist);

	put_rangepages_list(rangepageslist);
	rangepageslist = NULL;

	if (active_writer->ops.writer.post_load_ranges)
		active_writer->ops.writer.post_load_ranges();

write_image_header_abort1:
	if (!rangepageslist) {
		rangepageslist = get_rangepages_list();
		if (!rangepageslist)
			panic("Unable to allocate rangepageslist.\n");
	}

	absolutise_chain(&pagedir1.origranges, rangepageslist);
	print_chain(SUSPEND_HIGH, &pagedir1.origranges, 0);
	absolutise_chain(&pagedir1.destranges, rangepageslist);
	print_chain(SUSPEND_HIGH, &pagedir1.destranges, 0);
	absolutise_chain(&pagedir1.allocdranges, rangepageslist);
	print_chain(SUSPEND_HIGH, &pagedir1.allocdranges, 0);

	put_rangepages_list(rangepageslist);

	free_pages((unsigned long) header_buffer, 0);
	return -1;
}

#ifdef CONFIG_FBCON_SPLASHSCREEN
extern int splash_verbose(void);
#else
#define splash_verbose()
#endif

/* sanity_check_failed()
 * Description:	Complain when the image doesn't appear to match the booted
 * 		kernel. The user may press Control to invalidate the image
 * 		and continue booting, or Shift to reboot.
 * Arguments:	Char *. Pointer to a string explaining why we're moaning.
 */
int sanity_check_failed(char *reason)
{
	kd_mksound(300,HZ/4);
	mdelay(300);
	splash_verbose();
	printk(KERN_EMERG "=== Software Suspend ===\n\n");
	printk(KERN_EMERG "BIG FAT WARNING!! %s\n\n", reason);
	printk(KERN_EMERG "If you want to use the current suspend image, press Shift to reboot\n");
	printk(KERN_EMERG "and try again with the same kernel that you suspended from.\n\n");
	printk(KERN_EMERG "If you want to forget that image, press Control to continue.\n");
	printk(KERN_EMERG "Press SHIFT to reboot or CONTROL to continue booting with this kernel\n");
	while (!(shift_state & ((1 << KG_SHIFT) | (1 << KG_CTRL))))
		schedule();
	if (shift_state & (1 << KG_SHIFT))
		machine_restart(NULL);
	/* Wait for release */
	while (shift_state) 
		schedule(); 
	return -EPERM;
}

/* sanity_check()
 *
 * Description:	Perform a few checks, seeking to ensure that the kernel being
 * 		booted matches the one suspended. They need to match so we can
 * 		be _sure_ things will work. It is not absolutely impossible for
 * 		resuming from a different kernel to work, just not assured.
 * Arguments:	Struct suspend_header. The header which was saved at suspend
 * 		time.
 */
static int sanity_check(struct suspend_header *sh)
{
	if (sh->version_code != LINUX_VERSION_CODE)
		return sanity_check_failed("Incorrect kernel version");
	
	if (sh->num_physpages != num_physpages)
		return sanity_check_failed("Incorrect memory size");

	if (strncmp(sh->machine, system_utsname.machine, 65))
		return sanity_check_failed("Incorrect machine type");

	if (strncmp(sh->version, system_utsname.version, 65))
		return sanity_check_failed("Incorrect version");

	if (sh->num_cpus != NUM_CPUS)
		return sanity_check_failed("Incorrect number of cpus");

	if (sh->page_size != PAGE_SIZE)
		return sanity_check_failed("Incorrect PAGE_SIZE");

	return 0;
}

/* noresume_reset_plugins
 *
 * Description:	When we read the start of an image, plugins (and especially the
 * 		active writer) might need to reset data structures if we decide
 * 		to invalidate the image rather than resuming from it.
 */

static void noresume_reset_plugins(void)
{
	struct swsusp_plugin_ops * this_plugin;
	struct list_head *plugin;
	
	list_for_each(plugin, &swsusp_plugins) {
		this_plugin = list_entry(plugin, struct swsusp_plugin_ops, plugin_list);
		if (this_plugin->noresume_reset)
			this_plugin->noresume_reset();
	}
}

/* __read_primary_suspend_image
 *
 * Description:	Test for the existence of an image and attempt to load it.
 * Returns:	Int. Zero if image found and pageset1 successfully loaded.
 * 		Error if no image found or loaded.
 */
static int __read_primary_suspend_image(void)
{			
	int i, result = 0;
	unsigned long * range_pages;
	char * header_buffer = (char *) get_zeroed_page(GFP_ATOMIC);
	struct suspend_header * suspend_header;
	struct range * last_range_page = NULL;

	if (!header_buffer)
		return -ENOMEM;
	
	software_suspend_state |= SOFTWARE_SUSPEND_RUNNING;

	/* Check for an image */
	if (!(result = active_writer->ops.writer.image_exists())) {
		result = -ENODATA;
		noresume_reset_plugins();
		goto out;
	}

	/* Check for noresume command line option */
	if (software_suspend_state & SOFTWARE_SUSPEND_NORESUME_SPECIFIED) {
		active_writer->ops.writer.invalidate_image();
		result = -EINVAL;
		noresume_reset_plugins();
		goto out;
	}

	/* 
	 * Prepare the active writer for reading the image header. The
	 * activate writer might read its own configuration or set up
	 * a network connection here.
	 * 
	 * NB: This call may never return because there might be a signature
	 * for a different image such that we warn the user and they choose
	 * to reboot. (If the device ids look erroneous (2.4 vs 2.6) or the
	 * location of the image might be unavailable if it was stored on a
	 * network connection.
	 */
	if ((result = active_writer->ops.writer.read_header_init())) {
		noresume_reset_plugins();
		goto out;
	}
	
	/* Read suspend header */
	if ((result = active_writer->ops.writer.read_header_chunk(header_buffer, sizeof(struct suspend_header))) < 0) {
		noresume_reset_plugins();
		goto out;
	}
	
	suspend_header = (struct suspend_header *) header_buffer;

	/*
	 * NB: This call may also result in a reboot rather than returning.
	 */
	if (sanity_check(suspend_header)) { /* Is this the same machine? */
		active_writer->ops.writer.invalidate_image();
		result = -EINVAL;
		noresume_reset_plugins();
		goto out;
	}

	/*
	 * ---------------------------------------------------- 
	 * We have an image and it looks like it will load okay.
	 * ---------------------------------------------------- 
	 */

	/* Get metadata from header. Don't override commandline parameters.
	 *
	 * We don't need to save swsusp_state[4] because it's not used during resume
	 * and will be restored with the image anyway
	 */
	
	orig_mem_free = suspend_header->orig_mem_free;
	memcpy((char *) &pagedir_resume, (char *) &suspend_header->pagedir, sizeof(pagedir_resume));
	unused_ranges = suspend_header->unused_ranges;
	num_range_pages = suspend_header->num_range_pages;
	swsusp_result = suspend_header->param0;
	if (!swsusp_act_used)
		swsusp_action = suspend_header->param1;
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
	if (!swsusp_dbg_used)
		swsusp_debug_state = suspend_header->param2;
#endif
	if (!swsusp_lvl_used)
		console_loglevel = suspend_header->param3;
	pagedir1.pageset_size = pagedir_resume.pageset_size;
	pagedir2.pageset_size = suspend_header->pageset_2_size;

	/* Prepare the console */
	now_resuming = 1;
	if (pm_prepare_console())
		printk(name_suspend "Can't allocate a console... proceeding\n");

	/* Read plugin configurations */
	if ((result = read_plugin_configs())) {
		noresume_reset_plugins();
		goto out;
	}

	printnolog(SUSPEND_IO, SUSPEND_VERBOSE, 1, "");

	/* Read range pages */
	check_shift_keys(1, "About to read pagedir. ");

	printlog(SUSPEND_IO, SUSPEND_VERBOSE,name_suspend "Reading %d range pages.\n", num_range_pages);

	for (i=0; i < num_range_pages; i++) {
		/* Get a page into which we will load the data */
		struct range * this_range_page = (struct range *)get_zeroed_page(GFP_ATOMIC);
		if (!this_range_page) {
			abort_suspend("Unable to allocate a pagedir.");
			result = -ENOMEM;
			noresume_reset_plugins();
			goto outfreeingrangepages;
		}

		/* Link to previous page */
		if (i == 0)
			first_range_page = this_range_page;
		else
			*RANGEPAGELINK(last_range_page) = (i | (unsigned long) this_range_page);

		/* Read this page */
		printnolog(SUSPEND_IO, SUSPEND_VERBOSE, 0, "Reading range page %d [%p].\n", i, this_range_page);
		if ((result = active_writer->ops.writer.read_header_chunk((char *) this_range_page, PAGE_SIZE)) < 0) {
			printk("Active writer's read_header_chunk routine returned %d.\n", result);
			free_page((unsigned long) this_range_page);
			noresume_reset_plugins();
			goto outfreeingrangepages;
		}
		
		last_range_page = this_range_page;
	}
	
	/* Set the last page's link to its index */
	*RANGEPAGELINK(last_range_page) = i;

	/* Clean up after reading the header */
	if ((result = active_writer->ops.writer.read_header_cleanup())) {
		noresume_reset_plugins();
		goto outfreeingrangepages;
	}

	printlog(SUSPEND_IO, SUSPEND_VERBOSE, "Pagedir_resume.pageset_size is %d\n", pagedir_resume.pageset_size);

	check_shift_keys(1, NULL);

	/* Okay.
	 *
	 * Now we need to move the range pages to a place such that they won't
	 * get overwritten while being used when copying the original kernel
	 * back. To achieve this, we need to absolutise them where they are
	 * now, prepare a bitmap of pages that collide and then relativise the
	 * range pages again. Having done that, we can relocate the range
	 * pages so that they don't collide with the image being restored,
	 * and absolutise them in that location.
	 */
	
	range_pages = get_rangepages_list();
	
	if (!range_pages) {
		printk("Unable to allocate rangepageslist.\n");
		result = -ENOMEM;
		noresume_reset_plugins();
		goto outfreeingrangepages;
	}
	
	/* Absolutise ranges so they can be used for building the map of
	 * pages that will be overwritten. */
	absolutise_ranges(range_pages);
	absolutise_chain(&pagedir_resume.origranges, range_pages);

	/* Mark the pages used by the current and original kernels */
 	warmup_collision_cache();

	/* Prepare to move the pages so they don't conflict */
	relativise_chain(&pagedir_resume.origranges);
	relativise_ranges();
 
	/* Relocate the pages */
	relocate_rangepages(range_pages);

	/* Absolutise in final place */
	absolutise_ranges(range_pages);
	
	/* Done.
	 *
	 * Now we can absolutise all the pointers to the range chains.
	 */

	set_chain_names(&pagedir_resume);
	
	absolutise_chain(&pagedir_resume.origranges, range_pages);
	print_chain(SUSPEND_HIGH, &pagedir_resume.origranges, 0);
	
	/*
	 * We don't want the original destination ranges (the locations where
	 * the atomic copy of pageset1 was stored at suspend time); we release
	 * the chain's elements before getting new ones. (The kernel running
	 * right now could be using pages that were free when we suspended).
	 */

	absolutise_chain(&pagedir_resume.destranges, range_pages);
	
	absolutise_chain(&pagedir_resume.allocdranges, range_pages);
	print_chain(SUSPEND_HIGH, &pagedir_resume.allocdranges, 0);

	if (unused_ranges)
		unused_ranges = RANGE_ABSOLUTE(unused_ranges, range_pages);
	
	put_rangepages_list(range_pages);

	/* 
	 * The active writer should be using chains to record where it stored
	 * the data. Give it a chance to absolutise them.
	 */
	if (active_writer->ops.writer.post_load_ranges)
		active_writer->ops.writer.post_load_ranges();

	/* 
	 * Get the addresses of pages into which we will load the kernel to
	 * be copied back
	 */
	put_range_chain(&pagedir_resume.destranges);

	if (get_pageset1_load_addresses()) {
		result = -ENOMEM;
		noresume_reset_plugins();
		goto outfreeingrangepages;
	}

	print_chain(SUSPEND_HIGH, &pagedir_resume.destranges, 0);
	
	printlog(SUSPEND_IO, SUSPEND_VERBOSE,"\n");

	/* Read the original kernel back */
	check_shift_keys(1, "About to read pageset 1. ");

	if (read_pageset(&pagedir_resume, 1, 0)) {
		result = -EPERM;
		noresume_reset_plugins();
		goto outfreeingrangepages;
	}

	printlog(SUSPEND_IO, SUSPEND_VERBOSE,"\n|\n");
	
	PRINTFREEMEM("after loading image.");
	check_shift_keys(1, "About to restore original kernel. ");
	result = 0;
out:
	free_pages((unsigned long) header_buffer, 0);
	return result;
outfreeingrangepages:
	//FIXME Test i post loop and reset memory structures.
	{
		int j;
		struct range * this_range_page = first_range_page;
		struct range * next_range_page;
		for (j = 0; j < i; j++) {
			next_range_page = (struct range *) (((unsigned long) *RANGEPAGELINK(this_range_page)) & PAGE_MASK);
			free_page((unsigned long) this_range_page);
			this_range_page = next_range_page;
		}
	}
	goto out;
}

/* read_primary_suspend_image()
 *
 * Description:	Attempt to read the header and pageset1 of a suspend image.
 * 		Handle the outcome, complaining where appropriate.
 */
int read_primary_suspend_image(void)
{
	int error;

	beepOK;
	error = __read_primary_suspend_image();

	switch (error) {
		case 0:
		case -ENODATA:
			beepOK;
		case -EINVAL:	/* non fatal error */
			software_suspend_state &= ~3;
			MDELAY(1000);
			return error;
			break;
		case -EIO:
			printk(KERN_CRIT name_suspend "I/O error\n");
			beepERR;
			break;
		case -ENOENT:
			printk(KERN_CRIT name_suspend "No such file or directory\n");
			break;
		case -EPERM:
			printk(KERN_CRIT name_suspend "Sanity check error\n");
			beepERR;
			break;
		default:
			printk(KERN_CRIT name_suspend "Error %d resuming\n", error);
			beepERR;
			break;
	}
	abort_suspend("Error %d in read_primary_suspend_image",error);
	return error;
}

/* read_secondary_pagedir()
 *
 * Description:	Read in part or all of pageset2 of an image, depending upon
 * 		whether we are suspending and have only overwritten a portion
 * 		with pageset1 pages, or are resuming and need to read them 
 * 		all.
 * Arguments:	Int. Boolean. Read only pages which would have been
 * 		overwritten by pageset1?
 * Returns:	Int. Zero if no error, otherwise the error value.
 */
int read_secondary_pagedir(int overwrittenpagesonly)
{
	int result = 0;

	if (!pageset2_size)
		return 0;

	printlog(SUSPEND_IO, SUSPEND_VERBOSE, "Beginning of read_secondary_pagedir: "); 

	result = read_pageset(&pagedir2, 2, overwrittenpagesonly);

	update_status(100, 100, NULL);
	check_shift_keys(1, "Pagedir 2 read. ");

	printlog(SUSPEND_IO, SUSPEND_VERBOSE,"\n");
	return result;
}
