/*
 * kernel/power/lzf_compress.c
 *
 * Copyright (C) 2003 Marc Lehmann <pcg@goof.com>
 * Copyright (C) 2003,2004 Nigel Cunningham <ncunningham@users.sourceforge.net>
 *
 * This file is released under the GPLv2.
 *
 * This file contains data compression routines for swsusp,
 * using LZH compression.
 *
 */

#include <linux/suspend-common.h>

#define FLAG_DISABLE 1
static unsigned long flags;
static int expected_lzf_compression = 0;

#define DISABLED (flags & FLAG_DISABLE)

/*
 * size of hashtable is (1 << HLOG) * sizeof (char *)
 * decompression is independent of the hash table size
 * the difference between 15 and 14 is very small
 * for small blocks (and 14 is also faster).
 * For a low-memory configuration, use HLOG == 13;
 * For best compression, use 15 or 16.
 */
#ifndef HLOG
# define HLOG 14
#endif

/*
 * sacrifice some compression quality in favour of compression speed.
 * (roughly 1-2% worse compression for large blocks and
 * 9-10% for small, redundant, blocks and >>20% better speed in both cases)
 * In short: enable this for binary data, disable this for text data.
 */
#ifndef ULTRA_FAST
# define ULTRA_FAST 1
#endif

#define STRICT_ALIGN 0
#define USE_MEMCPY 0
#define INIT_HTAB 0

#include "lzf/lzf_c.c"
#include "lzf/lzf_d.c"

static struct swsusp_plugin_ops lzf_compression_ops;
static struct swsusp_plugin_ops * next_driver;

static void *compression_workspace = NULL;
static u8 *local_buffer = NULL;
static u8 *page_buffer = NULL;
static unsigned int bufofs;

static __nosavedata unsigned long bytes_in = 0, bytes_out = 0;

/* allocate_compression_space
 *
 * Description:	Allocate space for use in [de]compressing our data.
 *		Each call must have a matching call to free_memory.
 * Returns:	Int: Zero if successful, -ENONEM otherwise.
 */

static inline int allocate_compression_space(void)
{
	BUG_ON(compression_workspace);

	compression_workspace = vmalloc_32((1<<HLOG)*sizeof(char *));
	if (!compression_workspace) {
		printk(KERN_WARNING "Failed to allocate %d bytes for lzf workspace\n", (1<<HLOG)*sizeof(char *));
		return -ENOMEM;
	}
	
	return 0;
}

/* free_zlib_memory
 *
 * Description:	Frees memory allocated by the allocation routine (above).
 */

static inline void free_memory(void)
{
	if (!compression_workspace)
		return;

	vfree(compression_workspace);
	compression_workspace = NULL;
}

/* ---- Local buffer management ---- */

/* allocate_local_buffer
 *
 * Description:	Allocates a page of memory for buffering output.
 * Returns:	Int: Zero if successful, -ENONEM otherwise.
 */

static int allocate_local_buffer(void)
{
	if (!local_buffer) {
		local_buffer = (char *) get_zeroed_page(GFP_ATOMIC);
	
		if (!local_buffer) {
			printk(KERN_ERR "Failed to allocate the local buffer for lzf compression driver.\n");
			return -ENOMEM;
		}
	}

	if (!page_buffer) {
		page_buffer = (char *) get_zeroed_page(GFP_ATOMIC);
	
		if (!page_buffer) {
			printk(KERN_ERR "Failed to allocate the page buffer for lzf compression driver.\n");
			return -ENOMEM;
		}
	}

	return 0;
}

/* free_local_buffer
 *
 * Description:	Frees memory allocated for buffering output.
 */

static inline void free_local_buffer(void)
{
	if (local_buffer)
		free_pages((unsigned long) local_buffer, 0);

	local_buffer = NULL;

	if (page_buffer)
		free_pages((unsigned long) page_buffer, 0);

	page_buffer = NULL;
}

/* ---- Exported functions ---- */

/* write_init()
 *
 * Description:	Allocate buffers and prepare to compress data.
 * Arguments:	Stream_number:	Ignored.
 * Returns:	Zero on success, -ENOMEM if unable to vmalloc.
 */

static int lzf_write_init(int stream_number)
{
	int result;
	
	next_driver = get_next_filter(&lzf_compression_ops);

	if (DISABLED)
		return 0;

	if ((result = allocate_compression_space()))
		return result;
	
	if ((result = allocate_local_buffer()))
		return result;

	/* Only reset the stats if starting to write an image */
	if (stream_number == 2)
	        bytes_in = bytes_out = 0;
	
        bufofs = 0;
        
	return 0;
}

/* lzf_write()
 *
 * Description:	Helper function for write_chunk. Write the compressed data.
 * Arguments:	u8*:		Output buffer to be written.
 * 		unsigned int:	Length of buffer.
 * Return:	int:		Result to be passed back to caller.
 */

static int lzf_write (u8 *buffer, unsigned int len)
{
	int ret;

        bytes_out += len;

	while (len + bufofs > PAGE_SIZE) {
		unsigned int chunk = PAGE_SIZE - bufofs;
        	memcpy (local_buffer + bufofs, buffer, chunk);
                buffer += chunk;
                len -= chunk;
                bufofs = 0;
                if ((ret = next_driver->write_chunk(local_buffer)) < 0)
                	return ret;
        }
        memcpy (local_buffer + bufofs, buffer, len);
        bufofs += len;
        return 0;
}

/* lzf_write_chunk()
 *
 * Description:	Compress a page of data, buffering output and passing on
 * 		filled pages to the next plugin in the pipeline.
 * Arguments:	Buffer_start:	Pointer to a buffer of size PAGE_SIZE, 
 * 				containing data to be compressed.
 * Returns:	0 on success. Otherwise the error is that returned by later
 * 		plugins, -ECHILD if we have a broken pipeline or -EPERM if
 * 		zlib errs.
 */

static int lzf_write_chunk(char * buffer_start)
{
	int ret; 
        u16 len;
	
	if (!next_driver) {
		printk("Compression Driver: Argh! No one wants my output!");
		return -ECHILD;
	}

	/* Pass data through if compression disabled */
	if (DISABLED) {
		/* Push directly to next driver */
		return next_driver->write_chunk(buffer_start);
	}

        bytes_in += PAGE_SIZE;
		
        len = lzf_compress (buffer_start, PAGE_SIZE, page_buffer, PAGE_SIZE - 3, compression_workspace);
        if ((ret = lzf_write ((u8 *)&len, 2)) < 0)
        	return ret;

        if (len) // some compression
        	return lzf_write (page_buffer, len);
        else
               	return lzf_write (buffer_start, PAGE_SIZE);
}

/* write_cleanup()
 *
 * Description: Write unflushed data and free workspace.
 * Returns:	Result of writing last page.
 */

static int lzf_write_cleanup(void)
{
	int ret;
	
	if (DISABLED)
		return 0;

       	ret = next_driver->write_chunk(local_buffer);

	free_memory();
	free_local_buffer();

	return ret;
}

/* read_init()
 *
 * Description:	Prepare to read a new stream of data.
 * Arguments:	int: Section of image about to be read.
 * Returns:	int: Zero on success, error number otherwise.
 */

static int lzf_read_init(int stream_number)
{
	int result;

	next_driver = get_next_filter(&lzf_compression_ops);

	if (DISABLED)
		return 0;

	if ((result = allocate_local_buffer()))
		return result;

        bufofs = PAGE_SIZE;

	return 0;
}

/* lzf_read()
 *
 * Description:	Read data into compression buffer.
 * Arguments:	u8 *:		Address of the buffer.
 * 		unsigned int:	Length
 * Returns:	int:		Result of reading the image chunk.
 */

static int lzf_read (u8 *buffer, unsigned int len)
{
	int ret;

	if (DISABLED)
		return next_driver->read_chunk(buffer, PAGE_SIZE);

	while (len + bufofs > PAGE_SIZE) {
        	unsigned int chunk = PAGE_SIZE - bufofs;
        	memcpy (buffer, local_buffer + bufofs, chunk);
                buffer += chunk;
                len -= chunk;
                bufofs = 0;
                if ((ret = next_driver->read_chunk(local_buffer, SWSUSP_SYNC)) < 0)
                        return ret;
        }
        memcpy (buffer, local_buffer + bufofs, len);
        bufofs += len;
        return 0;
}

/* lzf_read_chunk()
 *
 * Description:	Retrieve data from later plugins and decompress it until the
 * 		input buffer is filled.
 * Arguments:	Buffer_start: 	Pointer to a buffer of size PAGE_SIZE.
 * 		Sync:		Whether the previous plugin (or core) wants its
 * 				data synchronously.
 * Returns:	Zero if successful. Error condition from me or from downstream
 * 		on failure.
 */

static int lzf_read_chunk(char * buffer_start, int sync)
{
	int ret; 
        u16 len;
	struct swsusp_plugin_ops * next_driver = get_next_filter(&lzf_compression_ops);

	if (!next_driver) {
		printk("Compression Driver: Argh! No one wants to feed me data!");
		return -ECHILD;
	}
	
	/* Pass request through if compression disabled */
	if (DISABLED) {
		/* Push request directly to next driver */
		return next_driver->read_chunk(buffer_start, sync);
	}
	
	/* 
	 * All our reads must be synchronous - we can't decompress
	 * data that hasn't been read yet.
	 */

        if ((ret = lzf_read ((u8 *)&len, 2)) < 0)
        	return ret;

        if (len == 0) { // uncompressed
        	return lzf_read (buffer_start, PAGE_SIZE);
        } else { // compressed
                if ((ret = lzf_read (page_buffer, len)) < 0)
                	return ret;
                ret = lzf_decompress (page_buffer, len, buffer_start, PAGE_SIZE);
                if (ret != PAGE_SIZE)
			return -EPERM; // why EPERM??
                return 0;
        }

        BUG_ON (1);
}

/* read_cleanup()
 *
 * Description:	Clean up after reading part or all of a stream of data.
 * Returns:	int: Always zero. Never fails.
 */

static int lzf_read_cleanup(void)
{
	if (DISABLED)
		return 0;

	free_local_buffer();
	return 0;
}

/* lzf_print_debug_stats
 *
 * Description:	Print information to be recorded for debugging purposes into a
 * 		buffer.
 * Arguments:	buffer: Pointer to a buffer into which the debug info will be
 * 			printed.
 * 		size:	Size of the buffer.
 * Returns:	Number of characters written to the buffer.
 */

static int lzf_print_debug_stats(char * buffer, int size)
{
	int pages_in = bytes_in >> PAGE_SHIFT, pages_out = bytes_out >> PAGE_SHIFT;
	int len;
	
	if (DISABLED) {
		len = suspend_snprintf(buffer, size, "- LZF Compression disabled.\n");
		return len;
	}
		
	//Output the compression ratio achieved.
	len = suspend_snprintf(buffer, size, "- LZF Compressor enabled.\n");
	if (pages_in)
		len+= suspend_snprintf(buffer+len, size - len,
		  "  Compressed %ld bytes into %ld.\n  Image compressed by %d percent.\n",
		  bytes_in, bytes_out, (pages_in - pages_out) * 100 / pages_in);
	return len;
}

/* compression_memory_needed
 *
 * Description:	Tell the caller how much memory we need to operate during
 * 		suspend/resume.
 * Returns:	Unsigned long. Maximum number of bytes of memory required for
 * 		operation.
 */

static unsigned long lzf_memory_needed(void)
{
	if (DISABLED)
		return 0;

	return PAGE_SIZE * 2 + (1<<HLOG)*sizeof(char *);
}

/* lzf_save_config_info
 *
 * Description:	Save informaton needed when reloading the image at resume time.
 * Arguments:	Buffer:		Pointer to a buffer of size PAGE_SIZE.
 * Returns:	Number of bytes used for saving our data.
 */

static int lzf_save_config_info(char * buffer)
{
	*((unsigned long *) buffer) = bytes_in;
	*((unsigned long *) (buffer + sizeof(unsigned long))) = bytes_out;
	*((unsigned long *) (buffer + 2 * sizeof(unsigned long))) = flags;
	return 3 * sizeof(unsigned long);
}

/* lzf_load_config_info
 *
 * Description:	Reload information needed for decompressing the image at 
 * 		resume time.
 * Arguments:	Buffer:		Pointer to the start of the data.
 *		Size:		Number of bytes that were saved.
 */

static void lzf_load_config_info(char * buffer, int size)
{
	if (size != 3 * sizeof(unsigned long)) {
		printk("Huh? Size of plugin data is not right. (%d instead of %d).\n",
			size, 3 * sizeof(unsigned long));
		return;
	}

	bytes_in = *((unsigned long *) buffer);
	bytes_out = *((unsigned long *) (buffer + sizeof(unsigned long)));
	flags = *((unsigned long *) (buffer + 2 * sizeof(unsigned long)));
	return;
}

/* lzf_get_expected_compression
 * 
 * Description:	Returns the expected ratio between data passed into this plugin
 * 		and the amount of data output when writing.
 * Returns:	100 if the plugin is disabled. Otherwise the value set by the
 * 		user via our proc entry.
 */

static int lzf_get_expected_compression(void)
{
	if (DISABLED)
		return 100;

	return 100 - expected_lzf_compression;
}

/*
 * data for our proc entries.
 */

static struct swsusp_proc_data expected_compression_proc_data = {
	.filename			= "expected_lzf_compression",
	.permissions			= PROC_RW,
	.type				= SWSUSP_PROC_DATA_INTEGER,
	.data = {
		.integer = {
			.variable	= &expected_lzf_compression,
			.minimum	= 0,
			.maximum	= 100,
		}
	}
};

static struct swsusp_proc_data disable_compression_proc_data = {
	.filename			= "disable_lzf_compression",
	.permissions			= PROC_RW,
	.type				= SWSUSP_PROC_DATA_INTEGER,
	.data = {
		.ul = {
			.variable	= &flags,
			.minimum	= 0,
			.maximum	= 1,
		}
	}
};

/*
 * Ops structure.
 */

static struct swsusp_plugin_ops lzf_compression_ops = {
	.type			= FILTER_PLUGIN,
	.name			= "LZF Page Compressor",
	.memory_needed 		= lzf_memory_needed,
	.print_debug_info	= lzf_print_debug_stats,
	.save_config_info	= lzf_save_config_info,
	.load_config_info	= lzf_load_config_info,
	.write_init		= lzf_write_init,
	.write_chunk		= lzf_write_chunk,
	.write_cleanup		= lzf_write_cleanup,
	.read_init		= lzf_read_init,
	.read_chunk		= lzf_read_chunk,
	.read_cleanup		= lzf_read_cleanup,
	.ops = {
		.filter = {
			.expected_compression = lzf_get_expected_compression,
		}
	}
};

/* ---- Registration ---- */

static __init int lzf_load(void)
{
	int result;

	printk("Software Suspend LZF Compression Driver v1.0\n");
	if (!(result = swsusp_register_plugin(&lzf_compression_ops))) {
		swsusp_register_procfile(&expected_compression_proc_data);
		swsusp_register_procfile(&disable_compression_proc_data);
	}
	return result;
}

__initcall(lzf_load);

