/*
 * kernel/power/memory_pool.c
 *  
 * Copyright (C) 2003,2004 Nigel Cunningham <ncunningham@users.sourceforge.net>
 *
 * This file is released under the GPLv2.
 *
 * It contains routines for managing the memory pool during software suspend
 * operation.
 * 
 * The memory pool is a pool of pages from which page allocations
 * are satisfied while we are suspending, and into which freed pages are
 * released. In this way, we can keep the image size static and consistent
 * while still using normal I/O routines to save the image and while saving
 * the image in two parts.
 * 
 * During suspend, almost all of the page allocations are order zero. Provision
 * is made for one order one and one order two allocation. This provision is 
 * utilised by the swapwriter for allocating memory which is used for structures
 * containing header page. (It could be made to use order zero allocations; this
 * just hasn't been done yet).
 */

#define SWSUSP_MEMORY_POOL_C
#include <linux/suspend-common.h>

/* The maximum order we handle */
#define MAX_POOL_ORDER 2

static unsigned long * memory_pool[MAX_POOL_ORDER + 1] = { NULL, NULL, NULL };
static int pool_level_limit[MAX_POOL_ORDER + 1] = { 0, 1, 1 };
static int pool_level[MAX_POOL_ORDER + 1] = { 0, 0, 0 };
static spinlock_t memory_pool_lock = SPIN_LOCK_UNLOCKED;

/* Abbreviations for debugging */
#define MEMPOOLVERBOSE SUSPEND_MEM_POOL, SUSPEND_VERBOSE
#define MEMPOOLMEDIUM SUSPEND_MEM_POOL, SUSPEND_MEDIUM
#define MEMPOOLHIGH SUSPEND_MEM_POOL, SUSPEND_HIGH

/* swsusp_pool_level()
 *
 * Description:	Returns the number of pages currently in the pool.
 * Returns:	Int.		Number of pages in the pool.
 */
int swsusp_pool_level(void)
{
	int order, sum = 0;

	for (order = 0; order <= MAX_POOL_ORDER; order++)
		sum+= pool_level[order] * (1 << order);

	return sum;
}

/* display_memory_pool_pages()
 *
 * Description:	Display the current contents of the memory pool.
 */	
void display_memory_pool_pages(void)
{
	unsigned long * next, flags;
	int order, count;

	swsusp_spin_lock_irqsave(&memory_pool_lock, flags);
	printlog(MEMPOOLVERBOSE, "Memory pool:\n");
	for (order = 0; order <= MAX_POOL_ORDER; order++) {
		printlog(MEMPOOLVERBOSE, "- Order %d:\n", order);
		next = memory_pool[order];
		count = 0;
		while (next) {
			printlog(MEMPOOLVERBOSE, "[%p] ", next);
			count++;
			next = (unsigned long *) *next;
		}
		if (count)
			printlog(MEMPOOLVERBOSE, "(%d entries)\n", count);
		else
			printlog(MEMPOOLVERBOSE, "(empty)\n");
	}
	spin_unlock_irqrestore(&memory_pool_lock, flags);
}

/* fill_swsusp_memory_pool()
 *
 * Description:	Fill the memory pool from the main free memory pool in the
 * 		first instance, or grabbed pages if that fails.
 * 		We allocate @sizesought order 0 pages, plus 1 each
 * 		of the higher order allocations.
 * Arguments:	int.		Number of order zero pages requested.
 * Returns:	int.		Number of order zero pages obtained.
 */
int fill_swsusp_memory_pool(int sizesought)
{
	int i = 0, order, orig_state = swsusp_state & USE_MEMORY_POOL;
	unsigned long *this = NULL;
	unsigned long flags;

	spin_lock_irqsave(&memory_pool_lock, flags);
	
	/* Pool must not be active for this to work */
	swsusp_state &= ~USE_MEMORY_POOL;

	pool_level_limit[0] = sizesought;

	for (order = MAX_POOL_ORDER; order >= 0; order--) {
		for (i = pool_level[order]; i < pool_level_limit[order]; i++) {
			this = (unsigned long *) __get_free_pages(GFP_ATOMIC, order);
			if (!this)
				this = (unsigned long *) get_grabbed_pages(order);
			if (!this) {
				printk(name_suspend "Error. Wanted %d pages wanted of order %d for swsusp memory pool, got %d.\n", 
					pool_level_limit[order], order, i - 1);
				break;
			}
			printlog(MEMPOOLHIGH, "Adding page %p to memory pool.\n", this);
			*this = (unsigned long) memory_pool[order];
			memory_pool[order] = this;
			pool_level[order]++;
		}
	}

	swsusp_state |= orig_state;

	spin_unlock_irqrestore(&memory_pool_lock, flags);

	display_memory_pool_pages();

	return 0;
}

/* empty_swsusp_memory_pool()
 *
 * Description:	Drain our memory pool.
 */
void empty_swsusp_memory_pool(void)
{
	unsigned long flags;
	int order;

	check_shift_keys(1, "About to drain swsusp memory pool.\n");
	display_memory_pool_pages();
	swsusp_spin_lock_irqsave(&memory_pool_lock, flags);

	/* Pool must not be active for this to work */
	swsusp_state &= ~USE_MEMORY_POOL;

	for (order = 0; order <= MAX_POOL_ORDER; order++) {
		while (memory_pool[order]) {
			unsigned long next = *memory_pool[order];
			printlog(MEMPOOLHIGH, 
				"Removing page %d (%lx) from memory pool.\n", 
				pool_level[order], memory_pool[order]);
			if (page_count(virt_to_page((unsigned long) memory_pool[order])) != 1)
				printk("Error freeing page %p from memory pool. Page count is %d (should be 1).\n",
				memory_pool[order],
				page_count(virt_to_page((unsigned long) memory_pool[order])));
			*memory_pool[order] = 0;
			free_pages((unsigned long) memory_pool[order], order);
			memory_pool[order] = (unsigned long *) next;
			pool_level[order]--;
		}
	}
	
	swsusp_spin_unlock_irqrestore(&memory_pool_lock, flags);
}

/* get_swsusp_pool_page()
 * 
 * Description:	Our equivalent to __alloc_pages (minus zone mask).
 * 		May be called from interrupt context.
 * Arguments:	unsigned int:	Mask. We really only care about __GFP_WAIT.
 * 		order:		The number of pages (1 << order) wanted.
 * Returns:	struct page *:	Pointer (possibly NULL) to pages allocated.
 */
struct page * get_swsusp_pool_page(unsigned int gfp_mask, unsigned int order)
{
	unsigned long * this = NULL, flags;
	struct page * page;
	static int printederror = 0;

	if (order > 0) {
		if (order <= MAX_POOL_ORDER) {
			swsusp_spin_lock_irqsave(&memory_pool_lock, flags);
			if (!pool_level[order]) {
				printk("No order %d allocation available.\n", order);
				swsusp_spin_unlock_irqrestore(&memory_pool_lock, flags);
				return NULL;
			}
			goto check_and_return;
		}
		
		if (!printederror) {
			printederror = 1;
			printk("Swsusp memory pool: Rejecting allocation of order %d.\n", order);
		}
		return NULL;
	}

try_again:
	if ((!memory_pool[0]) && (!(gfp_mask & __GFP_WAIT)))
		return NULL;

	while(!memory_pool[0]) {
		active_writer->ops.writer.wait_on_io(0);
		schedule();
	}

	swsusp_spin_lock_irqsave(&memory_pool_lock, flags);
	if (!memory_pool[order]) {
		swsusp_spin_unlock_irqrestore(&memory_pool_lock, flags);
		goto try_again;
	}
check_and_return:
	this = memory_pool[order];
	memory_pool[order] = (unsigned long *) *this;
	*this = 0;
	pool_level[order]--;
	printlog(MEMPOOLMEDIUM,
		"Allocating page %p from memory pool for %s (%d). %d remaining.\n",
		this, current->comm, current->pid, pool_level[order]);
	spin_unlock_irqrestore(&memory_pool_lock, flags);

	page = virt_to_page(this);
	
	if (page_count(page) != 1)
		printk("Error getting page %p from memory pool. Page count is %d (should be 1).\n",
		this,
		page_count(page));
	if (PageLRU(page))
		BUG();
	if (PageActive(page))
		BUG();

	return page;
}

/* free_swsusp_pool_pages()
 *
 * Description:	Our equivalent to  __free_pages. Put freed pages into the pool.
 * Arguments:	Struct page *:	First page to be freed.
 * 		Unsigned int:	Size of allocation being freed.
 */
void free_swsusp_pool_pages(struct page *page, unsigned int order)
{
	unsigned long *new_head, flags;
	int i;

	swsusp_spin_lock_irqsave(&memory_pool_lock, flags);
	
	if (order && (order <= MAX_POOL_ORDER) && (pool_level[order] < pool_level_limit[order])) {
		new_head = (unsigned long *) page_address(page);
		set_page_count(page, 1);
		*new_head = (unsigned long) memory_pool[order];
		memory_pool[order] = new_head;
		pool_level[order]++;
		swsusp_spin_unlock_irqrestore(&memory_pool_lock, flags);
		return;
	}

	for (i = 0; i < (1 << order); i++) {
		new_head = (unsigned long *) page_address(page + i);
		set_page_count(page + i, 1);
		*new_head = (unsigned long) memory_pool[0];
		memory_pool[0] = new_head;
		pool_level[0]++;
		printlog(MEMPOOLMEDIUM,
			"%s (%d) freeing page %p to memory pool. Now %d free.\n",
			current->comm, current->pid,
			new_head, pool_level[0]);
	}
	swsusp_spin_unlock_irqrestore(&memory_pool_lock, flags);
}
