/*
 * kernel/power/nulltransformer.c
 *
 * Copyright (C) 2003,2004 Nigel Cunningham <ncunningham@clear.net.nz>
 *
 * This file is released under the GPLv2.
 *
 * This file contains a dummy page transformer for testing and use as
 * a template for new page transformers.
 */

#include <linux/suspend-common.h>

struct swsusp_plugin_ops nulltransformer_ops;

/* write_init()
 * 
 * Description:	Routine called when Suspend wants to start to write a new
 * 		stream of data in the image. If this plugin supports being
 * 		disabled (by registering a proc entry itself), it may do
 * 		nothing here and in read init, and may simply pass data
 * 		through to the next plugin in the read/write_chunk routines.
 * Arguments:	Int. Stream number. Currently 1 or 2 depending upon which
 * 		pageset we are writing.
 * Returns:	Int. Zero on success. Non zero return values are interpreted
 * 		as error conditions and will lead to aborting saving the image.
 */
static int nulltrans_write_init(int stream_number)
{
	return 0;
}

/* write_chunk()
 *
 * Description:	Routine called when we are writing a chunk of the image.
 * 		The transformer should use the entire input buffer (size
 * 		PAGE_SIZE), not assuming that the data pointed to will be
 * 		available once we return. It should buffer output, before
 * 		sending it to the next filter one page at a time.
 * Arguments:	Char *. Pointer to a buffer of size PAGE_SIZE, containing
 * 		data to be transformed and/or passed on to the next plugin.
 * Returns:	Zero if no error condition exists. Error conditions from
 * 		later plugins should be passed back verbatim.
 */

static int nulltrans_write_chunk(char * buffer_start)
{
	struct swsusp_plugin_ops * next_driver = get_next_filter(&nulltransformer_ops);
	
	if (!next_driver) {
		printk("Null Pagetransformer: Argh! No one wants my output!");
		return -ECHILD;
	}

	return next_driver->write_chunk(buffer_start);
}

/* write_cleanup()
 *
 * Description:	Cleanup after writing. This routine should flush buffered
 * 		output (calling write_chunk of the next writer as usual),
 * 		and free any buffers allocated.
 * Returns: 	Zero. Never fails.
 */

static int nulltrans_write_cleanup(void)
{
	return 0;
}

/* read_init
 *
 * Description:	Prepare to read a new stream of data. Like write_init, this
 * 		routine should allocate space needed for buffering and perform
 * 		any other operations needed prior to reading the first chunk of
 * 		data.
 * Arguments:	Int. Portion of the image to be reread.
 * Returns:	Zero if no error. Error value otherwise.
 */

static int nulltrans_read_init(int stream_number)
{
	return 0;
}

/* read_chunk()
 *
 * Description:	Fill an input buffer with data, reversing any transformation
 * 		made when writing the image.
 * Arguments:	Char *. Pointer to the buffer (size PAGE_SIZE) to be filled.
 * 		Int Sync. Whether the reading of this chunk must be completed
 * 		immediately (1) or can be completed asynchronously (0). Since
 * 		we can't decompress/encrypt data that hasn't yet been read,
 * 		we should call later plugins with 1 and use the argument we
 * 		received if disabled. The advantage to using this argument is
 * 		that if no compression/encryption is enabled, the core code
 * 		will be able to request asynchronous reads from the writer,
 * 		resulting in faster throughput. Note that the writer may
 * 		also implement readahead, helping throughput when we are
 * 		operating synchronously here.
 * Returns:	Int: Zero if no error, error value of this or a later plugin
 * 		otherwise. (Errors from later plugins should be passed back
 * 		verbatim).
 */

static int nulltrans_read_chunk(char * buffer_start, int sync)
{
	struct swsusp_plugin_ops * next_driver = get_next_filter(&nulltransformer_ops);

	if (!next_driver) {
		printk("Null Pagetransformer: Argh! No one wants to feed me data!");
		return -ECHILD;
	}
	
	return next_driver->read_chunk(buffer_start, sync);
}

/* read_cleanup()
 *
 * Description:	Clean up after reading part or all of a stream of data.
 * 		(If the user aborts while writing pageset1, we reread
 * 		the part of pageset2 that was overwritten by the atomic
 * 		copy of the kernel). This routine should also free any
 * 		buffers allocated by the read_init routine.
 * Returns:	Zero on success. Should always succeed!
 */

static int nulltrans_read_cleanup(void)
{
	return 0;
}

/* print_debug_stats()
 *
 * Description:	Write debugging information to a buffer, for display
 * 		at the end of a cycle or when the user does cat
 * 		/proc/swsusp/debug_info. To prevent buffer overruns,
 * 		suspend_snprintf should be used. It is like vsn_printf,
 * 		but returns the number of bytes actually written (not
 * 		what would have been written if the buffer was big
 * 		enough! - see vsnprint man page).
 * Arguments:	Char *. Pointer to the buffer in which to store info.
 * 		Int. Size of the buffer. Probably less than PAGE_SIZE!
 * Returns:	Number of bytes written.
 */
extern int suspend_snprintf(char * buffer, int buffer_size, const char *fmt, ...);

static int nulltrans_print_debug_stats(char * buffer, int size)
{
	return 0;
}

static unsigned long nulltrans_memory_needed(void)
{
	return 0;
}

struct swsusp_plugin_ops nulltransformer_ops = {
	.type			= FILTER_PLUGIN,
	.name			= "Null Page Transformer",
	.memory_needed 		= nulltrans_memory_needed,
	.print_debug_info	= nulltrans_print_debug_stats,
	.write_init		= nulltrans_write_init,
	.write_chunk		= nulltrans_write_chunk,
	.write_cleanup		= nulltrans_write_cleanup,
	.read_init 		= nulltrans_read_init,
	.read_chunk		= nulltrans_read_chunk,
	.read_cleanup		= nulltrans_read_cleanup,
};

/* ---- Registration ---- */

static __init int nulltrans_load(void)
{
	printk("Software Suspend Null Page Transformer v1.0\n");
	return swsusp_register_plugin(&nulltransformer_ops);
}

__initcall(nulltrans_load);

