/*
 * Nullwriter.c
 *
 * Copyright 2003 Nigel Cunningham <ncunningham@clear.net.nz>
 *
 * Distributed under GPLv2.
 * 
 * This file encapsulates functions for testing the calls to
 * read/write an image.
 *
 */

#include <linux/suspend-common.h>

#define CAPACITY 200
static unsigned long available = CAPACITY;
static unsigned long mainpool_allocated = 0;
static unsigned long header_allocated = 0;

static unsigned long nullwriter_storage_allocated(void)
{
	return mainpool_allocated + header_allocated;
}

static long nullwriter_storage_available(void)
{
	return (long) (mainpool_allocated + header_allocated + available);
}

static int nullwriter_release_storage(void)
{
#ifdef CONFIG_SOFTWARE_SUSPEND_KEEP_IMAGE
	if ((TEST_ACTION_STATE(SUSPEND_KEEP_IMAGE)) && (now_resuming))
		return 0;
#endif

	available += (mainpool_allocated + header_allocated);
	header_allocated = mainpool_allocated = 0;

	return 0;
}

static long nullwriter_allocate_header_space(unsigned long space_requested)
{
	long extra_wanted = (space_requested - header_allocated);

	if (extra_wanted <= 0)
		return space_requested;
	
	if (extra_wanted < available) {
		long stolen = MIN((long) mainpool_allocated, (extra_wanted - (long) available));
		header_allocated += stolen;
		mainpool_allocated -= stolen;
	}

	header_allocated += MIN(extra_wanted, (long) available);

	return header_allocated;
}

static int nullwriter_allocate_storage(unsigned long space_requested)
{
	
	long extra_wanted = (space_requested - mainpool_allocated);
	long got = MIN(extra_wanted, (long) available);

	if (extra_wanted <= 0)
		return space_requested;

	mainpool_allocated += got;
	available -= got;

	return mainpool_allocated;
}

static int nullwriter_write_header_init(void)
{
	return 0;
}

static int nullwriter_write_header_chunk(char * buffer, int buffer_size)
{
	return 0;
}

static int nullwriter_write_header_cleanup(void)
{
	return 0;
}

/* ------------------------- HEADER READING ------------------------- */

static int nullwriter_read_header_init(void)
{
	return 0;
}

static int nullwriter_read_header_chunk(char * buffer, int buffer_size)
{
	return buffer_size;
}

static int nullwriter_read_header_cleanup(void)
{
	return 0;
}

static int nullwriter_prepare_save_ranges(void)
{
	return 0;
}

static int nullwriter_post_load_ranges(void)
{
	return 0;
}

static int nullwriter_write_init(int stream_number)
{
	return 0;
}

static int nullwriter_write_chunk(char * buffer)
{
	return 0;
}

static int nullwriter_write_cleanup(void)
{
	return 0;
}

static int nullwriter_read_init(int stream_number)
{
	return 0;
}

static int nullwriter_read_chunk(char * buffer, int sync)
{
	return 0;
}

static int nullwriter_read_cleanup(void)
{
	return 0;
}

static int nullwriter_invalidate_image(void)
{
	return 0;
}

/*
 * workspace_size
 *
 * Description:
 * Returns the number of bytes of RAM needed for this
 * code to do its work. (Used when calculating whether
 * we have enough memory to be able to suspend & resume).
 *
 */
static unsigned long nullwriter_memory_needed(void)
{
	return 0;
}

/*
 * Storage needed
 *
 * Returns amount of space in the header required
 * for the nullwriter's data.
 *
 */
static unsigned long nullwriter_storage_needed(void)
{
	return 0;
}

/*
 * Wait on I/O
 *
 */

static int nullwriter_wait_on_io(int flush_all)
{
	return 0;
}

/*
 * Image_exists
 *
 */

static int nullwriter_image_exists(void)
{
	return 0;
}

/*
 * Parse Image Location
 *
 */

static int nullwriter_parse_image_location(char * commandline, int boot_time)
{
	if (strncmp(commandline, "null:", 5)) {
		printk(name_suspend "Commandline doesn't begin with 'null:'... nullwriter ignoring.\n");
		return 0;
	}
	
	return 1;
}

struct swsusp_plugin_ops nullwriterops = {
	.type					= WRITER_PLUGIN,
	.name					= "Null Writer",
	.memory_needed				= nullwriter_memory_needed,
	.storage_needed				= nullwriter_storage_needed,
	.write_init				= nullwriter_write_init,
	.write_chunk				= nullwriter_write_chunk,
	.write_cleanup				= nullwriter_write_cleanup,
	.read_init				= nullwriter_read_init,
	.read_chunk				= nullwriter_read_chunk,
	.read_cleanup				= nullwriter_read_cleanup,
	.ops = {
		.writer = {
			.storage_available 	= nullwriter_storage_available,
			.storage_allocated	= nullwriter_storage_allocated,
			.release_storage	= nullwriter_release_storage,
			.allocate_header_space	= nullwriter_allocate_header_space,
			.allocate_storage	= nullwriter_allocate_storage,
			.image_exists		= nullwriter_image_exists,
			.write_header_init	= nullwriter_write_header_init,
			.write_header_chunk	= nullwriter_write_header_chunk,
			.write_header_cleanup	= nullwriter_write_header_cleanup,
			.read_header_init	= nullwriter_read_header_init,
			.read_header_chunk	= nullwriter_read_header_chunk,
			.read_header_cleanup	= nullwriter_read_header_cleanup,
			.prepare_save_ranges	= nullwriter_prepare_save_ranges,
			.post_load_ranges	= nullwriter_post_load_ranges,
			.invalidate_image	= nullwriter_invalidate_image,
			.wait_on_io		= nullwriter_wait_on_io,
			.parse_image_location	= nullwriter_parse_image_location,
		}
	}
};

/* ---- Registration ---- */

static __init int nullwriter_load(void)
{
	printk("Software Suspend Null Writer v1.0\n");
	return swsusp_register_plugin(&nullwriterops);
}

__initcall(nullwriter_load);

