/*
 * kernel/power/pagedir.c
 *
 * Copyright (C) 1998-2001 Gabor Kuti <seasons@fornax.hu>
 * Copyright (C) 1998,2001,2002 Pavel Machek <pavel@suse.cz>
 * Copyright (C) 2002-2003 Florent Chabaud <fchabaud@free.fr>
 * Copyright (C) 2002-2004 Nigel Cunningham <ncunningham@clear.net.nz>
 *
 * This file is released under the GPLv2.
 *
 * Routines for handling pagesets.
 * Note that pbes aren't actually stored as such. They're stored as
 * ranges (extents is the term, I'm told).
 */

#define SWSUSP_PAGEDIR_C
#include <linux/suspend-common.h>
extern struct pagedir pagedir1, pagedir2, pagedir_resume;

/* setup_pbe_variable
 *
 * Description:	Set up one variable in a page backup entry from the range list.
 * Arguments:	unsigned long:		The variable which will contain the value.
 * 		struct range**:		Address of the pointer to the current range.
 * 		struct rangechain*:	Address of the rangechain we are traversing.
 */
void setup_pbe_variable(unsigned long * variable, struct range ** currentrange, struct rangechain * chain)
{
	*currentrange = chain->first;
	if (chain->first)
		*variable = chain->first->minimum;
	else {
		*variable = 0;
		printnolog(SUSPEND_RANGES, SUSPEND_VERBOSE, 1, "Initialising variable from empty chain (%s).\n", chain->name);
	}
}

/* get_first_pbe
 *
 * Description:	Get the first page backup entry for a pagedir.
 * Arguments:	struct pbe *:	Address of the page backup entry we're populating.
 * 		struct pagedir: Pagedir providing the data.
 */
void get_first_pbe(struct pbe * pbe, struct pagedir * pagedir)
{
	unsigned long currentorig, currentaddress;

	pbe->pagedir = pagedir;

	/* Get raw initial values */
	setup_pbe_variable((unsigned long *) &pbe->origaddress, &pbe->currentorigrange, &pagedir->origranges);
	setup_pbe_variable((unsigned long *) &pbe->address, &pbe->currentdestrange, &pagedir->destranges);

	/* Convert to range values */
	currentorig = (unsigned long) pbe->origaddress;
	currentaddress = (unsigned long) pbe->address;

	pbe->origaddress = mem_map + currentorig;
	pbe->address = mem_map + currentaddress;

	if ((currentaddress < 0) || (currentaddress > max_mapnr))
		panic("Argh! Destination range value %ld is invalid!", currentaddress);

	printnolog(SUSPEND_PAGESETS, SUSPEND_VERBOSE, 1,
		"origaddress: %lx. address: %lx.",
		pbe->origaddress,
		pbe->address);
}

/* get_next_pbe
 *
 * Description:	Get the next page backup entry in a pagedir.
 * Arguments:	struct pbe *:	Address of the pbe we're updating.
 */
void get_next_pbe(struct pbe * pbe)
{
	unsigned long currentorig, currentaddress;

	/* Convert to range values */
	currentorig = (pbe->origaddress - mem_map);
	currentaddress = (pbe->address - mem_map);

	/* Update values */
	GET_RANGE_NEXT(pbe->currentorigrange, currentorig);
	GET_RANGE_NEXT(pbe->currentdestrange, currentaddress);
	
	pbe->origaddress = mem_map + currentorig;
	pbe->address = mem_map + currentaddress;
	
	printnolog(SUSPEND_PAGESETS, SUSPEND_VERBOSE, 1,
		"origaddress: %lx. address: %lx.",
		pbe->origaddress,
		pbe->address);
}

/*
 * --------------------------------------------------------------------------------------
 *
 * 	Local Page Flags routines.
 *
 * 	Rather than using the rare and precious flags in struct page, we allocate
 * 	our own bitmaps dynamically.
 * 
 */

/* clear_map
 *
 * Description:	Clear an array used to store local page flags.
 * Arguments:	unsigned long *:	The pagemap to be cleared.
 */

static void clear_map(unsigned long * pagemap)
{
	int i;
	for(i=0; i <= ((max_mapnr - 1) >> (generic_fls(sizeof(unsigned long) * 8 - 1))); i++)
		pagemap[i]=0;
}

/* allocatemap
 *
 * Description:	Allocate a bitmap for local page flags.
 * Arguments:	unsigned long **:	Pointer to the bitmap.
 * 		int:			Whether to set nosave flags for the
 * 					newly allocated pages.
 * Note:	This looks suboptimal, but remember that we might be allocating
 * 		the Nosave bitmap here.
 */
int allocatemap(unsigned long ** pagemap, int setnosave)
{
	unsigned long * check;
	int i;
	if (*pagemap) {
		printk("Error. Pagemap already allocated.\n");
		clear_map(*pagemap);
	} else {
		check = (unsigned long *) __get_free_pages(GFP_ATOMIC, BITMAP_ORDER);
		if (!check) {
			abort_suspend("Error. Unable to allocate memory for pagemap.");
			return 1;
		}
		clear_map(check);
		*pagemap = check;
		if (setnosave) {
			struct page * firstpage = virt_to_page((unsigned long) check);
			for (i = 0; i < (1 << BITMAP_ORDER); i++)
				SetPageNosave(firstpage + i);
		}
	}
	return 0;
}

/* freemap
 *
 * Description:	Free a local pageflags bitmap.
 * Arguments:	unsigned long **: Pointer to the bitmap being freed.
 * Note:	Map being freed might be Nosave.
 */
int freemap(unsigned long ** pagemap)
{
	int i;
	if (!*pagemap)
		return 1;
	else {
		struct page * firstpage = virt_to_page((unsigned long) *pagemap);
		for (i = 0; i < (1 << BITMAP_ORDER); i++)
			ClearPageNosave(firstpage + i);
		free_pages((unsigned long) *pagemap, BITMAP_ORDER);
		*pagemap = NULL;
		return 0;
	}
}

//---------------------------------------------------------------------------------

/* copy_pageset1
 *
 * Description:	Make the atomic copy of pageset1.
 */
void copy_pageset1(void)
{
	int i = 0;
	struct pbe pbe;

	get_first_pbe(&pbe, &pagedir1);

	for (i = 0; i < pageset1_size; i++) {
		copy_page((char *) page_address(pbe.address), 
				(char *) page_address(pbe.origaddress));
		get_next_pbe(&pbe);
	}
}

/* free_pagedir
 *
 * Description:	Free a previously allocated pagedir.
 * Arguments:	struct pagedir *:	Pointer to the pagedir being freed.
 */
void free_pagedir(struct pagedir * p)
{
	PRINTFREEMEM("at start of free_pagedir");

	if (p->allocdranges.first) {
		/* Free allocated pages */
		struct range * rangepointer;
		unsigned long pagenumber;
		range_for_each(&p->allocdranges, rangepointer, pagenumber) {
			ClearPageNosave(mem_map+pagenumber);
			free_page((unsigned long) page_address(mem_map+pagenumber));
			check_shift_keys(0, NULL);
		}
	}

	/* For pagedir 2, destranges == origranges */
	if (p->pagedir_num == 2)
		p->destranges.first = NULL;
	
	put_range_chain(&p->origranges);
	put_range_chain(&p->destranges);
	put_range_chain(&p->allocdranges);

	PRINTFREEMEM("at end of free_pagedir");
	printlog(SUSPEND_SWAP, SUSPEND_MEDIUM, "Pageset size was %d.\n", p->pageset_size);
	p->pageset_size = 0;
}

/* allocate_extra_pagedir_memory
 *
 * Description:	Allocate memory for making the atomic copy of pagedir1 in the
 * 		case where it is bigger than pagedir2.
 * Arguments:	struct pagedir *: 	The pagedir for which we should 
 * 					allocate memory.
 * 		int:			Size of pageset 1.
 * 		int:			Size of pageset 2.
 * Result:	int. Zero on success. One if unable to allocate enough memory.
 */
int allocate_extra_pagedir_memory(struct pagedir * p, int pageset_size, int alloc_from)
{
	int num_to_alloc = pageset_size - alloc_from - p->allocdranges.size;

	prepare_status(0, 0, "Preparing page directory.");

	PRINTFREEMEM("at start of allocate_extra_pagedir_memory");

	if (num_to_alloc < 1)
		num_to_alloc = 0;

	if (num_to_alloc) {
		int i, numnosaveallocated=0;
		int origallocd = p->allocdranges.size;
	
		PRINTFREEMEM("prior to attempt");
		for(i=1; i <= num_to_alloc; i++) {
			struct page * newpage;
			unsigned long virt;
			virt = get_grabbed_pages(0);
			if (!virt) {
				p->pageset_size = i;
				printlog(SUSPEND_PAGESETS, SUSPEND_VERBOSE,
					"   Allocated (extra) memory for pages from %d-%d (%d pages).\n",
					alloc_from, pageset_size - 1, p->allocdranges.size - origallocd);
				printk("Couldn't get enough yet. %d pages short.\n",
					num_to_alloc - i + 1);
				PRINTFREEMEM("at abort of allocate_extra_pagedir_memory");
				return 1;
			}
			newpage = virt_to_page(virt);
			SetPageNosave(newpage);
			add_to_range_chain(&p->allocdranges, newpage - mem_map);
			numnosaveallocated++;
			check_shift_keys(0, NULL);
		}
		printlog(SUSPEND_PAGESETS, SUSPEND_VERBOSE,"   Allocated (extra) memory for pages from %d-%d (%d pages).\n", alloc_from, pageset_size - 1, p->allocdranges.size - origallocd);
	}

	p->pageset_size = pageset_size;

	PRINTFREEMEM("at end of allocate_extra_pagedir_memory");
	return 0;
}

/* mark_page_pageset2
 * 
 * Description:	This looks a bit pointless, but it seems necessary for 2.6.
 * 		I'll look at it some more.
 * Arguments:	struct page *:	The page to be marked as Pageset2.
 */
static inline void mark_page_pageset2(struct page * page)
{
	SetPagePageset2(page);
}

/* mark_pages_for_pageset2
 *
 * Description:	Mark unshared pages in processes not needed for suspend as
 * 		being able to be written out in a separate pagedir.
 * 		HighMem pages are simply marked as pageset2. They won't be
 * 		needed during suspend.
 * 		Since we don't know which highmem pages are free, we can't
 * 		return a count any more.
 */

void mark_pages_for_pageset2(void)
{
	int i, numpageset2 = 0;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,99)
	struct zone * zone;
#endif

	if (max_mapnr != num_physpages) {
		abort_suspend("mapnr is not expected");
		return;
	}
	
	/* Start by marking no pages as pageset2 */
	for (i = 0; i < max_mapnr; i++)
		ClearPagePageset2(mem_map+i);

	/* Add LRU pages */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,99)
	for_each_zone(zone) {
		spin_lock(&zone->lru_lock);
		INACTIVE_LIST_FOR_EACH(SetPagePageset2);
		ACTIVE_LIST_FOR_EACH(SetPagePageset2);
		spin_unlock(&zone->lru_lock);
	}
#else
	spin_lock(&pagemap_lru_lock);
	INACTIVE_LIST_FOR_EACH(SetPagePageset2);
	ACTIVE_LIST_FOR_EACH(SetPagePageset2);
	spin_unlock(&pagemap_lru_lock);
#endif

	/* Ensure range pages are not Pageset2 */
	if (num_range_pages) {
		unsigned long * range_pages;
		range_pages = get_rangepages_list();
		if (!range_pages) {
			abort_suspend("Unable to get list of range pages. Number of range pages is %d.", num_range_pages);
			return;
		}
		for (i = 1; i <= num_range_pages; i++) {
			struct page * page;
			page = virt_to_page(range_pages[i]);
			if (PagePageset2(page)) {	// Must be assigned by the time recalc stats is called
				printlog(SUSPEND_PAGESETS, SUSPEND_ERROR, "Pagedir[%d] was marked as pageset2 - unmarking.\n", i);
				ClearPagePageset2(page);
				numpageset2--;
			}
		}
		put_rangepages_list(range_pages);
	}

	/* Finally, ensure HighMem pages are pageset 2 and that Slab pages are
	 * not. */

	for (i = 0; i < max_mapnr; i++) {
		if (PageHighMem(mem_map+i))
			if (!TestAndSetPagePageset2(mem_map+i))
				numpageset2++;
		if (PageSlab(mem_map+i))
			if (TestAndClearPagePageset2(mem_map+i)) {
				printk("Found slab page marked pageset 2.\n");
				numpageset2--;
			}
	}
}

/* warmup_collision_cache
 *
 * Description:	Mark the pages which are used by the original kernel.
 */
void warmup_collision_cache(void) {
	int i;
	struct range * rangepointer = NULL;
	unsigned long pagenumber;
	
	allocatemap(&inusemap, 0);/* Doesn't get deallocated because forgotten
				   when we copy PageDir1 back. Doesn't matter
				   if collides because not used during copy back.
				 */
	printlog(SUSPEND_IO, SUSPEND_VERBOSE, "Setting up pagedir cache...");
	for (i = 0; i < max_mapnr; i++)
		ClearPageInUse(mem_map+i);

	range_for_each(&pagedir_resume.origranges, rangepointer, pagenumber)
		SetPageInUse(mem_map+pagenumber);

	print_chain(SUSPEND_MEDIUM, &pagedir_resume.origranges, 0);
	printlog(SUSPEND_IO, SUSPEND_VERBOSE, "%d", i);
	printlog(SUSPEND_IO, SUSPEND_VERBOSE, "|\n");
}

/* get_pageset1_load_addresses
 * 
 * Description: We check here that pagedir & pages it points to won't collide
 * 		with pages where we're going to restore from the loaded pages
 * 		later.
 * Returns:	Zero on success, one if couldn't find enough pages (shouldn't
 * 		happen).
 */

int get_pageset1_load_addresses(void)
{
	int i, nrdone = 0, result = 0;
	void **eaten_memory = NULL;
	void **c = eaten_memory, *f, *addr;
	struct page * pageaddr = NULL;

	// Because we're trying to make this work when we're saving as much memory as possible
	// we need to remember the pages we reject here and then free them when we're done.
	
	printlog(SUSPEND_IO, SUSPEND_VERBOSE,"\nAllocating memory.\n");
	for(i=0; i < pagedir_resume.pageset_size; i++) {
		while ((addr = (void *) get_zeroed_page(GFP_ATOMIC))) {
			memset(addr, 0, PAGE_SIZE);
			pageaddr = virt_to_page(addr);
			if (!PageInUse(pageaddr)) {
				break;
			}
			eaten_memory = addr;
			*eaten_memory = c;
			c = eaten_memory;
		}
		if (!addr) {
			abort_suspend("Error: Ran out of memory seeking locations for reloading data.\n");
			result = 1;
			break;
		}
		add_to_range_chain(&pagedir_resume.destranges, pageaddr - mem_map);
		nrdone++;
	}

	// Free unwanted memory
	c = eaten_memory;
	while(c) {
		f = c;
		c = *c;
		if (f)
			free_page((unsigned long) f);
	}
	eaten_memory = NULL;
	
	printlog(SUSPEND_IO, SUSPEND_VERBOSE,"Check_pagedir: Prepared %d of %d pages.\n", nrdone, pagedir_resume.pageset_size);
	
	check_shift_keys(1, "Pagedir prepared. ");

	return result;
}

char* origrangesname = "original addresses";
char* destrangesname = "destination addresses";
char* allocdrangesname = "allocated addresses";

/* set_chain_names
 *
 * Description:	Set the chain names for a pagedir. (For debugging).
 * Arguments:	struct pagedir: The pagedir on which we want to set the names.
 */

void set_chain_names(struct pagedir * p)
{
	p->origranges.name = origrangesname;
	p->destranges.name = destrangesname;
	p->allocdranges.name = allocdrangesname;
}
