/*
 * Prepare_image.c
 *
 * We need to eat memory until we can:
 * 1. Perform the save without changing anything (RAM_NEEDED < max_mapnr)
 * 2. Fit it all in available space (active_writer->(available_space() >= STORAGE_NEEDED)
 * 3. Reload the pagedir and pageset1 to places that don't collide with their
 *    final destinations, not knowing to what extent the resumed kernel will
 *    overlap with the one loaded at boot time. I think the resumed kernel should overlap
 *    completely, but I don't want to rely on this as it is an unproven assumption. We
 *    therefore assume there will be no overlap at all (worse case).
 * 4. Meet the user's requested limit (if any) on the size of the image.
 *    The limit is in MB, so pages/256 (assuming 4K pages).
 *
 *    (Final test in save_image doesn't use EATEN_ENOUGH_MEMORY)
 */

#define SWSUSP_PREPARE_IMAGE_C

#include <linux/suspend-common.h>
#include <linux/notifier.h>
extern int pageset2_sizelow;
extern unsigned long orig_mem_free;
extern void mark_pages_for_pageset2(void);
extern int image_size_limit;
extern int fill_swsusp_memory_pool(int sizesought);

static int amount_eaten = 0, arefrozen = 0, numnosave = 0;

static void generate_free_page_map(void) 
{
	int i, loop;
	struct page * page;
	pg_data_t *pgdat = pgdat_list;
	unsigned type;
	unsigned long flags;

	for(i=0; i < max_mapnr; i++)
		SetPageInUse(mem_map+i);
	
	for (type=0;type < MAX_NR_ZONES; type++) {
		ZONE_TYPE *zone = pgdat->node_zones + type;
		int order = MAX_ORDER - 1;
		FREE_AREA_TYPE *area;
		struct list_head *head, *curr;
		swsusp_spin_lock_irqsave(&zone->lock, flags);
		do {
			int first_entry = 1;
			area = zone->free_area + order;
			head = &area->free_list;
			curr = head;

			for(;;) {
				if(!curr)
					break;
				if (first_entry) 
					first_entry--;
				else {
					page = list_entry(curr, struct page, list);
					for(loop=0; loop < (1 << order); loop++)
						ClearPageInUse(page+loop);
				}

				curr = curr->next;
				if (curr == head)
					break; 				
			}
		} while(order--);
		swsusp_spin_unlock_irqrestore(&zone->lock, flags);
	}
}

static int is_head_of_free_region(struct page * page)
{
	struct page * posn = page;

	while (((posn-mem_map) < max_mapnr) && (!PageInUse(posn))) 
		posn++;
	return (posn - page);
}

/*
 * header_storage_for_plugins
 *
 * Returns the amount of space needed to store configuration
 * data needed by the plugins prior to copying back the original
 * kernel. We can exclude data for pageset2 because it will be
 * available anyway once the kernel is copied back.
 */
unsigned long header_storage_for_plugins(void)
{
	struct list_head *plugin;
	struct swsusp_plugin_ops * this_plugin;
	unsigned long bytes = 0;
	
	list_for_each(plugin, &swsusp_plugins) {
		this_plugin = list_entry(plugin, struct swsusp_plugin_ops, plugin_list);
		if (this_plugin->storage_needed)
			bytes += this_plugin->storage_needed();
	}

	return ((bytes + PAGE_SIZE - 1) >> PAGE_SHIFT);
}

/*
 * expected_compression_ratio
 *
 * Returns the expected ratio between the amount of memory
 * to be saved and the amount of space required on the
 * storage device.
 */
int expected_compression_ratio(void)
{
	struct list_head *filter;
	struct swsusp_plugin_ops * this_filter;
	int ratio = 100;
	
	list_for_each(filter, &swsusp_filters) {
		this_filter = list_entry(filter, struct swsusp_plugin_ops, ops.filter.filter_list);
		if (this_filter->ops.filter.expected_compression)
			ratio = ratio * this_filter->ops.filter.expected_compression() / 100;
	}

	return ratio;
}

/*
 * memory_for_plugins
 *
 * Returns the amount of memory requested by plugins for
 * doing their work during the cycle.
 */

unsigned long memory_for_plugins(void)
{
	unsigned long bytes = 0;
	struct list_head *plugin;
	struct swsusp_plugin_ops * this_plugin;

	list_for_each(plugin, &swsusp_plugins) {
		this_plugin = list_entry(plugin, struct swsusp_plugin_ops, plugin_list);
		if (this_plugin->memory_needed)
			bytes += this_plugin->memory_needed();
	}

	return ((bytes + PAGE_SIZE - 1) >> PAGE_SHIFT);
}

static struct pageset_sizes_result count_data_pages(void)
{
	int chunk_size, loop, numfree = 0;
	int ranges = 0, currentrange = 0;
	int usepagedir2;
	int rangemin = 0;
	struct pageset_sizes_result result;
	struct range * rangepointer;
	unsigned long value;

	result.size1 = 0;
	result.size2 = 0;
	result.size2low = 0;
	result.needmorespace = 0;

	numnosave = 0;

	put_range_chain(&pagedir1.origranges);
	put_range_chain(&pagedir1.destranges);
	put_range_chain(&pagedir2.origranges);
	pagedir2.destranges.first = NULL;
	pagedir2.destranges.size = 0;

	generate_free_page_map();

	if (TEST_RESULT_STATE(SUSPEND_ABORTED)) {
		result.size1 = -1;
		result.size2 = -1;
		result.size2low = -1;
		result.needmorespace = 0;
		return result;
	}

	if (max_mapnr != num_physpages) {
		abort_suspend("Max_mapnr is not equal to num_physpages.");
		result.size1 = -1;
		result.size2 = -1;
		result.size2low = -1;
		result.needmorespace = 0;
		return result;
	}
	for (loop = 0; loop < max_mapnr; loop++) {
		if (!PageReserved(mem_map+loop)) {
			if (PageNosave(mem_map+loop)) {
				numnosave++;
				if (currentrange) {
					append_to_range_chain(currentrange, rangemin, loop - 1);
					rangemin = loop;
					currentrange = 0;
				}
				continue;
			}

			if ((chunk_size=is_head_of_free_region(mem_map+loop))!=0) {
				if (currentrange) {
					append_to_range_chain(currentrange, rangemin, loop - 1);
					rangemin = loop;
					currentrange = 0;
				}
				numfree += chunk_size;
				loop += chunk_size - 1;
				continue;
			}
		} else {
			if (PageNosave(mem_map+loop)) {
				/*
				abort_suspend("Reserved page marked nosave! (%lx)", ADDRESS(loop));
				result.size1 = -1;
				result.size2 = -1;
				result.size2low = -1;
				result.needmorespace = 0;
				return result;
				*/

				/* AGP pages can be Reserved and Nosave */
				numnosave++;
				if (currentrange) {
					append_to_range_chain(currentrange, rangemin, loop - 1);
					rangemin = loop;
					currentrange = 0;
				}
			}

			/*
			 * Just copy whole code segment. Hopefully it is not that big.
			 */
			if ((ADDRESS(loop) >= (unsigned long) &__nosave_begin) && 
			    (ADDRESS(loop) < (unsigned long)&__nosave_end)) {
				numnosave++;
				if (currentrange) {
					append_to_range_chain(currentrange, rangemin, loop - 1);
					rangemin = loop;
					currentrange = 0;
				}
				continue;
			}
			/* Hmm, perhaps copying all reserved pages is not too healthy as they may contain 
			   critical bios data? */

		};

		usepagedir2 = (PageHighMem(mem_map+loop) || PagePageset2(mem_map+loop));
		
		if (currentrange != (1 + usepagedir2)) {
			if (currentrange)
				append_to_range_chain(currentrange, rangemin, loop - 1);
			currentrange = usepagedir2 + 1;
			rangemin = loop;
			ranges++;
		}
		
		if (PagePageset2(mem_map+loop)) {
			result.size2++;
			if (!PageHighMem(mem_map+loop))
				result.size2low++;
		} else
			result.size1++;
	}
	
	if (currentrange)
		append_to_range_chain(currentrange, rangemin, loop - 1);
	check_shift_keys(0, NULL);

	if ((pagedir1.pageset_size) && (result.size1 > pagedir1.pageset_size))
		result.needmorespace = 1;
	if ((pagedir2.pageset_size) && (result.size2 > pagedir2.pageset_size))
		result.needmorespace = 1;
	printnolog(SUSPEND_RANGES, SUSPEND_MEDIUM, 0, "Counted %d ranges.\n", ranges);
	pagedir2.destranges.first = pagedir2.origranges.first;
	pagedir2.destranges.size = pagedir2.origranges.size;
	range_for_each(&pagedir1.allocdranges, rangepointer, value) {
		add_to_range_chain(&pagedir1.destranges, value);
		check_shift_keys(0, NULL);
	}

	printnolog(SUSPEND_EAT_MEMORY, SUSPEND_MEDIUM, 0,
		"Count data pages: Set1 (%d) + Set2 (%d) + Nosave (%d) + NumFree (%d) = %d.\n",
		result.size1, result.size2, numnosave, numfree,
		result.size1 + result.size2 + numnosave + numfree);
	return result;
}

static int amount_needed(void)
{

	int max1 = MAX( RAM_TO_SUSPEND - nr_free_pages() - amount_eaten,
			(STORAGE_NEEDED -  active_writer->ops.writer.storage_available()));
	return MAX( max1,
		    (image_size_limit > 0) ? (STORAGE_NEEDED - (image_size_limit << 8)) : 0);
}

#define EATEN_ENOUGH_MEMORY() (amount_needed() < 1)
unsigned long storage_available = 0;

void display_stats(void)
{ 
	unsigned long storage_allocated = active_writer->ops.writer.storage_allocated();
	printlog(SUSPEND_EAT_MEMORY, SUSPEND_MEDIUM, 
		"Free:%d+%d+%d=%d. Sets:%d,%d(%d). Header:%d. Nosave:%d-%d-%d=%d. Storage:%d/%lu(%lu). Needed:%d|%d|%d\n", 
		
		/* Free */
		nr_free_pages(), amount_eaten, swsusp_pool_level(),
		nr_free_pages() + amount_eaten + swsusp_pool_level(),
		
		/* Sets */
		pageset1_size,
		pageset2_size, pageset2_sizelow,

		/* Header */
		num_range_pages,

		/* Nosave */
		numnosave, pagedir1.allocdranges.size, amount_eaten,
		numnosave - pagedir1.allocdranges.size - amount_eaten,

		/* Storage - converted to pages for comparison */
		storage_allocated,
		STORAGE_NEEDED,
		storage_available,

		/* Needed */
		RAM_TO_SUSPEND - nr_free_pages() - amount_eaten,
		STORAGE_NEEDED - storage_available, 
		(image_size_limit > 0) ? (STORAGE_NEEDED - (image_size_limit << 8)) : 0);
}

/*
 * Eaten is the number of pages which have been eaten.
 * Pagedirincluded is the number of pages which have been allocated for the pagedir.
 */
extern int allocate_extra_pagedir_memory(struct pagedir * p, int pageset_size, int alloc_from);
extern unsigned long space_for_image;
extern unsigned char swsusp_memory_pool_active;

struct pageset_sizes_result recalculate_stats(void) 
{
	struct pageset_sizes_result result;

	mark_pages_for_pageset2();  /* Need to call this before getting pageset1_size! */
	result = count_data_pages();
	pageset2_sizelow = result.size2low;
	pageset2_size = result.size2;
	pageset1_size = result.size1;
	storage_available = active_writer->ops.writer.storage_available();
	return result;
}

static int update_image(void) 
{ 
	struct pageset_sizes_result result;
	int iteration = 0, orig_num_range_pages;
	int alloc_from;

	do {
		iteration++;

		orig_num_range_pages = num_range_pages;

		printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,"-- Iteration %d.\n", iteration);

		if (allocate_extra_pagedir_memory(&pagedir2, pageset2_size, pageset2_size)) {
			printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,
				"Still need to get more pages for pagedir 2.\n");
			return 1;
		}

		alloc_from = (pageset1_size > pageset2_sizelow ? pageset2_sizelow : pageset1_size);

		if (allocate_extra_pagedir_memory(&pagedir1, pageset1_size, alloc_from)) {
			printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,
				"Still need to get more pages for pagedir 1.\n");
			return 1;
		}

		if (active_writer->ops.writer.allocate_storage(MAIN_STORAGE_NEEDED)) {
			printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,
				"Still need to get more storage space for the image proper.\n");
			return 1;
		}

		if (active_writer->ops.writer.allocate_header_space(HEADER_STORAGE_NEEDED)) {
			printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,
				"Still need to get more storage space for header.\n");
			return 1;
		}

		result = recalculate_stats();
		display_stats();

	} while (((orig_num_range_pages < num_range_pages) || 
		   result.needmorespace ||
		   active_writer->ops.writer.storage_allocated() < (HEADER_STORAGE_NEEDED + MAIN_STORAGE_NEEDED)) 
		 && (!TEST_RESULT_STATE(SUSPEND_ABORTED)));
	
	return (amount_needed() > 0);
}

struct eaten_memory_t
{
	void * next;
	int order; /* Order of _this_ allocation */
};

struct eaten_memory_t *eaten_memory = NULL;

static void grab_free_memory(void)
{
	int order, k;
	struct eaten_memory_t *prev = eaten_memory;

	/*
	 * First, quickly eat all memory that's already free.
	 */
	
	for (order = MAX_ORDER - 1; order > -1; order--) {
		eaten_memory = (struct eaten_memory_t *) __get_free_pages(EAT_MEMORY_FLAGS, order);
		while (eaten_memory) {
			struct page * page = virt_to_page(eaten_memory);
			printnolog(SUSPEND_MEMORY, SUSPEND_HIGH, 0,
				"Grab free memory got %x, order %d.\n",
				eaten_memory, order);
			check_shift_keys(0, NULL);
			eaten_memory->next = prev;
			eaten_memory->order = order;
			prev = eaten_memory;
			amount_eaten += (1 << order);
			for (k=0; k < (1 << order); k++)
				SetPageNosave(page + k);
			eaten_memory = (struct eaten_memory_t *) __get_free_pages(EAT_MEMORY_FLAGS, order);
		}
	}

	eaten_memory = prev;
}

static void free_grabbed_memory(void)
{
	struct eaten_memory_t *next = NULL;
	int j, num_freed = 0;

	/* Free all eaten pages immediately */
	while(eaten_memory) {
		struct page * page = virt_to_page(eaten_memory);
		next = eaten_memory->next;
		for (j=0; j < (1 << (eaten_memory->order)); j++) {
			ClearPageNosave(page + j);
			num_freed++;
		}
		printnolog(SUSPEND_MEMORY, SUSPEND_HIGH, 0,
			"Free grabbed pages returning %x, order %d. (Fast path).\n",
			eaten_memory, eaten_memory->order);
		check_shift_keys(0, NULL);
		free_pages((unsigned long) eaten_memory, eaten_memory->order);
		eaten_memory = next;
	}
	amount_eaten -= num_freed;
}

unsigned long get_grabbed_pages(int order)
{
	struct eaten_memory_t *prev = NULL, *next=NULL, *this=eaten_memory;
	struct eaten_memory_t *alternative = NULL, *alternative_prev = NULL;
	int alternative_order = 999;
	unsigned long result;

	/* Get a single grabbed page for swsusp's use */
	/* We only look at pages of the desired order */
	/* Code could be added to split a higher order) */
	printnolog(SUSPEND_MEMORY, SUSPEND_HIGH, 0,
		"\nNew call to get_grabbed_pages. Seeking order %d.\n",
		order);

	while(this) {
		next = this->next;
		if (this->order == order) {
			struct page * page = virt_to_page(this);
			int j;
			if (prev)
				prev->next = next;
			else
				eaten_memory = next;
			for (j=0; j < (1 << order); j++) {
				ClearPageNosave(page + j);
				clear_page(page_address(page + j));
			}
			amount_eaten -= (1 << order);
			printnolog(SUSPEND_MEMORY, SUSPEND_HIGH, 0,
				"Get grabbed pages returning %x, order %d. (Fast path).\n",
				this, order);
			check_shift_keys(0, NULL);
			return (unsigned long) this;
		}
		if ((this->order < alternative_order) && (this->order > order)) {
			alternative_prev = prev;
			alternative = this;
			alternative_order = this->order;
			printnolog(SUSPEND_MEMORY, SUSPEND_HIGH, 0,
				"New alternative: %p, order %d.\n", alternative, alternative->order);
		}
		prev = this;
		this = next;
	}

	/* Maybe we didn't eat any memory - try normal get */
	if (!alternative) {
		printnolog(SUSPEND_MEMORY, SUSPEND_HIGH, 0,
			"No alternative. Trying __get_free_pages...");
		this = (struct eaten_memory_t *) __get_free_pages(GFP_ATOMIC, order);
		printnolog(SUSPEND_MEMORY, SUSPEND_HIGH, 0,
			"Result is %p, order %d.\n",
			this, order);
		check_shift_keys(0, NULL);
		return (unsigned long) this;
	}

	if (alternative_prev)
		alternative_prev->next = alternative->next;
	else
		eaten_memory = alternative->next;


	{
		struct page * page = virt_to_page(alternative);
		int j;
		for (j=0; j < (1 << (alternative_order)); j++) {
			ClearPageNosave(page + j);
			clear_page(page_address(page + j));
		}
		printnolog(SUSPEND_MEMORY, SUSPEND_HIGH, 0,
			"Freeing %p, Order %d.\n",
			alternative, alternative_order);
		free_pages((unsigned long) alternative, alternative_order);
		amount_eaten -= (1 << alternative_order);
	}

	/* Get the chunk we want to return */
	result = __get_free_pages(EAT_MEMORY_FLAGS, order);
	printnolog(SUSPEND_MEMORY, SUSPEND_HIGH, 0,
		"Got %p, Order %d back.\n",
		result, order);
	clear_page(result);

	/* Grab the rest */
	grab_free_memory();
	
	printnolog(SUSPEND_MEMORY, SUSPEND_HIGH, 0,
		"Get grabbed pages returning %x, order %d. (Slow path).\n",
		result, order);
	check_shift_keys(0, NULL);
	return result;
}

#define MAX_ATTEMPTS 3
extern int freeze_processes(int no_progress);

static int attempt_to_freeze(void)
{
	int lastfreezeresult, freezeattemptsleft = MAX_ATTEMPTS;

	/* Stop processes before checking again */
	do {
		thaw_processes();
		prepare_status(1, 1, "Freezing processes: Attempt %d", MAX_ATTEMPTS + 1 - freezeattemptsleft);
		lastfreezeresult = freeze_processes(0);
		printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 0, "- Freeze_processes returned %d.\n",
			lastfreezeresult);
		freezeattemptsleft--;
		//printk(KERN_ERR "Last freeze result was %d.\n", lastfreezeresult);
	} while ((freezeattemptsleft && lastfreezeresult) &&
		 (!TEST_RESULT_STATE(SUSPEND_ABORTED)));


	//printk("Exiting attempt to freeze with result %d.\n", lastfreezeresult);
	if (lastfreezeresult) {
		SET_RESULT_STATE(SUSPEND_ABORTED);
		SET_RESULT_STATE(SUSPEND_FREEZING_FAILED);
	} else
		arefrozen = 1;

	return lastfreezeresult;
}

static int eat_memory(void)
{
	int orig_memory_still_to_eat, last_amount_needed = 0, times_criteria_met = 0;
	
	/*
	 * Note that if we have enough storage space and enough free memory, we may
	 * exit without eating anything. We give up when the last 10 iterations ate
	 * no extra pages because we're not going to get much more anyway, but
	 * the few pages we get will take a lot of time.
	 *
	 * We freeze processes before beginning, and then unfreeze them if we
	 * need to eat memory until we think we have enough. If our attempts
	 * to freeze fail, we give up and abort.
	 */

	/* ----------- Stage 1: Freeze Processes ------------- */

	
	prepare_status(0, 1, "Eating memory.");

	recalculate_stats();
	display_stats();

	orig_memory_still_to_eat = amount_needed();
	last_amount_needed = orig_memory_still_to_eat;

	if ((image_size_limit == -1) && (orig_memory_still_to_eat)) {
		SET_RESULT_STATE(SUSPEND_ABORTED);
		SET_RESULT_STATE(SUSPEND_WOULD_EAT_MEMORY);
	}
		
	/* ----------- Stage 2: Eat memory ------------- */

	while ((!EATEN_ENOUGH_MEMORY()) && (!TEST_RESULT_STATE(SUSPEND_ABORTED)) && (times_criteria_met < 10)) {
		if (orig_memory_still_to_eat)
			update_status(orig_memory_still_to_eat - amount_needed(), orig_memory_still_to_eat, " Image size %d ", MB(STORAGE_NEEDED));
		
		if ((last_amount_needed - amount_needed()) < 10)
			times_criteria_met++;
		else
			times_criteria_met = 0;
		last_amount_needed = amount_needed();
		try_to_free_pages_swsusp(amount_needed());
		grab_free_memory();
		recalculate_stats();
		display_stats();

		check_shift_keys(0, NULL);
	}

	grab_free_memory();

	printlog(SUSPEND_EAT_MEMORY, SUSPEND_VERBOSE, "\n");
	
	printlog(SUSPEND_EAT_MEMORY, SUSPEND_VERBOSE,
		"(Freezer exit:) Swap needed calculated as (%d+%d)*%d/100+%d+1+%d=%d.\n",
		pageset1_size,
		pageset2_size,
		expected_compression_ratio,
		num_range_pages,
	 	header_storage_for_plugins(),
		STORAGE_NEEDED);

	/* Blank out image size display */
	update_status(100, 100, "                   ");
	if ((amount_needed() > 0)) {
		printk("Unable to free sufficient memory to suspend. Still need %d pages.\n",
			amount_needed());
		SET_RESULT_STATE(SUSPEND_ABORTED);
		SET_RESULT_STATE(SUSPEND_UNABLE_TO_FREE_ENOUGH_MEMORY);
	}
	
	check_shift_keys(1, "Memory eating completed. ");
	return 0;
}

int prepare_image(void)
{
	int result = 1, sizesought;

	arefrozen = 0;

	sizesought = 100 + memory_for_plugins();

	if (fill_swsusp_memory_pool(sizesought))
		return 1;

	if (attempt_to_freeze())
		return 1;

	if (!active_writer->ops.writer.storage_available()) {
		printk(KERN_ERR "You need some storage available to be able to suspend.\n");
		SET_RESULT_STATE(SUSPEND_ABORTED);
		SET_RESULT_STATE(SUSPEND_NOSTORAGE_AVAILABLE);
		return 1;
	}

	do {
		if (eat_memory() || TEST_RESULT_STATE(SUSPEND_ABORTED))
			break;

		/* Top up */
		if (fill_swsusp_memory_pool(sizesought))
			continue;
	
		do_suspend_sync();

		result = update_image();

	} while ((result) && (!TEST_RESULT_STATE(SUSPEND_ABORTED)));

	PRINTFREEMEM("after preparing image");

	/* Release memory that has been eaten */
	free_grabbed_memory();
	
	if (!TEST_RESULT_STATE(SUSPEND_ABORTED))
		swsusp_state |= USE_MEMORY_POOL;

	return result;
}
