/*
 * /drivers/swsusp/proc.c
 *
 * Copyright (C) 2002-2003 Nigel Cunningham <ncunningham@clear.net.nz>
 *
 * This file is released under the GPLv2.
 *
 * This file is to realize proc entries for tuning Software Suspend.
 *
 * Versions:
 * 1: /proc/sys/kernel/swsusp the only tuning interface
 * 2: Initial version of this file
 * 3: Removed KDB parameter.
 *    Added checkpage parameter (for checking checksum of a page over time).
 * 4: Added entry for maximum granularity in splash screen progress bar.
 *    (Progress bar is slow, but the right setting will vary with disk &
 *    processor speed and the user's tastes).
 * 5: Added enable_escape to control ability to cancel aborting by pressing
 *    ESC key.
 * 6: Removed checksumming and checkpage parameter. Made all debugging proc
 *    entries dependant upon debugging being compiled in.
 *    Meaning of some flags also changed in this version.
 * 7: Added header_locations entry to simplify getting the resume= parameter for
 *    swapfiles easy and swapfile entry for automatically doing swapon/off from
 *    swapfiles as well as partitions.
 * 8: Added option for marking process pages as pageset 2 (processes_pageset2).
 * 9: Added option for keep image mode.
 *    Enumeration patch from Michael Frank applied.
 * 10: Various corrections to when options are disabled/enabled;
 *     Added option for specifying expected compression.
 * 11: Added option for freezer testing. Debug only.
 * 12: Removed test entries no_async_[read|write], processes_pageset2 and
 *     NoPageset2.
 * 13: Make default_console_level available when debugging disabled, but limited
 *     to 0 or 1.
 * 14: Rewrite to allow for dynamic registration of proc entries and smooth the
 *     transition to kobjects in 2.6.
 * 15: Add setting resume2 parameter without rebooting (still need to run lilo
 *     though!).
 */

#define SWSUSP_PROC_C

static int swsusp_proc_version = 15;
static int proc_initialised = 0;

#include <linux/suspend-common.h>
#include <linux/suspend-debug.h>
#include <linux/proc_fs.h>

static struct list_head swsusp_proc_entries;
static struct proc_dir_entry *swsusp_dir, *compat_entry, *compat_parent;

extern void swsusp_console_proc_init(void);

/*
 * swsusp_write_compat_proc.
 *
 * This entry allows all of the settings to be set at once. 
 * It was originally for compatibility with pre- /proc/swsusp
 * versions, but has been retained because it makes saving and
 * restoring the configuration simpler.
 */
static int swsusp_write_compat_proc(struct file *file, const char * buffer,
		unsigned long count, void * data)
{
	char * buf = (char *) get_zeroed_page(GFP_ATOMIC), *lastbuf;
	int i;
	unsigned long nextval;

	if (!buf)
		return -ENOMEM;

	if (count > PAGE_SIZE)
		count = PAGE_SIZE;

	if (copy_from_user(buf, buffer, count))
		return -EFAULT;
	for (i = 0; i < 6; i++) {
		if (!buf)
			break;
		lastbuf = buf;
		nextval = simple_strtoul(buf, &buf, 0);
		if (buf == lastbuf)
			break;
		switch (i) {
			case 0:
				swsusp_result = nextval;
				break;
			case 1:
				swsusp_action = nextval;
				break;
			case 2:
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
				swsusp_debug_state = nextval;
#endif
				break;
			case 3:
				swsusp_default_console_level = nextval;
#ifndef CONFIG_SOFTWARE_SUSPEND_DEBUG
				if (swsusp_default_console_level > 1)
					swsusp_default_console_level = 1;
#endif
				break;
			case 4:
				image_size_limit = nextval;
				break;
			case 5:
				max_async_ios = nextval;
				break;

		}
		buf++;
		while (*buf == ' ')
			buf++;
	}
	buf[count] = 0;
	free_pages((unsigned long) buf, 0);
	return count;
}

/*
 * swsusp_read_compat_proc.
 *
 * Like it's _write_ sibling, this entry allows all of the settings
 * to be read at once.
 * It too was originally for compatibility with pre- /proc/swsusp
 * versions, but has been retained because it makes saving and
 * restoring the configuration simpler.
 */
static int swsusp_read_compat_proc(char * page, char ** start, off_t off, int count,
		int *eof, void *data)
{
	int len = 0;

	len = sprintf(page, "%ld %ld %ld %d %d %d\n",
			swsusp_result,
			swsusp_action,
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
			swsusp_debug_state,
#else
	  	        0,
#endif
			swsusp_default_console_level,
			image_size_limit,
			max_async_ios);
	*eof = 1;
	return len;
}

/*
 * proc_software_suspend_pending
 *
 * This routine initiates a suspend cycle when /proc/swsusp/activity is
 * written to. The value written is ignored.
 */

static int proc_software_suspend_pending(struct file *file, const char *buffer,
        unsigned long count, void *data)
{
	software_suspend_pending();
	return count;
}

extern char resume_file[256];	/* For resume= kernel option */

static int resume2_read_proc(char * page, char ** start, off_t off, int count,
		int *eof, void *data)
{
	*eof = 1;
	return (sprintf(page, "%s\n", resume_file));
}

extern int attempt_to_parse_resume_device(int boot_time);

static int resume2_write_proc(struct file *file, const char * buffer,
		unsigned long count, void * data)
{
	if (count > 255)
		count = 255;

	if (copy_from_user(resume_file, buffer, count))
		return -EFAULT;
	
	if ((count) && (resume_file[count - 1] == '\n'))
		resume_file[count - 1] = 0;

	resume_file[count] = 0;

	if (count)
		attempt_to_parse_resume_device(0);
	return count;
}
/*
 * debuginfo_read_proc
 * Functionality   : Displays information that may be helpful in debugging
 *                   software suspend.
 */
extern void start_kswsuspd(void * data);

static int debuginfo_read_proc(char * page, char ** start, off_t off, int count,
		int *eof, void *data)
{
	static struct debug_info_data kswsuspd_data;

	kswsuspd_data.action = 2;
	kswsuspd_data.buffer = page;
	kswsuspd_data.buffer_size = count;
	start_kswsuspd(&kswsuspd_data);
	*eof = 1;
	return kswsuspd_data.bytes_used;
}

/*
 * version_read_proc
 *
 * Displays the version of software suspend installed in the
 * currently running kernel.
 */
static int version_read_proc(char * page, char ** start, off_t off, int count,
		int *eof, void *data)
{
	*eof = 1;
	return sprintf(page, "%s\n", swsusp_version);
}

/*
 * generic_read_proc
 *
 * Generic handling for reading the contents of bits, integers
 * and unsigned longs.
 */
static int generic_read_proc(char * page, char ** start, off_t off, int count,
		int *eof, void *data)
{
	int len = 0;
	struct swsusp_proc_data * proc_data = (struct swsusp_proc_data *) data;

	switch (proc_data->type) {
		case SWSUSP_PROC_DATA_CUSTOM:
			printk("Error! /proc/swsusp/%s marked as having custom routines, but the generic read routine has been invoked.\n",
				proc_data->filename);
			break;
		case SWSUSP_PROC_DATA_BIT:
			len = sprintf(page, "%d\n", -test_bit(proc_data->data.bit.bit, proc_data->data.bit.bit_vector));
			break;
		case SWSUSP_PROC_DATA_INTEGER:
			{
				int * variable = proc_data->data.integer.variable;
				len = sprintf(page, "%d\n", *((int *) variable));
				break;
			}
		case SWSUSP_PROC_DATA_UL:
			{
				long * variable = proc_data->data.ul.variable;
				len = sprintf(page, "%lu\n", *((long *) variable));
				break;
			}
	}
	*eof = 1;
	return len;
}

/*
 * generic_write_proc
 *
 * Generic routine for handling writing to files representing
 * bits, integers and unsigned longs.
 */

static int generic_write_proc(struct file *file, const char * buffer,
		unsigned long count, void * data)
{
	struct swsusp_proc_data * proc_data = (struct swsusp_proc_data *) data;
	char * buf = (char *) get_zeroed_page(GFP_ATOMIC);

	if (!buf)
		return -ENOMEM;

	if (count > PAGE_SIZE)
		count = PAGE_SIZE;

	if (copy_from_user(buf, buffer, count))
		return -EFAULT;
	buf[count] = 0;

	switch (proc_data->type) {
		case SWSUSP_PROC_DATA_CUSTOM:
			printk("Error! /proc/swsusp/%s marked as having custom routines, but the generic write routine has been invoked.\n",
				proc_data->filename);
			break;
		case SWSUSP_PROC_DATA_BIT:
			{
			int value = simple_strtoul(buf, NULL, 0);
			if (value)
				set_bit(proc_data->data.bit.bit, (proc_data->data.bit.bit_vector));
			else
				clear_bit(proc_data->data.bit.bit, (proc_data->data.bit.bit_vector));
			}
			break;
		case SWSUSP_PROC_DATA_INTEGER:
			{
				int * variable = proc_data->data.integer.variable;
				int minimum = proc_data->data.integer.minimum;
				int maximum = proc_data->data.integer.maximum;
				*variable = simple_strtoul(buf, NULL, 0);
				/* Set minimum == -1 for no minimum */
				if (((*variable) < minimum) && (minimum != -1))
					*variable = minimum;

				/* Likewise for no maximum */
				if (((*variable) > maximum) && (maximum != -1))
					*variable = maximum;
				break;
			}
		case SWSUSP_PROC_DATA_UL:
			{
				unsigned long * variable = proc_data->data.ul.variable;
				unsigned long minimum = proc_data->data.ul.minimum;
				unsigned long maximum = proc_data->data.ul.maximum;
				*variable = simple_strtoul(buf, NULL, 0);
				/* Set minimum == -1 (well yes, 0xFFFFFFFF) for no minimum */
				if ((minimum != (unsigned long) -1) && ((*variable) < minimum))
					*variable = minimum;

				/* Likewise for no maximum */
				if ((minimum != (unsigned long) -1) && ((*variable) > maximum))
					*variable = maximum;
				break;
			}
			break;
	}
	free_pages((unsigned long) buf, 0);
	return count;
}

/*
 * Non-plugin proc entries.
 *
 * This array contains entries that are automatically registered at
 * boot. Plugins and the console code register their own entries separately.
 */

static struct swsusp_proc_data proc_params[] = {
	{ .filename			= "activate",
	  .permissions			= PROC_WRITEONLY,
	  .type				= SWSUSP_PROC_DATA_CUSTOM,
	  .data = {
		  .special = {
			  .write_proc	= proc_software_suspend_pending
		  }
	  }
	},

	{ .filename			= "all_settings",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_CUSTOM,
	  .data = {
		  .special = {
			  .read_proc	= swsusp_read_compat_proc,
			  .write_proc	= swsusp_write_compat_proc,
		  }
	  }
	},

	{ .filename			= "async_io_limit", 
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_INTEGER,
	  .data = {
		  .integer = {
			  .variable	= &max_async_ios,
			  .minimum	= 1,
			  .maximum	= 5000,
		  }
	  }
	},
		
	{ .filename			= "debug_info",
	  .permissions			= PROC_READONLY,
	  .type				= SWSUSP_PROC_DATA_CUSTOM,
	  .data = {
		  .special = {
			  .read_proc	= debuginfo_read_proc,
		  }
	  }
	},
	
	{ .filename			= "image_size_limit",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_INTEGER,
	  .data = {
		  .integer = {
			  .variable	= &image_size_limit,
			  .minimum	= 0,
			  .maximum	= -1,
		  }
	  }
	},

	{ .filename			= "interface_version",
	  .permissions			= PROC_READONLY,
	  .type				= SWSUSP_PROC_DATA_INTEGER,
	  .data = {
		  .integer = {
			  .variable	= &swsusp_proc_version,
		  }
	  }
	},
	
	{ .filename			= "last_result",
	  .permissions			= PROC_READONLY,
	  .type				= SWSUSP_PROC_DATA_UL,
	  .data = {
		  .ul = {
			  .variable	=  &swsusp_result,
		  }
	  }
	},

	{ .filename			= "version",
	  .permissions			= PROC_READONLY,
	  .type				= SWSUSP_PROC_DATA_CUSTOM,
	  .data = {
		  .special = {
			  .read_proc	= version_read_proc,
		  }
	  }
	},

#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
	{ .filename			= "freezer_test",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_BIT,
	  .data = {
		  .bit = {
			  .bit_vector	= &swsusp_action,
			  .bit		= SUSPEND_FREEZER_TEST,
		  }
	  }
	},

	{ .filename			= "reboot",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_BIT,
	  .data = {
		  .bit = {
			  .bit_vector	= &swsusp_action,
			  .bit		= SUSPEND_REBOOT,
		  }
	  }
	},
	  
	{ .filename			= "resume2",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_CUSTOM,
	  .data = {
		  .special = {
			  .read_proc	= resume2_read_proc,
			  .write_proc	= resume2_write_proc,
		  }
	  }
	},

	{ .filename			= "slow",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_BIT,
	  .data = {
		  .bit = {
			  .bit_vector	= &swsusp_action,
			  .bit		= SUSPEND_SLOW,
		  }
	  }
	},
#endif
	  
#ifdef CONFIG_SOFTWARE_SUSPEND_KEEP_IMAGE
	{ .filename			= "keep_image",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_BIT,
	  .data = {
		  .bit = {
			  .bit_vector	= &swsusp_action,
			  .bit		= SUSPEND_KEEP_IMAGE,
		  }
	  }
	},
#endif
};

/*
 * swsusp_initialise_proc
 *
 * Initialise the /proc/swsusp tree.
 *
 */

int swsusp_initialise_proc(void)
{
	swsusp_dir = proc_mkdir("swsusp", NULL);
	
	if (swsusp_dir == NULL) {
		printk(KERN_ERR "Failed to create /proc/swsusp.\n");
		return 1;
	}

	INIT_LIST_HEAD(&swsusp_proc_entries);

	proc_initialised = 1;

	return 0;
}

/*
 * swsusp_register_procfile
 *
 * Helper for registering a new /proc/swsusp entry.
 */

struct proc_dir_entry * swsusp_register_procfile(struct swsusp_proc_data * swsusp_proc_data)
{
	struct proc_dir_entry * new_entry;
	
	if ((!proc_initialised) && (swsusp_initialise_proc()))
		return NULL;

	new_entry = create_proc_entry(
			swsusp_proc_data->filename,
			swsusp_proc_data->permissions, 
			swsusp_dir);
	if (new_entry) {
		list_add_tail(&swsusp_proc_data->proc_data_list, &swsusp_proc_entries);
		if (swsusp_proc_data->type) {
			new_entry->read_proc = generic_read_proc;
			new_entry->write_proc = generic_write_proc;
		} else {
			new_entry->read_proc = swsusp_proc_data->data.special.read_proc;
			new_entry->write_proc = swsusp_proc_data->data.special.write_proc;
		}
		new_entry->data = swsusp_proc_data;
	}
	return new_entry;
}

/*
 * swsusp_unregister_procfile
 *
 * Helper for removing unwanted /proc/swsusp entries.
 *
 */
void swsusp_unregister_procfile(struct swsusp_proc_data * swsusp_proc_data)
{
	remove_proc_entry(
		swsusp_proc_data->filename,
		swsusp_dir);
	list_del(&swsusp_proc_data->proc_data_list);
}

/* 
 * find_proc_dir_entry.
 *
 * Based on remove_proc_entry.
 * This will go shortly, once user space utilities
 * are updated to look at /proc/swsusp/all_settings.
 */

struct proc_dir_entry * find_proc_dir_entry(const char *name, struct proc_dir_entry *parent)
{
	struct proc_dir_entry **p;
	int len;

	len = strlen(name);
	for (p = &parent->subdir; *p; p=&(*p)->next ) {
		if (proc_match(len, name, *p)) {
			return *p;
		}
	}
	return NULL;
}

/*
 * swsusp_init_proc
 * 
 * Initialise proc file support at boot time. This is called
 * by swsusp2.c::software_resume2.
 */
int swsusp_init_proc(void)
{
	int i;
	int numfiles = sizeof(proc_params) / sizeof(struct swsusp_proc_data);
	
	if ((!proc_initialised) && swsusp_initialise_proc())
		return 1;

	for (i=0; i< numfiles; i++)
		swsusp_register_procfile(&proc_params[i]);

	swsusp_console_proc_init();

	compat_parent = find_proc_dir_entry("sys", &proc_root);
	if (compat_parent) {
		compat_parent = find_proc_dir_entry("kernel", compat_parent);
		if (compat_parent) {
			compat_entry = create_proc_entry("swsusp",
				0600, 
				compat_parent);
			compat_entry->read_proc = swsusp_read_compat_proc;
			compat_entry->write_proc = swsusp_write_compat_proc;
		}
	}

	return 0;
}
