/*
 * kernel/power/freeze_and_free.c
 *
 * Copyright (C) 1998-2001 Gabor Kuti <seasons@fornax.hu>
 * Copyright (C) 1998,2001,2002 Pavel Machek <pavel@suse.cz>
 * Copyright (C) 2002-2003 Florent Chabaud <fchabaud@free.fr>
 * Copyright (C) 2002-2003 Nigel Cunningham <ncunningham@clear.net.nz>
 *
 * This file is released under the GPLv2.
 *
 * Freeze_and_free contains the routines software suspend uses to freeze other
 * processes during the suspend cycle and to (if necessary) free up memory in
 * accordance with limitations on the image size.
 *
 * Ideally, the image saved to disk would be an atomic copy of the entire 
 * contents of all RAM and related hardware state. One of the first 
 * prerequisites for getting our approximation of this is stopping the activity
 * of other processes. We can't stop all other processes, however, since some 
 * are needed in doing the I/O to save the image. Freeze_and_free.c contains 
 * the routines that control suspension and resuming of these processes.
 * 
 * Under high I/O load, we need to be careful about the order in which we
 * freeze processes. If we freeze processes in the wrong order, we could
 * deadlock others. The freeze_order array this specifies the order in which
 * critical processes are frozen. All others are suspended after these have
 * entered the refrigerator.
 *
 * Another complicating factor is that freeing memory requires the processes
 * to not be frozen, but at the end of freeing memory, they need to be frozen
 * so that we can be sure we actually have eaten enough memory. This is why
 * freezing and freeing are in the one file. The freezer is not called from
 * the main logic, but indirectly, via the code for eating memory. The eat
 * memory logic is iterative, first freezing processes and checking the stats,
 * then (if necessary) unfreezing them and eating more memory until it looks 
 * like the criteria are met (at which point processes are frozen & stats
 * checked again).
 */

#define SWSUSP_FREEZER_C

#include <linux/suspend-common.h>
#include <linux/notifier.h>

char idletimeout;
atomic_t swsusp_num_active = { 0 };
atomic_t __nosavedata swsusp_cpu_counter = { 0 }; 

#if 0
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
extern void ide_suspend(void);
extern void ide_unsuspend(void);
#endif
#endif

unsigned char software_suspend_state = SOFTWARE_SUSPEND_DISABLED;
unsigned int suspend_task = 0;
unsigned char __nosavedata swsusp_state = 0;

unsigned long swsusp_action = 0;
unsigned long swsusp_result = 0;

/* Locks */
spinlock_t suspend_irq_lock = SPIN_LOCK_UNLOCKED;
unsigned long swsuspirqflags;

int now_resuming = 0;

int swsusp_default_console_level = 0;

/* ------------------------------------------------------------------------ */


#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,99)
extern int wakeup_bdflush(long nr_pages);
void do_suspend_sync(void)
{
	wakeup_bdflush(0);
	blk_run_queues();
	io_schedule();
	yield();
}

#else
extern int nr_buffers_type[NR_LIST];
void do_suspend_sync(void)
{
	int i;
	int orig[NR_LIST];

	for (i = 0; i < NR_LIST; i++)
		orig[i] = nr_buffers_type[i];
	
	do { } while (sync_buffers(NODEV, 1));
	do { } while (fsync_dev(NODEV));

	while (1) {
		run_task_queue(&tq_disk);
		if (!TQ_ACTIVE(tq_disk))
			break;
		printk(KERN_ERR "Hm, tq_disk is not empty after run_task_queue\n");
	}
	
	if (nr_buffers_type[1] || nr_buffers_type[2]) {
		printk("Entry to do_suspend_sync:\n");
		for (i = 0; i < NR_LIST; i++)
			printk(" nr_buffers_type[i] = %d.\n", orig[i]);

		printk("Exit from do_suspend_sync:\n");
		for (i = 0; i < NR_LIST; i++)
			printk(" nr_buffers_type[i] = %d.\n", nr_buffers_type[i]);
		check_shift_keys(1, "Not all buffers clean when exiting do_suspend_sync.");
	}
}
#endif

#ifdef CONFIG_SMP
static void smp_pause(void * data)
{
	atomic_inc(&swsusp_cpu_counter);
	//printk("Processor %d paused. CPU counter ->%d\n", 
	//	smp_processor_id(), atomic_read(&swsusp_cpu_counter));
	while(swsusp_state & FREEZE_SMP) {
		cpu_relax();
		barrier();
		check_shift_keys(0, "");
	}
	atomic_dec(&swsusp_cpu_counter);
	//printk("Processor %d unpaused. CPU counter ->%d\n", 
	//	smp_processor_id(), atomic_read(&swsusp_cpu_counter));
	//printk("Processor %d unpaused.\n", smp_processor_id());
}

#endif

 /* Processes which need to be frozen after others in this particular order */
#if 0
static char * freeze_order[] = {"kjournald", "bdflush", "kupdated", "kswapd", "[OTHERS]" };
static const int freeze_order_num = 4;
static const int freeze_rest_index = 4;
#else
static char * freeze_order[] = {"[OTHERS]" };
static const int freeze_order_num = 0;
static const int freeze_rest_index = 0;
#endif
static int to_be_frozen(struct task_struct * p) {

	if ((p == current) ||
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	    (!(strcmp(p->comm, "dosexec"))) ||
#endif
	    (p->flags & PF_NOFREEZE) ||
	    (p->state == TASK_ZOMBIE) ||
	    (p->state == TASK_STOPPED))
		return 0;
	return 1;
}

static int processesfrozen = 0;
static int freezing = 0;

/**
 * refrigerator - idle routine for frozen processes
 * @flag: unsigned long, non zero if signals to be flushed.
 *
 * A routine for kernel threads which should not do work during suspend
 * to enter and spin in until the process is finished.
 *
 */

void refrigerator(unsigned long flag)
{
	unsigned long flags;
	long save;
	
	if (unlikely(current->flags & PF_NOFREEZE)) {
		swsusp_spin_lock_irqsave(PROCESS_SIG_MASK(current), flags);
		RECALC_SIGPENDING;
		swsusp_spin_unlock_irqrestore(PROCESS_SIG_MASK(current), flags);
		return;
	}

	/* You need correct to work with real-time processes.
	   OTOH, this way one process may see (via /proc/) some other
	   process in stopped state (and thereby discovered we were
	   suspended. We probably do not care). 
	 */
	processesfrozen++;
	if ((flag) && (current->flags & PF_FREEZE)) {

		printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 0, "%s (%d) refrigerated and sigpending recalculated.\n",
			current->comm, current->pid);
		swsusp_spin_lock_irqsave(PROCESS_SIG_MASK(current), flags);
		RECALC_SIGPENDING;
		swsusp_spin_unlock_irqrestore(PROCESS_SIG_MASK(current), flags);
	} else
		printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 0, "%s (%d) refrigerated.\n",
			current->comm, current->pid);

	if (swsusp_state) {
		save = current->state;
		current->flags |= PF_FROZEN;
		while (current->flags & PF_FROZEN) {
			current->state = TASK_STOPPED;
			schedule();
		}
		current->state = save;
	}
	current->flags &= ~PF_FREEZE;
	processesfrozen--;
}

/*
 * swsusp_activity_start
 */
void __swsusp_activity_start(int flags, const char * function, const char * file)
{
	if (unlikely((suspend_task) && (current->pid == suspend_task)))
		return;

	if (unlikely(current->flags & PF_FREEZE))
		refrigerator(flags);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	if (unlikely((swsusp_state & FREEZE_NEW_ACTIVITY) &&
		     strcmp(current->comm, "dosexec") &&
		     ((!(current->flags & PF_SYNCTHREAD)) ||
		      (swsusp_state & FREEZE_UNREFRIGERATED))))
#else
	if (unlikely((swsusp_state & FREEZE_NEW_ACTIVITY) &&
		     ((!(current->flags & PF_SYNCTHREAD)) ||
		      (swsusp_state & FREEZE_UNREFRIGERATED))))
#endif
		refrigerator(flags);
	
	if (unlikely(current->flags & PF_FRIDGE_WAIT)) {
		printk("__swsusp_activity_start: Task %s already has FRIDGE_WAIT set.\n",
			current->comm);
		return;
	}

	current->flags |= PF_FRIDGE_WAIT;
	atomic_inc(&swsusp_num_active);
}

/*
 * freeze_processes - Freeze processes prior to saving an image of memory to disk
 * 
 * Return value: 0 = success, else # of processes that we failed to stop 
 */
extern int sync_old_buffers(void);
extern void show_task(struct task_struct * p);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
extern spinlock_t io_request_lock;
#endif

/* Freeze_processes.
 * If the flag no_progress is non-zero, progress bars not be updated.
 * Debugging output is still printed.
 */
int freeze_processes(int no_progress)
{
	int todo = 0, threadcount, newtodo;
	int lastprocessesfrozen = processesfrozen;
	struct task_struct FOR_EACH_THREAD_TASK_STRUCTS;
	int index = 0, origfrozen = processesfrozen;
	int orignum_active, lastnum_active = 0;
	unsigned long start_time, last_time;
	int showidlelist;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	unsigned long iorequestflags = 0;
#endif

	//swap_list_lock();

	freezing = 1;
	
	showidlelist = 1;

	swsusp_state = FREEZE_NEW_ACTIVITY;
	swsusp_result = 0;	/* Might be called from pm_disk or swsusp - ensure reset */

	read_lock(&tasklist_lock);
	FOR_EACH_THREAD_START {
		if (p->flags & PF_SYNCTHREAD)
			printnolog(SUSPEND_FREEZER, SUSPEND_MEDIUM, 0,
				"%s (%d) is a syncthread at entrance to fridge\n",
				p->comm, p->pid);
		if (!(p->flags & PF_FRIDGE_WAIT))
			continue;
		printnolog(SUSPEND_FREEZER, SUSPEND_MEDIUM, 0,
			"%s (%d) busy at entrance to fridge.\n",
			p->comm, p->pid);
		show_task(p);
	} FOR_EACH_THREAD_END
	read_unlock(&tasklist_lock);
	printnolog(SUSPEND_FREEZER, SUSPEND_MEDIUM, 0, "\n");
	
	/* First, wait for syncing to finish */
	if (!no_progress)
		prepare_status(1, 1, "Freezing processes: Waiting for activity to finish.");
	lastnum_active = orignum_active = atomic_read(&swsusp_num_active);
	if (!no_progress)
		update_status(0, lastnum_active, " %d remaining.", lastnum_active);
	
	last_time = start_time = jiffies;
	while (atomic_read(&swsusp_num_active)) {
		if (TEST_RESULT_STATE(SUSPEND_ABORTED)) {
			read_lock(&tasklist_lock);
			FOR_EACH_THREAD_START {
				if (p->flags & PF_SYNCTHREAD)
					printk("%s (%d) is still a syncthread when aborting.\n",
						p->comm, p->pid);
				if (!(p->flags & PF_FRIDGE_WAIT))
					continue;
				printk("%s (%d) still busy when aborting.\n",
						p->comm,
						p->pid);

				if (!TEST_ACTION_STATE(SUSPEND_FREEZER_TEST_SHOWALL))
					show_task(p);
				todo++;
			} FOR_EACH_THREAD_END
			read_unlock(&tasklist_lock);
			if (TEST_ACTION_STATE(SUSPEND_FREEZER_TEST_SHOWALL))
				show_state();
			freezing = 0;
			idletimeout=0;
			//swap_list_unlock();
			return todo;
		}
		set_task_state(current, TASK_UNINTERRUPTIBLE);
		schedule_timeout(HZ/5);
		if (atomic_read(&swsusp_num_active) < lastnum_active) {
			lastnum_active = atomic_read(&swsusp_num_active);
			if (!no_progress)
				update_status(orignum_active - lastnum_active, 
				orignum_active, " %d remaining.", 
				lastnum_active);
			last_time = jiffies;
			showidlelist=1;
		} else if (jiffies > last_time + 250) {
			idletimeout = 1;
			printk("swsusp_num_active is %d, Idle for 250 jiffies\n",
				atomic_read(&swsusp_num_active));
			if (showidlelist) {
				showidlelist = 0;
				show_state();
			}
			last_time = jiffies;
		}
	}

	idletimeout=0;
	check_shift_keys(1, "I/O activity stopped.");

	if (TEST_RESULT_STATE(SUSPEND_ABORTED)) {
		//swap_list_unlock();
		return 1;
	}

	/* Now do our own sync, just in case one wasn't running already */
	if (!no_progress)
		prepare_status(1, 1, "Freezing processes: Syncing remaining I/O.");
	current->flags |= PF_SYNCTHREAD;
	do_suspend_sync();
	current->flags &= ~PF_SYNCTHREAD;

	check_shift_keys(1, "Syncing finished.");

	if (TEST_RESULT_STATE(SUSPEND_ABORTED)) {
		//swap_list_unlock();
		return 1;
	}

	suspend_task = current->pid;
	swsusp_state |= FREEZE_UNREFRIGERATED;

	/* Now freeze remaining processes */
	if (!no_progress)
		prepare_status(1, 1, "Freezing processes: Freezing remaining tasks.");
	read_lock(&tasklist_lock);
	FOR_EACH_THREAD_START {
		if (!to_be_frozen(p))
			continue;
		todo++;
	} FOR_EACH_THREAD_END
	read_unlock(&tasklist_lock);

	start_time = jiffies;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	swsusp_spin_lock_irqsave(&io_request_lock, iorequestflags);
#endif

	printnolog(SUSPEND_FREEZER, SUSPEND_MEDIUM, 0, " ** %s **\n",
		freeze_order[index]);

	do {
		int numsignalled = 0;
		newtodo = 0;
		threadcount = 0;

		printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 1, "SMP Freeze.       ");

		/* 
		 * Pause the other processors so we can safely
		 * change threads' flags
		 */
		swsusp_state |= FREEZE_SMP;
		smp_call_function(smp_pause, NULL, 0, 0);

		printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 1, "Wait on SMP Freeze.");
		while ((atomic_read(&swsusp_cpu_counter) < (NUM_CPUS - 1)) &&
			(!TEST_RESULT_STATE(SUSPEND_ABORTED))) {
			cpu_relax();
			barrier();
			check_shift_keys(0, "");
		}
		if (TEST_RESULT_STATE(SUSPEND_ABORTED)) 
			goto aborting;

		printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 1, "Signalling.        ");
		read_lock(&tasklist_lock);
		FOR_EACH_THREAD_START {
			unsigned long flags;
			int i;
			int thread_index = -1;
			if (!to_be_frozen(p))
				continue;
			for (i = 0; i < freeze_order_num; i++)
				if (!strcmp(p->comm, freeze_order[i])) {
					thread_index = i;
					break;
				}
			if (thread_index > -1) {
				if (thread_index != index)
					continue;
			} else
				if (freeze_rest_index != index)
					continue;
			threadcount++;
			/* 
			 * We don't signal sync; new instances automatically
			 * enter freezer (on swsusp_state flag) and we wait
			 * for old ones to finish and exit.
			 */
			if (!(p->flags & PF_FREEZE)) {
				numsignalled++;
				printnolog(SUSPEND_FREEZER, SUSPEND_MEDIUM, 0, 
						"%s%s thread %d",
						numsignalled > 1 ? "," : "",
						index == freeze_order_num ?
							p->comm : "",
						p->pid);
				p->flags |= PF_FREEZE;
				/* FIXME: smp problem here: we may not access other process' flags
				   without locking */
				swsusp_spin_lock_irqsave(PROCESS_SIG_MASK(p), flags);
				WAKE_UP(p);
				swsusp_spin_unlock_irqrestore(PROCESS_SIG_MASK(p), flags);
			}
		} FOR_EACH_THREAD_END
		read_unlock(&tasklist_lock);
		printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 1, "Unfreeze SMP.      ");
		swsusp_state &= ~FREEZE_SMP;
		while ((atomic_read(&swsusp_cpu_counter)) &&
			(!TEST_RESULT_STATE(SUSPEND_ABORTED))) {
			cpu_relax();
			barrier();
			check_shift_keys(0, "");
		}
		if (numsignalled)
			printnolog(SUSPEND_FREEZER, SUSPEND_MEDIUM, 0, ".\n");
		if (!threadcount && index < freeze_order_num) {
			index++;
			printnolog(SUSPEND_FREEZER, SUSPEND_MEDIUM, 0, "\n ** %s **\n",
				freeze_order[index]);
		}
		printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 1, "Schedule.          ");
		set_task_state(current, TASK_INTERRUPTIBLE);
		schedule_timeout(HZ/10);
		printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 1, "Examine effect.    ");
		read_lock(&tasklist_lock);
		FOR_EACH_THREAD_START {
			if (!to_be_frozen(p))
				continue;
			/*
			 * We signal uninterruptible processes, but don't have a hernia
			 * if they don't enter the refrigerator. The signal ensures they
			 * do refrigerate if they wake up. If they don't, they will
			 * harmlessly run through the freezer after suspend.
			 *
			 * Note that these uninterruptible processes are counted in the
			 * number to be frozen (for the sake of the progress bar), but
			 * won't be shown as processes that refused to be frozen if
			 * freezing fails.
			 */
			if (p->state == TASK_UNINTERRUPTIBLE)
				continue;
			/* Count all processes yet to be frozen */
			newtodo++;
		} FOR_EACH_THREAD_END
		read_unlock(&tasklist_lock);
		printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 1, "Show results.      ");
		if (lastprocessesfrozen < processesfrozen) {
			printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 0, "Waiting on:");
			read_lock(&tasklist_lock);
			FOR_EACH_THREAD_START {
				if (!to_be_frozen(p))
					continue;
				if (p->state == TASK_UNINTERRUPTIBLE)
					continue;
				/* Only print a process if we're waiting for it now */
				if (p->flags & PF_FREEZE)
					printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 0, "%s (%d); ", 
						p->comm, p->pid);
			} FOR_EACH_THREAD_END
			read_unlock(&tasklist_lock);
			printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 0, "\n");
		}
		lastprocessesfrozen = processesfrozen;
		printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 1, "Check time.        ");
		if ((time_after(jiffies, start_time + TIMEOUT)) && (newtodo)) {
			printk(KERN_ERR name_suspend "Stopping tasks failed.\n");
			printk(KERN_ERR "Tasks that refused to be refrigerated and haven't since exited:\n");
			read_lock(&tasklist_lock);
			FOR_EACH_THREAD_START {
				if (!to_be_frozen(p)) 
					continue;
				if (p->state == TASK_UNINTERRUPTIBLE)
					continue;
				/* 
				 * We still let the process enter the refrigerator when it's 
				 * ready. It will just flush signals and exit immediately 
				 * because suspend_task isn't set.
				 */
				if (p->flags & PF_FREEZE) {
					printk(" - %s (#%d) signalled but didn't enter refrigerator.\n", 
						p->comm, p->pid);
					show_task(p);
				} else
					printk(" - %s (#%d) wasn't signalled.\n", 
						p->comm, p->pid);
			} FOR_EACH_THREAD_END
			read_unlock(&tasklist_lock);
			goto aborting;
		}
		printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 1, "Update status.     ");
		if ((todo > newtodo) && (!no_progress))
			update_status(todo - newtodo, todo,
					"%d/%d", todo - newtodo, todo);
		printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 1, "Post update status ");
	} while(newtodo && (!TEST_RESULT_STATE(SUSPEND_ABORTED)));

out:
	printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 1, "Left freezer loop.\n");
	freezing = 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	swsusp_spin_unlock_irqrestore(&io_request_lock, iorequestflags);
#endif

	check_shift_keys(1, "Freezing processes completed. ");

	swsusp_state &= ~FREEZE_SMP;

	while (atomic_read(&swsusp_cpu_counter)) {
		cpu_relax();
		barrier();
		check_shift_keys(0, "");
	}

	//swap_list_unlock();

	return newtodo;
aborting:
	newtodo = todo - (processesfrozen - origfrozen);
	goto out;
}

void thaw_processes(void)
{
	struct task_struct FOR_EACH_THREAD_TASK_STRUCTS;
	printnolog(SUSPEND_FREEZER, SUSPEND_LOW, 1, "Thawing tasks\n");
	
	suspend_task = 0;
	swsusp_state = 0;
	IDE_UNSUSPEND
	
	/* 
	 * Pause the other processors so we can safely
	 * change threads' flags
	 */

	swsusp_state |= FREEZE_SMP;
	smp_call_function(smp_pause, NULL, 0, 0);

	while (atomic_read(&swsusp_cpu_counter) < (NUM_CPUS - 1)) {
		cpu_relax();
		barrier();
		check_shift_keys(0, "");
	}

	read_lock(&tasklist_lock);

	FOR_EACH_THREAD_START {
		if (p->flags & PF_FROZEN) {
			printnolog(SUSPEND_FREEZER, SUSPEND_VERBOSE, 0, "Waking %5d: %s.\n", p->pid, p->comm);
			p->flags &= ~PF_FROZEN;
			wake_up_process(p);
		}
	} FOR_EACH_THREAD_END

	read_unlock(&tasklist_lock);

	swsusp_state = 0;

	while (atomic_read(&swsusp_cpu_counter)) {
		cpu_relax();
		barrier();
		check_shift_keys(0, "");
	}
}

/*
 * The following code is based on the reboot notifier code in kernel/sys.c.
 * It is used to notify drivers when a suspend cycle finishes, so that
 * timers which have been stopped can be restarted.
 */


/*
 *	Notifier list for kernel code which wants to be called
 *	at resume. This is used to restart timers which were
 *	stopped during suspend.
 */

static struct notifier_block *resume_notifier_list;
rwlock_t swsusp_notifier_lock = RW_LOCK_UNLOCKED;

/**
 *	register_resume_notifier - Register function to be called at resume time
 *	@nb: Info about notifier function to be called
 *
 *	Registers a function with the list of functions
 *	to be called at resume time.
 *
 *	Currently always returns zero, as notifier_chain_register
 *	always returns zero.
 */
 
int register_resume_notifier(struct notifier_block * nb)
{
	return notifier_chain_register(&resume_notifier_list, nb);
}

/**
 *	unregister_resume_notifier - Unregister previously registered resume notifier
 *	@nb: Hook to be unregistered
 *
 *	Unregisters a previously registered resume
 *	notifier function.
 *
 *	Returns zero on success, or %-ENOENT on failure.
 */
 
int unregister_resume_notifier(struct notifier_block * nb)
{
	return notifier_chain_unregister(&resume_notifier_list, nb);
}

inline int notify_resume(void)
{
	return notifier_call_chain(&resume_notifier_list, 0, NULL);
}

EXPORT_SYMBOL(__swsusp_activity_start);
EXPORT_SYMBOL(swsusp_num_active);
EXPORT_SYMBOL(swsusp_state);
EXPORT_SYMBOL(idletimeout);
