/* pm_disk routines for manipulating ranges.
 *
 * (C) 2003, Nigel Cunningham.
 *
 * These encapsulate the manipulation of ranges. I learnt after writing this
 * code that ranges are more commonly called extents. They work like this:
 *
 * A lot of the data that pm_disk saves involves continguous ranges of memory
 * or storage. Let's say that we're storing data on disk in blocks 1-32768 and
 * 49152-49848 of a swap partition. Rather than recording 1, 2, 3... in arrays
 * pointing to the locations, we simply use:
 *
 * struct range {
 * 	unsigned long min;
 * 	unsigned long max;
 * 	struct range * next;
 * }
 *
 * We can then store 1-32768 and 49152-49848 in 2 struct ranges, using 24 bytes
 * instead of something like 133,860. This is of course inefficient where a range
 * covers only one or two values, but the benefits gained by the much larger
 * ranges more than outweight these instances.
 *
 * Whole pages are allocated to store ranges, with unused structs being chained
 * together and linked into an unused_ranges list:
 *
 * struct range * unused_ranges; (just below).
 *
 * We can fit 341 ranges in a 4096 byte page (rangepage), with 4 bytes left over.
 * These four bytes, referred to as the RangePageLink, are used to link the pages
 * together. The RangePageLink is a pointer to the next page, or'd with the index
 * number of the page.
 *
 * (The following will apply once the range page code is merged and the data is
 * switched over to being stored in ranges)...
 * 
 * RangePages are stored in the header of the suspend image. For portability
 * between suspend time and resume time, we 'relativise' the contents of each page
 * before writing them to disk. That is, each .next and each RangePageLink is
 * changed to point not to an absolute location, but to the relative location in
 * the list of pages. This makes all the information valid and usable (after it
 * has been absolutised again, of course) regardless of where it is reloaded to
 * at resume time.
 */

#include <linux/suspend-common.h>

/* Abbreviations for debugging */
#define RANGE_VERBOSE SUSPEND_RANGES, SUSPEND_VERBOSE
#define RANGE_HIGH SUSPEND_RANGES, SUSPEND_HIGH

struct range * unused_ranges = NULL;
int nr_unused_ranges = 0;
int max_ranges_used = 0;
int num_range_pages = 0;
static unsigned long ranges_allocated = 0;
struct range * first_range_page = NULL, * last_range_page = NULL;

/* Add_range_page
 * Allocates and initialises new pages for storing ranges.
 * Returns 1 on failure to get a page.
 * Otherwise adds the new pages to the unused_ranges pool
 * and returns 0.
 * During resuming, it ensures the page added doesn't
 * collide with memory that will be overwritten when
 * copying the original kernel back.
 */

static int add_range_pages(int number_requested)
{
	int i, j;
	struct range * ranges;
	unsigned long next;
	void **eaten_memory = NULL, **c = NULL, *f;

	for (j = 0; j < number_requested; j++) {
		if (now_resuming) {
			struct page * pageaddr;
			/* Make sure page doesn't collide when we're resuming */
			while ((next = get_zeroed_page(GFP_ATOMIC))) {
				pageaddr = virt_to_page(next);
				if (!PageInUse(pageaddr))
					break;
				eaten_memory = (void *) next;
				*eaten_memory = c;
				c = eaten_memory;
			}
			// Free unwanted memory
			c = eaten_memory;
			while(c) {
				f = c;
				c = *c;
				if (f)
					free_page((unsigned long) f);
			}
			eaten_memory = NULL;
		} else {
			next = get_zeroed_page(GFP_ATOMIC);
			if (!next)
				next = get_grabbed_pages(0);
		}


		if (!next) {
			printk("Failed to allocate a new range page.\n");
			return 1;
		}

		num_range_pages++;
		printnolog(SUSPEND_RANGES, SUSPEND_LOW, 0, "Got a new range page (#%d) at %lx\n",
			num_range_pages, next);
		if (!first_range_page)
			first_range_page = (struct range *) next;
		if (last_range_page)
			*RANGEPAGELINK(last_range_page) |= next;
		*RANGEPAGELINK(next) = num_range_pages;
		last_range_page = (struct range *) next;
		ranges = (struct range *) next;
		for (i = 0; i < RANGES_PER_PAGE; i++)
			(ranges+i)->next = (ranges+i+1);
		(ranges + i - 1)->next = unused_ranges;
		unused_ranges = ranges;
		nr_unused_ranges += i;
		if (console_loglevel > 5) {
			for (i = 0; i < RANGES_PER_PAGE; i++)
				printnolog(SUSPEND_RANGES, SUSPEND_VERBOSE, 0, "Ranges %d = %p. Next = %p\n", 
					i, 
					(ranges + i),
					(ranges + i)->next);
		}
	}
	return 0;
}


/* 
 * Free ranges.
 *
 * Frees pages allocated by add_range_pages()
 *
 * Checks that all ranges allocated have been freed and aborts
 * if this is not true.
 *
 * Ranges may not be in memory order but we don't
 * mind. We just look for a range that is on a
 * page boundary. That gives us the pages to be
 * freed. As we find them, we link them together
 * into a new chain (we're not going to use the
 * other ranges anyway) and then free the chain.
 */

int free_ranges(void)
{
	int i;
	struct range * this_range_page = first_range_page, * next_range_page = NULL;

	if (ranges_allocated)
		printk(" *** Warning: %ld ranges still allocated when free_ranges() called.\n", ranges_allocated);

	for (i = 0; i < num_range_pages; i++) {
		printnolog(SUSPEND_RANGES, SUSPEND_VERBOSE, 0, "Freeing range page %p.\n", this_range_page);
		next_range_page = (struct range *) (((unsigned long) (*RANGEPAGELINK(this_range_page))) & PAGE_MASK);
		free_pages((unsigned long) this_range_page, 0);
		this_range_page = next_range_page;
	}

	nr_unused_ranges = num_range_pages = ranges_allocated = 0;
	unused_ranges = last_range_page = first_range_page = NULL;

	return 0;
}

/* get_range
 *
 * Returns a free range, having removed it from the
 * unused list and having incremented the usage counter.
 * May imply allocating a new page and may therefore 
 * fail, returning NULL instead.
 *
 * No locking. This is because we are only called 
 * from swsusp, which is single threaded
 */

static struct range * get_range(void)
{
	struct range * result;
	
	if ((!unused_ranges) && (add_range_pages(1)))
		return NULL;

	result = unused_ranges;
	unused_ranges = unused_ranges->next;
	nr_unused_ranges--;
	ranges_allocated++;
	if (ranges_allocated > max_ranges_used)
		max_ranges_used++;
	result->minimum = result->maximum = 0;
	result->next = NULL;
	return result;
}

/*
 * put_range.
 *
 * Returns a range to the pool of unused pages and
 * decrements the usage counter.
 *
 * Assumes unlinking is done by the caller.
 */
void put_range(struct range * range)
{
	if (!range) {
		printk("Error! put_range called with NULL range.\n");
		return;
	}
	range->minimum = range->maximum = 0;
	range->next = unused_ranges;
	unused_ranges = range;
	ranges_allocated--;
	nr_unused_ranges++;
}

/*
 * put_range_chain.
 *
 * Returns a whole chain of ranges to the
 * unused pool.
 */
void put_range_chain(struct rangechain * chain)
{
	int count = 0;
	struct range * this;

	printnolog(RANGE_HIGH, 0, "\n");
	
	if (chain->first) {
		this = chain->first;
		while (this) {
			this->minimum = this->maximum = 0;
			this=this->next;
		}
		chain->last->next = unused_ranges;
		unused_ranges = chain->first;
		chain->first = NULL;
		count = chain->allocs - chain->frees;
		ranges_allocated -= count;
		nr_unused_ranges += count;

		printnolog(SUSPEND_RANGES, SUSPEND_MEDIUM, 0, "Emptied %s. For this chain, allocated %d and freed %d. Used optimisation %d times.\n",
				chain->name,
				chain->allocs,
				chain->frees,
				chain->timesusedoptimisation);
		chain->allocs = 0;
		chain->frees = 0;
		chain->size = 0;
		chain->timesusedoptimisation = 0;
		chain->lastaccessed = NULL;	/* Invalidate optimisation info */
		chain->last = NULL;
	}
}

/* printmethod:
 * 0: integer
 * 1: hex
 * 2: page number
 */
void print_chain(int debuglevel, struct rangechain * chain, int printmethod)
{
	struct range * this = chain->first;
	int count = 0, size = 0;
	
	if ((console_loglevel < debuglevel) || (!this)
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
	    || (!TEST_DEBUG_STATE(SUSPEND_RANGES)))
#else
	  )
#endif
		return;

	if (!chain->name)
		printnolog(RANGE_VERBOSE, 0,"Chain %p\n", chain);
	else
		printnolog(RANGE_VERBOSE, 0,"%s\n", chain->name);
	
	while (this) {
		/* 'This' is printed separately so it is displayed if an oops results */
		switch (printmethod) {
			case 0:
				printnolog(RANGE_VERBOSE, 0,"(%p) ",
					this);
				printnolog(RANGE_VERBOSE, 0,"%lx-%lx; ",
					this->minimum, this->maximum);
				break;
			case 1:
				printnolog(RANGE_VERBOSE, 0,"(%p)",
					this);
				printnolog(RANGE_VERBOSE, 0,"%lu-%lu; ",
					this->minimum, this->maximum);
				break;
			case 2:
				printnolog(RANGE_VERBOSE, 0,"(%p)",
					this);
				printnolog(RANGE_VERBOSE, 0,"%p-%p; ",
					page_address(mem_map+this->minimum),
					page_address(mem_map+this->maximum) + PAGE_SIZE - 1);
				break;
		}
		size+= this->maximum - this->minimum + 1;
		this = this->next;
		count++;
		if (!(count%4))
			printnolog(RANGE_VERBOSE, 0,"\n");
	}
	
	if ((count%4))
		printnolog(RANGE_VERBOSE, 0,"\n");

	printnolog(RANGE_VERBOSE, 0,"%d entries/%ld allocated. Allocated %d and freed %d. Size %d.",
			count, 
			ranges_allocated,
			chain->allocs,
			chain->frees,
			size);
	if (count != (chain->allocs - chain->frees)) {
		chain->debug = 1;
		check_shift_keys(1, "Discrepancy in chain.");
	}
	printnolog(RANGE_VERBOSE, 0,"\n");
}

static void check_chain(struct rangechain * chain, char * failuremessage)
{
	struct range * this = chain->first;
	int count = 0;
	
	while (this) {
		count++;
		this = this->next;
	}
	
	if (count != (chain->allocs - chain->frees)) {
		printk("Descrepancy in chain found %s.\n", failuremessage);
		print_chain(SUSPEND_ERROR, chain, 2);
	}
}

/*
 * add_to_range_chain.
 *
 * Takes a value to be stored and a pointer
 * to a chain and adds the value to the range
 * chain, merging with an existing range or 
 * adding a new entry as necessary. Ranges 
 * are stored in increasing order.
 *
 * Values should be consecutive, and so may
 * need to be transformed first. (eg for
 * pages, would want to call with page-mem_map).
 *
 * Important optimisation:
 * We store in the chain info the location of
 * the last range accessed or added (and its
 * previous). If the next value is outside this
 * range by one, we start from the previous
 * entry instead of the start of the chain.
 * In cases of heavy fragmentation, this saves
 * a lot of time searching.
 * 
 * Returns:
 * 0 if successful
 * 1 if the value is already included.
 * 2 if unable to allocate memory.
 * 3 if fall out bottom (shouldn't happen).
 */

int add_to_range_chain(struct rangechain * chain, unsigned long value)
{
	struct range * this, * prev = NULL, * prevtoprev = NULL;
	int usedoptimisation = 0;
	
	printnolog(RANGE_HIGH, 0, "\n");
	
	if (!chain->first) {	/* Empty */
		chain->last = chain->first = get_range();
		if (!chain->first) {
			printk("Error unable to allocate the first range for the chain.\n");
			return 2;
		}
		chain->allocs++;
		chain->first->maximum = value;
		chain->first->minimum = value;
		printnolog(RANGE_HIGH, 0, "%s: Added initial entry %p to chain.\n", 
				chain->name,
				chain->first);
		print_chain(SUSPEND_VERBOSE, chain, 0);
		chain->size++;
		return 0;
	}
	
	this = chain->first;

	if (chain->lastaccessed && chain->prevtolastaccessed && chain->prevtoprev) {
		if ((value + 1) == chain->lastaccessed->minimum) {
			prev = chain->prevtoprev;
			this = chain->prevtolastaccessed;
			usedoptimisation = 1;
		} else if (((value - 1) == chain->lastaccessed->maximum)) {
			prev = chain->prevtolastaccessed;
			this = chain->lastaccessed;
			usedoptimisation = 1;
		}
		
		/* prev wrong but doesn't matter */
		if (usedoptimisation) {
			printnolog(RANGE_HIGH, 0,
				"Adding %ld, using last accessed: (%ld-%ld)(%ld-%ld)\n",
				value,
				prev->minimum,
				prev->maximum,
				this->minimum,
				this->maximum);
			chain->timesusedoptimisation++;
		}
	}

	while (this) {
		/* Need new entry prior to this? */
		if ((value + 1) < this->minimum) {
			struct range * new = get_range();
			if (!new) {
				printk("Error unable to insert a new range for the chain.\n");
				return 2;
			}
			chain->allocs++;
			new->minimum = value;
			new->maximum = value;
			new->next = this;
			/* Prior to start of chain? */
			if (!prev)
				chain->first = new;
			else
				prev->next = new;
			printnolog(RANGE_HIGH, 0, 
					"%s: Adding new range %p prior to existing range.\n",
					chain->name,
					new);
			print_chain(SUSPEND_VERBOSE, chain, 0);
			if (!usedoptimisation) {
				chain->prevtoprev = prevtoprev;
				chain->prevtolastaccessed = prev;
				chain->lastaccessed = new;
			}
			if (TEST_ACTION_STATE(SUSPEND_RANGE_PARANOIA)) {
				if (usedoptimisation)
					check_chain(chain, "after adding new range prior to existing range with optimisation.\n");
				else
					check_chain(chain, "after adding new range prior to existing range.\n");
			}
			chain->size++;
			return 0;
		}

		if ((this->minimum <= value) && (this->maximum >= value)) {
			if (chain->name)
				printk("%s:", chain->name);
			else
				printk("%p:", chain);
			printk("Trying to add a value (%ld/0x%lx) already included in chain.\n",
					value, value);
			print_chain(SUSPEND_ERROR, chain, 0);
			check_shift_keys(1, NULL);
			return 1;
		}
		if ((value + 1) == this->minimum) {
			this->minimum = value;
			print_chain(SUSPEND_VERBOSE, chain, 0);
			if (!usedoptimisation) {
				chain->prevtoprev = prevtoprev;
				chain->prevtolastaccessed = prev;
				chain->lastaccessed = this;
			}
			if (TEST_ACTION_STATE(SUSPEND_RANGE_PARANOIA)) {
				if (usedoptimisation)
					check_chain(chain, "after expanding existing range downwards with optimisation.\n");
				else
					check_chain(chain, "after expanding existing range downwards..\n");
			}
			chain->size++;
			return 0;
		}
		if ((value - 1) == this->maximum) {
			if ((this->next) && (this->next->minimum == value + 1)) {
				struct range * oldnext = this->next;
				this->maximum = this->next->maximum;
				this->next = this->next->next;
				if ((chain->last) == oldnext)
					chain->last = this;
				put_range(oldnext);
				chain->lastaccessed = NULL;	/* Invalidate optimisation info */
				chain->frees++;
				printnolog(SUSPEND_RANGES,
						SUSPEND_HIGH,
						0,
						"%s: Putting merged range %p.\n", 
						chain->name,
						oldnext);
				print_chain(SUSPEND_VERBOSE, chain, 0);
				if (!usedoptimisation) {
					chain->prevtoprev = prevtoprev;
					chain->prevtolastaccessed = prev;
					chain->lastaccessed = this;
				}
				if (TEST_ACTION_STATE(SUSPEND_RANGE_PARANOIA)) {
					if (usedoptimisation)
						check_chain(chain, "after merging this and next ranges with optimisation.\n");
					else
						check_chain(chain, "after merging this and next ranges.\n");
				}
				chain->size++;
				return 0;
			} 
			this->maximum = value;
			print_chain(SUSPEND_VERBOSE, chain, 0);
			if (!usedoptimisation) {
				chain->prevtoprev = prevtoprev;
				chain->prevtolastaccessed = prev;
				chain->lastaccessed = this;
			}
			if (TEST_ACTION_STATE(SUSPEND_RANGE_PARANOIA)) {
				if (usedoptimisation)
					check_chain(chain, "after expanding a range upwards with optimisation.\n");
				else
					check_chain(chain, "after expanding a range upwards.\n");
			}
			chain->size++;
			return 0;
		}
		if (!this->next) {
			struct range * new = get_range();
			if (!new) {
				printk("Error unable to append a new range to the chain.\n");
				return 2;
			}
			chain->allocs++;
			new->minimum = value;
			new->maximum = value;
			new->next = NULL;
			this->next = new;
			chain->last = new;
			printnolog(SUSPEND_RANGES,
					SUSPEND_HIGH,
					0,
					"%s: Got new range %p for end of chain.\n",
					chain->name,
					new);
			print_chain(SUSPEND_VERBOSE, chain, 0);
			if (!usedoptimisation) {
				chain->prevtoprev = prev;
				chain->prevtolastaccessed = this;
				chain->lastaccessed = new;
			}
			if (TEST_ACTION_STATE(SUSPEND_RANGE_PARANOIA)) {
				if (usedoptimisation)
					check_chain(chain, "after appending a new range with optimisation.\n");
				else
					check_chain(chain, "after appending a new range.\n");
			}
			chain->size++;
			return 0;
		}
		prevtoprev = prev;
		prev = this;
		this = this->next;
	}
	printk("\nFell out the bottom of add_to_range_chain. This shouldn't happen!\n");
	SET_RESULT_STATE(SUSPEND_ABORTED);
	return 3;
}

/* append_range
 * Used where we know a range is to be added to the end of the list
 * and does not need merging with the current last range.
 * (count_data_pages only at the moment)
 */

int append_range_to_range_chain(struct rangechain * chain, unsigned long minimum, unsigned long maximum)
{
	struct range * newrange = NULL;

	newrange = get_range();
	if (!newrange) {
		printk("Error unable to append a new range to the chain.\n");
		return 2;
	}

	chain->allocs++;
	chain->size+= (maximum - minimum + 1);
	newrange->minimum = minimum;
	newrange->maximum = maximum;
	newrange->next = NULL;

	if (chain->last) {
		chain->last->next = newrange;
		chain->last = newrange;
	} else 
		chain->last = chain->first = newrange;

	printnolog(SUSPEND_RANGES,
		SUSPEND_HIGH,
		0,
		"%s: Got new range %p for end of chain.\n",
		chain->name,
		newrange);
	print_chain(SUSPEND_VERBOSE, chain, 0);
	/* No need to reset optimisation info since added to end */
	return 0;
}

int append_to_range_chain(int chain, unsigned long min, unsigned long max)
{
	int result = 0;

	switch (chain) {
		case 0:
			return 0;
		case 1:
			result = append_range_to_range_chain(&pagedir1.origranges, min, max);
			break;
		case 2:
			result = append_range_to_range_chain(&pagedir2.origranges, min, max);
			if (!result)
				result = append_range_to_range_chain(&pagedir1.destranges, min, max);
	}
	check_shift_keys(0, NULL);
	return result;
}

/* 
 * Prepare rangesets for save by translating addresses to relative indices.
 */
void relativise_ranges(void)
{
	struct range * this_range_page = first_range_page;
	int i;
	
	while (this_range_page) {
		struct range * this_range = this_range_page;
		for (i = 0; i < RANGES_PER_PAGE; i++) {
			printnolog(SUSPEND_RANGES, SUSPEND_VERBOSE, 0, "Setting pointer for range %p ",
				this_range);
				printnolog(SUSPEND_RANGES, SUSPEND_VERBOSE, 0, "Orig: %p ",
					this_range->next);
			if (this_range->next) {
				printnolog(SUSPEND_RANGES, SUSPEND_VERBOSE, 0, "New: %lx. ",
					(((unsigned long) this_range->next) & (PAGE_SIZE - 1)) |
						((*RANGEPAGELINK(this_range->next) & (PAGE_SIZE - 1)) << PAGE_SHIFT));
				printnolog(SUSPEND_RANGES, SUSPEND_VERBOSE, 0, "[%lx] %lx | %lx. ",
					*RANGEPAGELINK(this_range->next),
					(((unsigned long) this_range->next) & (PAGE_SIZE - 1)),
					(((*RANGEPAGELINK(this_range->next)) & (PAGE_SIZE - 1)) << PAGE_SHIFT));
			}

			if (this_range->next)
				this_range->next = RANGE_RELATIVE(this_range->next);
			
				printnolog(SUSPEND_RANGES, SUSPEND_VERBOSE, 0, " Value now %p.\n", this_range->next);
			this_range++;
		}
		this_range_page = (struct range *) ((*RANGEPAGELINK(this_range_page)) & PAGE_MASK);
	}
	printnolog(SUSPEND_RANGES, SUSPEND_VERBOSE, 0, "\n");
}

/* Convert ->next pointers for ranges back to absolute values.
 * The issue is finding out what page the absolute value is now at.
 * If we use an array of values, we gain speed, but then we need to
 * be able to allocate contiguous pages. Fortunately, this is done
 * prior to loading pagesets, so we can just allocate the pages
 * needed, set up our array and use it and then discard the data
 * before we exit.
 */

void absolutise_ranges(unsigned long * range_pages)
{
	struct range * this_range_page = first_range_page;
	int i;
	
	this_range_page = first_range_page;
	while (this_range_page) {
		struct range * this_range = this_range_page;
		for (i = 0; i < RANGES_PER_PAGE; i++) {
			printnolog(SUSPEND_RANGES, SUSPEND_VERBOSE, 0, "Setting range %p ",
				this_range);
			printnolog(SUSPEND_RANGES, SUSPEND_VERBOSE, 0, "Orig: %p ",
				this_range->next);
			if (this_range->next) {
				printnolog(SUSPEND_RANGES, SUSPEND_VERBOSE, 0, "New: %p. ",
					RANGE_ABSOLUTE(this_range->next, range_pages));
			}

			if (this_range->next) 
				this_range->next = RANGE_ABSOLUTE(this_range->next, range_pages);

			printnolog(SUSPEND_RANGES, SUSPEND_VERBOSE, 0, " Value is now %p.\n", this_range->next);
			this_range++;
		}
		printnolog(SUSPEND_RANGES, SUSPEND_VERBOSE, 0, "RangePagelink:");
		printnolog(SUSPEND_RANGES, SUSPEND_VERBOSE, 0, "%lx\n", *RANGEPAGELINK(this_range_page));
		this_range_page = (struct range *) ((*RANGEPAGELINK(this_range_page)) & PAGE_MASK);
	}
}

void absolutise_chain(struct rangechain * chain, unsigned long * range_pages)
{
	if (chain->first)
		chain->first = RANGE_ABSOLUTE(chain->first, range_pages);
	if (chain->last)
		chain->last = RANGE_ABSOLUTE(chain->last, range_pages);
	if (chain->lastaccessed)
		chain->lastaccessed = RANGE_ABSOLUTE(chain->lastaccessed, range_pages);
	if (chain->prevtolastaccessed)
		chain->prevtolastaccessed = RANGE_ABSOLUTE(chain->prevtolastaccessed, range_pages);
	if (chain->prevtoprev)
		chain->prevtoprev = RANGE_ABSOLUTE(chain->prevtoprev, range_pages);
}

void relativise_chain(struct rangechain * chain)
{
	if (chain->first)
		chain->first = RANGE_RELATIVE(chain->first);
	if (chain->last)
		chain->last = RANGE_RELATIVE(chain->last);
	if (chain->lastaccessed)
		chain->lastaccessed = RANGE_RELATIVE(chain->lastaccessed);
	if (chain->prevtolastaccessed)
		chain->prevtolastaccessed = RANGE_RELATIVE(chain->prevtolastaccessed);
	if (chain->prevtoprev)
		chain->prevtoprev = RANGE_RELATIVE(chain->prevtoprev);
}

#define POINTERS_PER_PAGE (PAGE_SIZE / sizeof(void *))
static int rangepages_list_order = -1;
unsigned long * get_rangepages_list(void)
{
	struct range * this_range_page = first_range_page;
	int i, pages_needed, pool_on = (swsusp_state & USE_MEMORY_POOL);
	unsigned long * range_pages;

	pages_needed = ((num_range_pages + POINTERS_PER_PAGE - 1) / POINTERS_PER_PAGE);
	printnolog(SUSPEND_RANGES, SUSPEND_LOW, 0, "Need %d pages for %d range pages.\n", pages_needed, num_range_pages);
	swsusp_state &= ~USE_MEMORY_POOL;
	rangepages_list_order = get_bitmask_order(pages_needed - 1);
	range_pages = (unsigned long *) __get_free_pages(GFP_ATOMIC, rangepages_list_order);
	swsusp_state |= pool_on;
	if (!range_pages)
		range_pages = (unsigned long *) get_grabbed_pages(get_bitmask_order(pages_needed - 1));
	if (!range_pages)
		return NULL;
	printnolog(SUSPEND_RANGES, SUSPEND_VERBOSE, 0, "Range pages is %p.\n", range_pages);
	
	for (i = 1; i <= num_range_pages; i++) {
		range_pages[i] = (unsigned long) this_range_page;

		printnolog(SUSPEND_RANGES, SUSPEND_VERBOSE, 0, "This range page is %p.", this_range_page);
		printnolog(SUSPEND_RANGES, SUSPEND_VERBOSE, 0, "Links to %lx.", *(RANGEPAGELINK(this_range_page)));
		this_range_page = (struct range *) (((unsigned long) (*RANGEPAGELINK(this_range_page))) & PAGE_MASK);
		printnolog(SUSPEND_RANGES, SUSPEND_VERBOSE, 0, "This range page is now %p.\n", this_range_page);
	}

	return range_pages;
}

void put_rangepages_list(unsigned long * range_pages)
{
	int pool_on = (swsusp_state & USE_MEMORY_POOL);

	swsusp_state &= ~USE_MEMORY_POOL;
	free_pages((unsigned long) range_pages, rangepages_list_order);
	swsusp_state |= pool_on;
	rangepages_list_order = -1;
}

int relocate_rangepages(unsigned long * range_pages)
{
	void **eaten_memory = NULL;
	void **c = eaten_memory, *m = NULL, *f;
	int oom = 0, i, numeaten = 0;

	printlog(SUSPEND_PAGESETS, SUSPEND_VERBOSE,"Relocating conflicting parts of range pages.\n");

	for (i = 1; i <= num_range_pages; i++) {
		int this_collides = 0;

		this_collides = PageInUse(virt_to_page(range_pages[i]));

		if (this_collides) {
			printlog(SUSPEND_PAGESETS, SUSPEND_VERBOSE,"\nRelocating range page %d.\n", i);
			while ((m = (void *) get_zeroed_page(GFP_ATOMIC))) {
				memset(m, 0, PAGE_SIZE);
				if (!PageInUse(virt_to_page(m))) {
					copy_page(m, (void *) range_pages[i]);
					free_page(range_pages[i]);
					printnolog(SUSPEND_PAGESETS, SUSPEND_VERBOSE, 0, "\nRelocated page %d from %lx to %p.\n",
							i,
							range_pages[i],
							m);
					if (i == 1)
						first_range_page = m;
					else
						*RANGEPAGELINK(range_pages[i-1]) = (i | (unsigned long) m);
					range_pages[i] = (unsigned long) m;
					break;
				}
				numeaten++;
				eaten_memory = m;
				printnolog(SUSPEND_PAGESETS, SUSPEND_VERBOSE, 0, "Eaten: %d. Still to try:%d\r", numeaten, nr_free_pages()); 
				*eaten_memory = c;
				c = eaten_memory;
			}

			if (!m) {
				printk("\nRan out of memory trying to relocate rangepages (tried %d pages).\n", numeaten);
				oom = 1;
				break;
			}
		}
	}
		
	printnolog(SUSPEND_PAGESETS, SUSPEND_VERBOSE, 0, "\nFreeing rejected memory locations...\n");

	c = eaten_memory;
	while(c) {
		f = c;
		c = *c;
		if (f)
			free_pages((unsigned long) f, 0);
	}
	eaten_memory = NULL;
	
	printlog(SUSPEND_PAGESETS, SUSPEND_VERBOSE,"\n");
	
	check_shift_keys(1, "Pagedir relocated. ");

	if (oom) 
		return -ENOMEM;
	else
		return 0;
}

