/*
 * Swapwriter.c
 *
 * Copyright 2003 Nigel Cunningham <ncunningham@clear.net.nz>
 *
 * Distributed under GPLv2.
 * 
 * This file encapsulates functions for usage of swap space as a
 * backing store.
 *
 * It currently contains the 2.4 and 2.5 versions in one file. This is
 * only TEMPORARY, as a means of not duplicating work while the big
 * cleanup is being done.
 *
 * Table of contents:
 * - Forward declarations of functions and structures
 * - Internal data structure declarations with helpers
 * - Internal variable declarations
 * - Higher level functions
 *   a) Lowlevel I/O
 *   b) Swap allocation
 * - Exported functions
 * - Ops declaration
 */

#include <linux/suspend-common.h>
struct swsusp_plugin_ops swapwriterops;

#define SIGNATURE_VER 6

/* ---------------------- 2.6 SPECIFIC --------------- */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
#include <linux/blkdev.h>
#include <linux/swapops.h>
#include <linux/buffer_head.h>

#define DEVICE_ID_TYPE dev_t
#define DEVICE_BLOCK_TYPE struct block_device *
#define DEVICE_BLOCK_NONE NULL
#define SYS_IO_STRUCT struct bio
#define SYS_ZONE_TYPE long
#define SIG_BDEV_START 2
#define SWP_OFFSET swp_offset
#define SWP_TYPE swp_type
#define SWP_ENTRY swp_entry
#define UNUSED_SWAP_ENTRY(i) (!swap_info[i].swap_file)
#define SWAP_FILE_INODE(i) swap_info[i].swap_file->f_dentry->d_inode
#define SWAP_IS_PARTITION(i) (S_ISBLK(SWAP_FILE_INODE(i)->i_mode))
#define SWAP_DEVICE_ID(i) (SWAP_FILE_INODE(i)->i_rdev)
#define SWAP_DEVICE_BDEV(i) (swap_info[i].bdev)
#define BDEV_TO_DEVICE_ID(bdev) ((bdev)->bd_dev)
#define SWAP_BLOCKSIZE(bdev) (block_size(bdev))
#define RESUME_BDEV(i) swap_info[i].bdev
#define RUN_IO do { blk_run_queues(); io_schedule(); } while(0)
#define SYNC_IO do { RUN_IO; sys_sync(); schedule(); check_shift_keys(0, NULL); } while(0)

/* ---------------------- 2.4 SPECIFIC --------------- */
#else
#include <linux/locks.h>
#include <linux/swapctl.h>
#include <linux/blk.h>

#define DEVICE_ID_TYPE kdev_t
#define DEVICE_BLOCK_TYPE kdev_t
#define DEVICE_BLOCK_NONE 0
#define SYS_IO_STRUCT struct buffer_head
#define SYS_ZONE_TYPE int
#define SIG_BDEV_START 3
#define UNUSED_SWAP_ENTRY(i) ((!swap_info[i].swap_device) && (!swap_info[i].swap_file))
#define SWAP_IS_PARTITION(i) (swap_info[i].swap_device)
#define SWAP_PARTITION_BDEV(i) swap_info[i].swap_device
#define SWAP_FILE_INODE(i) swap_info[i].swap_file->d_inode
#define SWAP_FILE_BDEV(i) SWAP_FILE_INODE(i)->i_dev
#define SWAP_DEVICE_ID(i) (SWAP_IS_PARTITION(i) ? SWAP_PARTITION_BDEV(i) : SWAP_FILE_BDEV(i))
#define SWAP_DEVICE_BDEV(i) SWAP_DEVICE_ID(i)
#define BDEV_TO_DEVICE_ID(bdev) (bdev)
#define SWAP_BLOCKSIZE(bdev) (blksize_size[MAJOR(bdev)][MINOR(bdev)])
#define SWAP_FILE_BLOCKSIZE(i) SWAP_FILE_INODE(i)->i_sb->s_blocksize
#define SWAP_DESCRIPTION(i) (SWAP_IS_PARTITION(i) ? SWAP_PARTITION_BDEV(i) : SWAP_FILE_BDEV(i))
#define RESUME_BDEV(i) swap_info[i].swap_device
#define RUN_IO do { run_task_queue(&tq_disk); } while(0)
static void sync_swap_partitions(void);
#define SYNC_IO do { sync_swap_partitions(); do_suspend_sync(); } while(0)

#endif

/* --------- END OF VERSION SPECIFIC DECLARES --------- */

// Forward Declarations

struct io_info;
struct submit_params;

/* --- Struct of pages stored on disk */

struct swaplink {
	char dummy[PAGE_SIZE - sizeof(swp_entry_t)];
	swp_entry_t next;
};

union diskpage {
	union swap_header swh;	/* swh.magic is the only member used */
	struct swaplink link;
	struct suspend_header sh;
};

union p_diskpage {
	union diskpage *pointer;
	char *ptr;
        unsigned long address;
};

#define SIGNATURE_LENGTH 10

// Lowlevel I/O functions
// - Get/set block size.
static inline int swsusp_get_block_size(DEVICE_BLOCK_TYPE bdev);
static inline int swsusp_set_block_size(DEVICE_BLOCK_TYPE bdev, int size);

// - Get blocks for a swap entry.
static int get_phys_params(swp_entry_t entry);

// - Get blocks for a header page (can't use chains).
static int get_header_params(struct submit_params * headerpage);

// - Manage structures that store info on in-progress I/O.
static struct io_info * get_io_info_struct(void);

// - Manage statistics for I/O structures.
void __inline reset_io_stats(void);
void check_io_stats(void);

// - Submit & cleanup I/O.
static struct io_info * start_one(int rw, struct submit_params * submit_info);
static void cleanup_one(struct io_info * io_info);
static inline void cleanup_completed_io(void);
static void finish_all_io(void);

// - Manage swap signature.
static int prepare_signature(struct submit_params * first_header_page, char * current_header);
static int parse_signature(char * signature, int restore);

// Higher Level

// - Synchronous (async + wait on completion)
static struct io_info * do_swsusp_io(int rw, struct submit_params * submit_info, int syncio);
static void bdev_page_io(int rw, DEVICE_BLOCK_TYPE bdev, long pos, struct page * page);

/* Don't set MAX_READAHEAD > 31. We're only using an unsigned long */
#define MAX_READAHEAD 31
static struct page * readahead_page[MAX_READAHEAD];
static int readahead_index = -1;
static unsigned long readahead_flags = 0;

/*
 * ---------------------------------------------------------------
 *
 *     Internal Data Structures
 *
 * ---------------------------------------------------------------
 */

/* header_data contains data that is needed to reload pagedir1, and
 * is therefore saved in the suspend header.
 *
 * Pagedir2 swap comes before pagedir1 swap (save order), and the first swap
 * entry for pagedir1 to use is set when pagedir2 is written (when we know how
 * much swap it used). Since this first entry is almost certainly not at the
 * start of a range, the firstoffset variable below tells us where to start in
 * the range. All of this means we don't have to worry about getting different
 * compression ratios for the kernel and cache (when compressing the image).
 * We can simply allocate one pool of swap (size determined using expected
 * compression ratio) and use it without worrying whether one pageset
 * compresses better and the other worse (this is what happens). As long as the
 * user gets the expected compression right, it will work.
 */

struct {
	/* Range chains for swap & blocks */
	struct rangechain swapranges;
	struct rangechain block_chain[MAX_SWAPFILES];
	
	/* Location of start of pagedir 1 */
	struct range * pd1start_block_range;
	unsigned long pd1start_block_offset;
	int pd1start_chain;

	/* Devices used for swap */
	DEVICE_ID_TYPE swapdevs[MAX_SWAPFILES];
	char blocksizes[MAX_SWAPFILES];

	/* Asynchronous I/O limit */
	int max_async_ios;
} header_data;


/*
 * ---------------------------------------------------------------
 *
 *     IO in progress information storage and helpers
 *
 * ---------------------------------------------------------------
 */

struct io_info {
	SYS_IO_STRUCT * sys_struct;
	SYS_ZONE_TYPE blocks[PAGE_SIZE/512];
	struct page * cache_page;
	struct page * buffer_page;
	struct page * data_page;
	unsigned long flags;
	DEVICE_BLOCK_TYPE dev;
	int blocks_used;
	int block_size;
	struct list_head list;
	int readahead_index;
};

/* [Max] number of pages used for above struct */
static int infopages = 0;
static int maxinfopages = 0;

/* [Max] number of I/O operations pending */
static int outstanding_io = 0;
static int max_outstanding_io = 0;

DEVICE_ID_TYPE header_device = 0;
DEVICE_BLOCK_TYPE header_block_device = DEVICE_BLOCK_NONE;
struct range * this_range_page = NULL, * next_range_page = NULL;
int headerblocksize = PAGE_SIZE;
int headerblock;

/* Bits in struct io_info->flags */
#define IO_WRITING 1
#define IO_RESTORE_PAGE_PROT 2
#define IO_AWAITING_READ 3
#define IO_AWAITING_WRITE 4
#define IO_CLEANUP_IN_PROGRESS 5
#define IO_HANDLE_PAGE_PROT 6

static LIST_HEAD(ioinfo_free);
static LIST_HEAD(ioinfo_ready_for_cleanup);
static LIST_HEAD(ioinfo_busy);
static spinlock_t ioinfo_lists_lock = SPIN_LOCK_UNLOCKED;
static spinlock_t readahead_flags_lock = SPIN_LOCK_UNLOCKED;

static char swapfilename[256] = "";	/* For swapfile automatically swapon/off'd. */
extern asmlinkage long sys_swapon(const char * specialfile, int swap_flags);

/* Must be silent - might be called from cat /proc/swsusp/debug_info
 * Returns 0 if was off, -EBUSY if was on, error value otherwise.
 */
static int enable_swapfile(void)
{
	int activateswapresult = -EINVAL;

	if (swapfilename[0]) {
		/* Attempt to swap on with maximum priority */
		activateswapresult = sys_swapon(swapfilename, 0xFFFF);
		if (activateswapresult && (activateswapresult != -EBUSY))
			printk(name_swsusp "The swapfile/partition specified by /proc/swsusp/swapfile (%s) could not be turned on (error %d). Attempting to continue.\n",
				swapfilename, activateswapresult);
	}
	return activateswapresult;
}

extern asmlinkage long sys_swapoff(const char * specialfile);
/* Returns 0 if was on, -EINVAL if was off, error value otherwise */
static int disable_swapfile(void)
{
	int result = -EINVAL;
	
	if (swapfilename[0]) {
		result = sys_swapoff(swapfilename);
		if (result == -EINVAL)
	 		return 0;	/* Wasn't on */
	}

	return result;
}

static int manage_swapfile(int enable)
{
	static int result;

	if (enable)
		result = enable_swapfile();
	else
		result = disable_swapfile();

	return result;
}
/* Returns pointer to page which was cleaned up on success, -ENODATA if none ready for cleanup */
static struct io_info * ioinfo_cleanup_one(void)
{
	struct io_info * this;
	int readahead_index;
	unsigned long flags;

	if (list_empty(&ioinfo_ready_for_cleanup))
		return ERR_PTR(-ENODATA);

	spin_lock_irqsave(&ioinfo_lists_lock, flags);
	this = list_entry(ioinfo_ready_for_cleanup.next, struct io_info, list);
	list_del(&this->list);
	spin_unlock_irqrestore(&ioinfo_lists_lock, flags);

	readahead_index = this->readahead_index;
	this->readahead_index = 0;

	cleanup_one(this);

	if (readahead_index) {
		spin_lock_irqsave(&readahead_flags_lock, flags);
		set_bit(readahead_index - 1, &readahead_flags);
		spin_unlock_irqrestore(&readahead_flags_lock, flags);
	}

	spin_lock_irqsave(&ioinfo_lists_lock, flags);
	list_add_tail(&this->list, &ioinfo_free);
	spin_unlock_irqrestore(&ioinfo_lists_lock, flags);
	return this;
}

static struct io_info * get_io_info_struct(void)
{
	unsigned long newpage = 0, flags;
	struct io_info * this = NULL;
	int iteration = 0, remaining = 0;

	do {
		if (!(iteration%8))
			RUN_IO;
		iteration++;
	
		/* We clean up an IO info struct if possible */
		ioinfo_cleanup_one();

		if ((max_async_ios) && (outstanding_io >= max_async_ios))
			continue;

		if (!list_empty(&ioinfo_free)) {
			spin_lock_irqsave(&ioinfo_lists_lock, flags);
			break;
		}

		/* Need to allocate a new page */
		newpage = get_zeroed_page(GFP_ATOMIC);
		if (!newpage)
			continue;

		printlog(SUSPEND_MEMORY, SUSPEND_VERBOSE, "[NewIOPage %lx]", newpage);
		infopages++;
		if (infopages > maxinfopages)
			maxinfopages++;
		this = (struct io_info *) newpage;
		remaining = PAGE_SIZE;
		spin_lock_irqsave(&ioinfo_lists_lock, flags);
		while (remaining >= (sizeof(struct io_info))) {
			list_add_tail(&this->list, &ioinfo_free);
			this = (struct io_info *) (((char *) this) + sizeof(struct io_info));
			remaining -= sizeof(struct io_info);
		}
		break;
	} while (1);
	
	this = list_entry(ioinfo_free.next, struct io_info, list);
	list_move_tail(&this->list, &ioinfo_busy);
	spin_unlock_irqrestore(&ioinfo_lists_lock, flags);
	return this;
}

/*
 * Finishes all IO and frees all IO info struct pages
 */
static void finish_all_io(void)
{
	struct io_info * this, * next = NULL;
	unsigned long flags;

	while (!list_empty(&ioinfo_busy))
		SYNC_IO;

	do { } while (!IS_ERR(ioinfo_cleanup_one()));

	/* Two stages, to avoid races. First free all io_info structs on
	 * a page except the first */
	spin_lock_irqsave(&ioinfo_lists_lock, flags);
	list_for_each_entry_safe(this, next, &ioinfo_free, list) {
		if (((unsigned long) this) & ~PAGE_MASK)
			list_del(&this->list);
	}

	/* Now we have only one reference to each page, and can safely
	 * free pages, knowing we're not going to be trying to access the
	 * same page after freeing it. */
	list_for_each_entry_safe(this, next, &ioinfo_free, list) {
		list_del(&this->list);
		free_pages((unsigned long) this, 0);
	}
	spin_unlock_irqrestore(&ioinfo_lists_lock, flags);
}

static void wait_on_one_page(struct io_info * io_info)
{
	do {
		SYNC_IO;
	} while (ioinfo_cleanup_one() != io_info);
}

void __inline reset_io_stats(void)
{
	max_outstanding_io = outstanding_io = 0;
	maxinfopages = infopages = 0;
}

void check_io_stats(void)
{
	if (outstanding_io)
		printnolog(SUSPEND_IO, SUSPEND_MEDIUM, 0, "Outstanding_io after writing is %d.\n", outstanding_io);
	printnolog(SUSPEND_IO, SUSPEND_LOW, 0, "Maximum outstanding_io was %d.\n", max_outstanding_io);
	if (infopages)
		printnolog(SUSPEND_IO, SUSPEND_MEDIUM, 0, "Info pages is %d.\n", infopages);
	printnolog(SUSPEND_IO, SUSPEND_LOW, 0, "Max info pages was %d.\n", maxinfopages);
}

struct submit_params {
	swp_entry_t swap_address;
	struct page * page;
	DEVICE_BLOCK_TYPE dev;
	long blocks[PAGE_SIZE/512];
	int blocks_used;
	int readahead_index;
};

/*
 * ---------------------------------------------------------------
 *
 *     Current state.
 *
 * ---------------------------------------------------------------
 */

/* Which pagedir are we saving/reloading? Needed so we can know whether to
 * remember the last swap entry used at the end of writing pageset2, and
 * get that location when saving or reloading pageset1.*/
static int current_stream = 0;

/* Pointer to current swap entry being loaded/saved. */
static struct range * currentblockrange = NULL;
static unsigned long currentblockoffset = 0;
static int currentblockchain = 0;

/* Last page added to cache (shot down so image kept consistent */
struct page * last_suspend_cache_page;

/* Header Page Information */
static int header_pageio_order = -1;
static int header_pages_allocated = 0;
static struct submit_params * header_page_info = NULL;

/*
 * ---------------------------------------------------------------
 *
 *     User Specified Parameters
 *
 * ---------------------------------------------------------------
 */

static int resume_firstblock = 0;
static int resume_firstblocksize = PAGE_SIZE;
static DEVICE_ID_TYPE resume_device = 0;
static DEVICE_BLOCK_TYPE resume_block_device = DEVICE_BLOCK_NONE;

/*
 * ---------------------------------------------------------------
 *
 *     Disk I/O routines
 *
 * ---------------------------------------------------------------
 */
extern char swapfilename[];
struct page * last_suspend_cache_page = NULL;

extern int max_async_ios;
extern int __nosavedata pages_written;
extern int expected_compression;

struct sysinfo swapinfo;

#define REAL_MAX_ASYNC ((max_async_ios ? max_async_ios : 32))

#define MARK_SWAP_SUSPEND 0
#define MARK_SWAP_RESUME 1

/* swap_entry_to_range_val & range_val_to_swap_entry: 
 * We are putting offset in the low bits so consecutive swap entries
 * make consecutive range values */

// ---------------------------------------------------------------------------
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,99)
// ---------------------------------------------------------------------------

extern dev_t name_to_dev_t(char *line) __init;
#define swap_entry_to_range_val(swp_entry) (swp_entry.val)
#define range_val_to_swap_entry(val) (swp_entry_t) { (val) }

// ---------------------------------------------------------------------------

/**
 *	Using bio to read from swap.
 *	This code requires a bit more work than just using buffer heads
 *	but, it is the recommended way for 2.5/2.6.
 */

static int swsusp_end_bio(struct bio * bio, unsigned int num, int err)
{
	struct io_info *io_info = (struct io_info *) bio->bi_private;
	unsigned long flags;
	
	spin_lock_irqsave(&ioinfo_lists_lock, flags);
	list_move_tail(&io_info->list, &ioinfo_ready_for_cleanup);
	spin_unlock_irqrestore(&ioinfo_lists_lock, flags);
	return 0;
}

/**
 *	submit - submit BIO request.
 *	@rw:	READ or WRITE.
 *	@io_info: IO info structure.
 *
 *	Straight from the textbook - allocate and initialize the bio.
 *	If we're writing, make sure the page is marked as dirty.
 *	Then submit it and carry on.
 */

static int submit(int rw, struct io_info * io_info)
{
	int error = 0;
	struct bio * bio = NULL;

	while (!bio) {
		bio = bio_alloc(GFP_ATOMIC,1);
		if (!bio) {
			blk_run_queues();
			io_schedule();
			ioinfo_cleanup_one();
		}
	}

	bio->bi_sector = io_info->blocks[0] << (PAGE_SHIFT - 9);
	bio->bi_bdev = io_info->dev;
	bio->bi_private = io_info;
	bio->bi_end_io = swsusp_end_bio;
	io_info->sys_struct = bio;

	if (bio_add_page(bio, io_info->buffer_page, PAGE_SIZE, 0) < PAGE_SIZE) {
		printk("ERROR: adding page to bio at %ld\n", io_info->blocks[0]);
		bio_put(bio);
		return -EFAULT;
	}

	if (rw == WRITE)
		bio_set_pages_dirty(bio);
	submit_bio(rw,bio);
	return error;
}

static inline int swsusp_set_block_size(struct block_device * bdev, int size)
{
	return set_blocksize(bdev, size);
}

static inline int swsusp_get_block_size(struct block_device * bdev)
{
	return SWAP_BLOCKSIZE(bdev);
}

// ---------------------------------------------------------------------------
#else /* 2.4 version */
// ---------------------------------------------------------------------------

extern kdev_t name_to_kdev_t(char *line) __init;
#define swap_entry_to_range_val(swp_entry) ((swp_entry.val >> 8) | ((swp_entry.val & 0x3f) << 24))
#define range_val_to_swap_entry(val) (swp_entry_t) { ((val >> 24) | ((val & 0xffffff) << 8)) }

static int swsusp_writepage(struct page *page);

static inline int sync_page(struct page *page)
{
	struct address_space *mapping = page->mapping;

	if (mapping && mapping->a_ops && mapping->a_ops->sync_page)
		return mapping->a_ops->sync_page(page);
	return 0;
}

static struct address_space_operations swsusp_aops = {
	writepage: swsusp_writepage,
	sync_page: block_sync_page,
};

struct address_space swsusp_space = {
	LIST_HEAD_INIT(swsusp_space.clean_pages),
	LIST_HEAD_INIT(swsusp_space.dirty_pages),
	LIST_HEAD_INIT(swsusp_space.locked_pages),
	0,				/* nrpages	*/
	&swsusp_aops,
};

/* 
 * Nearly the fs/buffer.c version, but we want to mark the page as done in
 * our own structures too.
 */
extern void end_buffer_io_async(struct buffer_head * bh, int uptodate);

static void swsusp_end_buffer_io_async(struct buffer_head * bh, int uptodate)
{
	struct page *page = bh->b_page;
	struct io_info * io_info = (struct io_info *) page->index;
	unsigned long flags;

	end_buffer_io_async(bh, uptodate);

	if (Page_Uptodate(page)) {
		spin_lock_irqsave(&ioinfo_lists_lock, flags);
		list_move_tail(&io_info->list, &ioinfo_ready_for_cleanup);
		spin_unlock_irqrestore(&ioinfo_lists_lock, flags);
	}

	return;
}

int swsusp_brw_page(int rw, struct page *page, DEVICE_BLOCK_TYPE dev, int b[], int size)
{
	struct buffer_head *head, *bh;

	if (!PageLocked(page))
		panic("swsusp_brw_page: page not locked for I/O");

	if (!page->buffers)
 		create_empty_buffers(page, dev, size);

	head = bh = page->buffers;

	do {
		lock_buffer(bh);
		bh->b_blocknr = *(b++);
		set_bit(BH_Mapped, &bh->b_state);
		if (rw == WRITE) {
			set_bit(BH_Uptodate, &bh->b_state);
			__mark_buffer_dirty(bh);
		} else {
			clear_bit(BH_Uptodate, &bh->b_state);
			atomic_set_buffer_clean(bh);
			__mark_buffer_clean(bh);
		}
		bh->b_flushtime = jiffies + 1;
		bh->b_end_io = swsusp_end_buffer_io_async;
		mark_buffer_async(bh, 1);
		bh = bh->b_this_page;
	} while (bh != head);

	do {
		struct buffer_head *next = bh->b_this_page;
		submit_bh(rw, bh);
		bh = next;
	} while (bh != head);

	return 0;
}

static inline int swsusp_get_block_size(dev_t bdev)
{
	int blksize;

	if (!blksize_size[MAJOR(bdev)])
		blksize = 1024;
	else
		blksize = blksize_size[MAJOR(bdev)][MINOR(bdev)];

	if (!blksize) {
		printk(name_suspend "%x: Blocksize not set?", bdev);
		blksize = PAGE_SIZE;
	}

	return blksize;
}

static inline int swsusp_set_block_size(dev_t bdev, int size)
{
	printk(name_suspend "Attempting to set blocksize for %x to %d.\n", resume_device, resume_firstblocksize);
	return set_blocksize(bdev, size);
}

static int swsusp_writepage(struct page *page)
{
	struct io_info * io_info = (struct io_info *) page->index;
	
	printlog(SUSPEND_IO, SUSPEND_MEDIUM, "Writepage %p.", page);
	if (remove_exclusive_swap_page(page)) {
		printlog(SUSPEND_IO, SUSPEND_VERBOSE, "Unlocking page %p. (1)", page);
		UnlockPage(page);
		return 0;
	}
 	swsusp_brw_page(WRITE, page, io_info->dev, io_info->blocks, io_info->block_size);
	return 0;
}

static void sync_swap_partitions(void)
{
	int i;
	
	swap_list_lock();
	for(i=0; i<MAX_SWAPFILES; i++)
		if (swap_info[i].flags)
			sync_buffers(swap_info[i].swap_device, 2);
	swap_list_unlock();
}

// ---------------------------------------------------------------------------
#endif
// ---------------------------------------------------------------------------

extern void free_single_cache_page(struct page * page);

static int get_phys_params(swp_entry_t entry)
{
	int swapfilenum = SWP_TYPE(entry);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,99)
	unsigned long offset = swp_offset(entry);
	struct swap_info_struct * sis = get_swap_info_struct(swapfilenum);
	sector_t sector = map_swap_page(sis, offset);

	printnolog(SUSPEND_BMAP,
		SUSPEND_VERBOSE,
		1,
		"Add swap partition entry to chain: Swap address %lx, chain %d, bdev %x, block %d (sector %x).",
		entry.val,
		swapfilenum,
		header_data.swapdevs[swapfilenum],
		offset,
		sector);
	add_to_range_chain(&header_data.block_chain[swapfilenum], sector);
#else
	unsigned long offset;
	int blocks[PAGE_SIZE/512];
	DEVICE_BLOCK_TYPE dev = 0;
	int block_size;
	struct inode *swapf = 0;
	get_swaphandle_info(entry, &offset, &dev, &swapf);
	if (dev) {
		/* We are assuming this.
		blocks_used = 1;
		block_size = PAGE_SIZE;
		*/
		printnolog(SUSPEND_BMAP,
			SUSPEND_VERBOSE,
			0,
			"Add swap partition entry to chain: Swap address %lx, chain %d, bdev %x, block %d.\n",
			entry.val,
			swapfilenum,
			header_data.swapdevs[swapfilenum],
			offset);
		add_to_range_chain(&header_data.block_chain[swapfilenum], offset);
	} else if (swapf) {
		int i, j;
		unsigned int block_num = offset
			<< (PAGE_SHIFT - swapf->i_sb->s_blocksize_bits);
		block_size = swapf->i_sb->s_blocksize;
		for (i=0, j=0; j< PAGE_SIZE ; i++, j += block_size)
			if (!(blocks[i] = bmap(swapf,block_num++))) {
				abort_suspend("get_phys_params: bad swap file");
				return 0;
			} else {
				printnolog(SUSPEND_BMAP,
					SUSPEND_VERBOSE,
					0,
					"Add swapfile to chain: Swap address %lx, chain %d, bdev %x, block %d/%ld-> block %d.\n",
					entry.val,
					swapfilenum,
					header_data.swapdevs[swapfilenum],
					i+1,
					PAGE_SIZE / block_size,
					blocks[i]);
				add_to_range_chain(&header_data.block_chain[swapfilenum], blocks[i]);
			}
	} else {
		printk("Warning! Trying to get invalid physical parameters! (Entry %lx).\n", entry.val);
		return 0;
	}
#endif
	return 1;
}

static int get_header_params(struct submit_params * headerpage)
{
	swp_entry_t entry = headerpage->swap_address;
	int swapfilenum = SWP_TYPE(entry);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,99)
	unsigned long offset = SWP_OFFSET(entry);
	struct swap_info_struct * sis = get_swap_info_struct(swapfilenum);
	sector_t sector = map_swap_page(sis, offset);

	headerpage->dev = sis->bdev,
	headerpage->blocks[0] = sector;
	headerpage->blocks_used = 1;
	headerpage->readahead_index = 0;
	printnolog(SUSPEND_BMAP, SUSPEND_VERBOSE, 1,
		"Header entry: %lx -> %x:%x.",
		headerpage->swap_address.val,
		headerpage->dev->bd_dev,
		headerpage->blocks[0]);
	/* We are assuming this.
	blocks_used = 1;
	block_size = PAGE_SIZE;
	*/
#else
	unsigned long offset;
	int blocks[PAGE_SIZE/512];
	DEVICE_BLOCK_TYPE dev = 0;
	int block_size;
	struct inode *swapf = 0;
	int i, j;
	get_swaphandle_info(headerpage->swap_address, &offset, &dev, &swapf);
	if (dev) {
		headerpage->dev = dev;
		headerpage->blocks[0] = offset;
		headerpage->blocks_used = 1;
		printnolog(SUSPEND_BMAP, SUSPEND_VERBOSE, 1,
			"Header entry: %lx -> %x:%x.",
			headerpage->swap_address.val,
			headerpage->dev,
			headerpage->blocks[0]);
		/* We are assuming this.
		blocks_used = 1;
		block_size = PAGE_SIZE;
		*/
	} else if (swapf) {
		unsigned int block_num = offset
			<< (PAGE_SHIFT - swapf->i_sb->s_blocksize_bits);
		block_size = swapf->i_sb->s_blocksize;
		headerpage->dev = SWAP_DEVICE_BDEV(swapfilenum);
		//printk("Swapfilenum is %d -> %d.\n", swapfilenum, headerpage->dev);
		headerpage->blocks_used = PAGE_SIZE / block_size;
		for (i=0, j=0; j< PAGE_SIZE ; i++, j += block_size)
			if (!(blocks[i] = bmap(swapf,block_num++))) {
				abort_suspend("get_header_params: bad swap file");
				return -EFAULT;
			} else {
				headerpage->blocks[i] = blocks[i];
				printnolog(SUSPEND_BMAP, SUSPEND_VERBOSE, 1,
					"Header entry: %lx -> %x:%x (%d/%d).",
					headerpage->swap_address.val,
					headerpage->dev,
					blocks[i],
					i+1,
					headerpage->blocks_used);
			}
	} else {
		printk("Warning! Trying to get invalid header params! (Entry %lx).\n", headerpage->swap_address.val);
		return -EFAULT;
	}
#endif
	return 0;
}

static struct io_info * start_one(int rw, struct submit_params * submit_info)
{
	struct io_info * io_info = get_io_info_struct();
	unsigned long buffer_virt;
	char * to, * from;
	struct page * buffer_page;
	int i;

	printlog(SUSPEND_IO, SUSPEND_MEDIUM, "Start_IO: [%p]", io_info);
	while (!(buffer_virt = get_zeroed_page(GFP_ATOMIC))) {
		RUN_IO;
		ioinfo_cleanup_one();
	}
	printlog(SUSPEND_IO, SUSPEND_HIGH, "[ALLOC BUFFER]->%d", nr_free_pages());
	buffer_page = virt_to_page(buffer_virt);
	
	io_info->data_page = submit_info->page;
	io_info->buffer_page = buffer_page;
	io_info->readahead_index = submit_info->readahead_index;

	if (rw == WRITE) {
		set_bit(IO_WRITING, &io_info->flags);

		to = (char *) buffer_virt;
		from = kmap(io_info->data_page);
		memcpy(to, from, PAGE_SIZE);
		flush_dcache_page(io_info->data_page);
		flush_dcache_page(buffer_page);
		kunmap(io_info->data_page);
	}

	atomic_inc(&buffer_page->count);
	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	lock_page(buffer_page);

	BUG_ON(buffer_page->mapping);
	buffer_page->mapping = &swsusp_space;
	buffer_page->index = (unsigned long) io_info;
#endif

	io_info->dev = submit_info->dev;
	for (i = 0; i < submit_info->blocks_used; i++)
		io_info->blocks[i] = submit_info->blocks[i];
	
	io_info->blocks_used = submit_info->blocks_used;
	io_info->block_size = PAGE_SIZE / submit_info->blocks_used;

	printlog(SUSPEND_IO, SUSPEND_HIGH, "-> (PRE BRW) %d\n", nr_free_pages());

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
 	swsusp_brw_page(rw, buffer_page, io_info->dev, io_info->blocks, io_info->block_size);
	if (PageError(buffer_page)) {
		TryLockPage(buffer_page); /* Ensure it's locked */

		if (!block_flushpage(buffer_page, 0))
			PAGE_BUG(buffer_page);
		return NULL;
	}
	io_info->sys_struct = buffer_page->buffers;
#else
 	submit(rw, io_info);
#endif
	
	io_info->cache_page = last_suspend_cache_page;
	last_suspend_cache_page = NULL;
	
	if (rw == READ)
		set_bit(IO_AWAITING_READ, &io_info->flags);
	else
		set_bit(IO_AWAITING_WRITE, &io_info->flags);
	outstanding_io++;
	if (outstanding_io > max_outstanding_io)
		max_outstanding_io++;
	return io_info;
}

static void cleanup_one(struct io_info * io_info)
{
	struct page * buffer_page;
	struct page * data_page;
	char *buffer_address, *data_address;
	int reading;

	buffer_page = io_info->buffer_page;
	data_page = io_info->data_page;

	if (test_and_set_bit(IO_CLEANUP_IN_PROGRESS, &io_info->flags))
		return;

	reading = test_bit(IO_AWAITING_READ, &io_info->flags);
	printlog(SUSPEND_IO, SUSPEND_MEDIUM, "Cleanup IO: [%p]\n", 
		io_info);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	TryLockPage(buffer_page); /* Ensure it's locked */

	if (!block_flushpage(buffer_page, 0))
		PAGE_BUG(buffer_page);
	
	sync_page(buffer_page);

	buffer_page->mapping = NULL;
	buffer_page->index = 0;
	UnlockPage(buffer_page);
#endif

	if (reading) {
		data_address = (char *) kmap(data_page);
		buffer_address = (char *) kmap(buffer_page);
		memcpy(data_address, buffer_address, PAGE_SIZE);
		flush_dcache_page(data_page);
		kunmap(data_page);
		kunmap(buffer_page);
	
	}

	if (atomic_read(&buffer_page->count) != 2)
		printk(KERN_EMERG "Cleanup IO: Page count is %d. Not good!\n", atomic_read(&buffer_page->count));
	atomic_dec(&buffer_page->count);
	__free_pages(buffer_page, 0);
	
	if (reading) {
		free_single_cache_page(io_info->cache_page);
		io_info->cache_page = NULL;
	}

	outstanding_io--;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,99)
	bio_put(io_info->sys_struct);
#endif
	io_info->sys_struct = NULL;
	io_info->flags = 0;
}

static inline void cleanup_completed_io(void)
{
	/* We clean up all we can at once so we don't only batch the first lot */
	do { } while (!ioinfo_cleanup_one());
}

static struct io_info * do_swsusp_io(int rw, struct submit_params * submit_info, int syncio)
{
	struct io_info * io_info = start_one(rw, submit_info);
	if (!io_info) {
		/* Wait on pages until no more */
		wait_on_one_page(ERR_PTR(-ENODATA));
	} else if (syncio) {
		//finish_all_io();
		wait_on_one_page(io_info);
	}
	return io_info;
} 

/*
 *
 */

int parse_signature(char * header, int restore)
{
	int type = -1;
	
	if (!memcmp("SWAP-SPACE",header,10))
		return 0;
	else if (!memcmp("SWAPSPACE2",header,10))
		return 1;

	else if (!memcmp("pmdisk", header,6))
		type = 2;
	
	else if (!memcmp("S1SUSP",header,6))
		type = 4;
	else if (!memcmp("S2SUSP",header,6))
		type = 5;
	
	else if (!memcmp("1R",header,2))
		type = 6;
	else if (!memcmp("2R",header,2))
		type = 7;
	
	else if (!memcmp("std",header,3))
		type = 8;
	else if (!memcmp("STD",header,3))
		type = 9;
	
	else if (!memcmp("sd",header,2))
		type = 10;
	else if (!memcmp("SD",header,2))
		type = 11;
	
	else if (!memcmp("z",header,1))
		type = 12;
	else if (!memcmp("Z",header,1))
		type = 13;
	
	/* Put bdev of suspend header in last byte of swap header (unsigned short) */
	if (type > 11) {
		DEVICE_ID_TYPE * header_ptr = (DEVICE_ID_TYPE *) &header[1];
		unsigned char * headerblocksize_ptr = (unsigned char *) &header[5];
		unsigned long * headerblock_ptr = (unsigned long *) &header[6];
		header_device = *header_ptr;
		headerblocksize = 512 * ((int) *headerblocksize_ptr);
		headerblock = *headerblock_ptr;
	}

	if ((restore) && (type > 5)) {
		/* We only reset our own signatures */
		if (type & 1)
			memcpy(header,"SWAPSPACE2",10);
		else
			memcpy(header,"SWAP-SPACE",10);
	}

	return type;
}

/*
 * prepare_signature
 */

static int prepare_signature(struct submit_params * header_page_info, char * current_header)
{
	int current_type = parse_signature(current_header, 0);
	DEVICE_ID_TYPE * header_ptr = (DEVICE_ID_TYPE *) (&current_header[1]);
	unsigned char * headerblocksize_ptr = (unsigned char *) (&current_header[5]);
	unsigned long * headerblock_ptr = (unsigned long *) (&current_header[6]);

	if ((current_type > 1) && (current_type < 6))
		return 1;

	if (current_type & 1)
		current_header[0] = 'Z';
	else
		current_header[0] = 'z';
	*header_ptr = BDEV_TO_DEVICE_ID(header_page_info->dev);
	*headerblocksize_ptr = (unsigned char) (PAGE_SIZE / 512 / header_page_info->blocks_used);
	*headerblock_ptr = (unsigned long) header_page_info->blocks[0]; /* prev is the first/last swap page of the resume area */
	printnolog(SUSPEND_IO, SUSPEND_HIGH, 0, "Saving header block size of %ld (%ld 512 byte blocks per page).\n",
			PAGE_SIZE / header_page_info->blocks_used,
			PAGE_SIZE / 512 / header_page_info->blocks_used);
	return 0;
}

/* We used to use bread here, but it doesn't correctly handle
 * blocksize != PAGE_SIZE. Now we create a submit_info to get the data we
 * want and use our normal routines (synchronously).
 */

static void bdev_page_io(int rw, DEVICE_BLOCK_TYPE bdev, long pos, struct page * page)
{
	struct submit_params submit_info;

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,99)
	submit_info.page = page;
	submit_info.dev = bdev;

	submit_info.blocks[0] = pos;
	submit_info.blocks_used = 1;
#else
	int block_size;
	int i, j;
	submit_info.page = page;
	submit_info.dev = bdev;

	block_size = blksize_size[MAJOR(bdev)][MINOR(bdev)];
	for (i=0, j=0; j< PAGE_SIZE ; i++, j += block_size)
		submit_info.blocks[i] = pos+i;
	submit_info.blocks_used = i;
#endif
	submit_info.readahead_index = 0;
	do_swsusp_io(rw, &submit_info, 1);
}

extern int signature_check(char * header, int fix);

static int free_swap_pages_for_header(void)
{
	int i;

	if (!header_page_info)
		return 1;

	PRINTFREEMEM("at start of free_swap_pages_for_header");

	for (i=1; i<=header_pages_allocated; i++)
		swap_free(header_page_info[i-1].swap_address);
	free_pages((unsigned long) header_page_info, header_pageio_order);
	printnolog(SUSPEND_SWAP, SUSPEND_LOW, 1, " Freed %d swap pages in free_swap_pages_for_header.\n", 
			header_pages_allocated);
	header_page_info = NULL;
	header_pages_allocated = 0;
	header_pageio_order = -1;
	PRINTFREEMEM("at end of free_swap_pages_for_header");
	return 0;
}

static void get_main_pool_phys_params(void)
{
	struct range * rangepointer = NULL;
	unsigned long address;
	int i;
	
	for (i = 0; i < MAX_SWAPFILES; i++)
		if (header_data.block_chain[i].first)
			put_range_chain(&header_data.block_chain[i]);

	range_for_each(&header_data.swapranges, rangepointer, address)
		get_phys_params(range_val_to_swap_entry(address));
}

extern void put_range(struct range * range);

static unsigned long swapwriter_storage_allocated(void)
{
	return (header_data.swapranges.size + header_pages_allocated);
}

static long swapwriter_storage_available(void)
{
	manage_swapfile(1);
	si_swapinfo(&swapinfo);
	return (swapinfo.freeswap + (long) swapwriter_storage_allocated());
}

static int swapwriter_release_storage(void)
{
	int i = 0, swapcount = 0;

#ifdef CONFIG_SOFTWARE_SUSPEND_KEEP_IMAGE
	if ((TEST_ACTION_STATE(SUSPEND_KEEP_IMAGE)) && (now_resuming))
		return 0;
#endif

	free_swap_pages_for_header();
	
	if (header_data.swapranges.first) {
		/* Free swap entries */
		struct range * rangepointer;
		unsigned long rangevalue;
		swp_entry_t entry;
		//printk("Freeing swap ranges chain.\n");
		range_for_each(&header_data.swapranges, rangepointer, rangevalue) {
			entry = range_val_to_swap_entry(rangevalue);
			printnolog(SUSPEND_SWAP, SUSPEND_VERBOSE, 1, "(Free %d %lx)", i,
				entry.val);
			swap_free(entry);
			swapcount++;
			check_shift_keys(0, NULL);
		}
		put_range_chain(&header_data.swapranges);
		for (i = 0; i < MAX_SWAPFILES; i++)
			if (header_data.block_chain[i].first)
				put_range_chain(&header_data.block_chain[i]);
	}
	
	printlog(SUSPEND_SWAP, SUSPEND_MEDIUM, "Freed %d swap pages in free_swap.\n", swapcount);

	manage_swapfile(0);

	return 0;
}

static long swapwriter_allocate_header_space(unsigned long space_requested)
{
	int numpages = space_requested; /* This was going to be in bytes... not yet */
	int i, new_order;

	PRINTFREEMEM("at start of allocate_header_space");
	
	/* num_range_pages + 1 because we also need a submit_info for saving the header */
	new_order = get_bitmask_order((((sizeof(struct submit_params) * numpages) + PAGE_SIZE - 1) / PAGE_SIZE) - 1);
	if (new_order != header_pageio_order) {
		if (header_pageio_order > -1)
			free_swap_pages_for_header();

		printnolog(SUSPEND_SWAP, SUSPEND_MEDIUM, 0,
			" Seeking order %d for %d pages.\n", 
			new_order, numpages);
		header_page_info = (struct submit_params *) get_grabbed_pages(new_order);
	
		if (!header_page_info) {
			printk("Unable to allocate page(s) for swap header info structures.\n");
			return -ENOSPC;
		}

		header_pageio_order = new_order;
	}

	for (i=(header_pages_allocated+1); i<=numpages; i++) {
		header_page_info[i-1].swap_address = get_swap_page();
		if ((!header_page_info[i-1].swap_address.val) && (header_data.swapranges.first)) {
			/*
			 *  Steal one from pageset swap chain. If, as a result,
			 *  it is too small, more swap will be allocated or
			 *  memory eaten.
			 */

			printk("Stealing a page from main swap pool. Pool size going from %d ",
				header_data.swapranges.size);
			
			header_page_info[i-1].swap_address = range_val_to_swap_entry(header_data.swapranges.first->minimum);
			if (header_data.swapranges.first->minimum < header_data.swapranges.first->maximum)
				header_data.swapranges.first->minimum++;
			else {
				struct range * oldfirst = header_data.swapranges.first;
				header_data.swapranges.first = oldfirst->next;
				header_data.swapranges.frees++;
				header_data.swapranges.lastaccessed = NULL;
				if (header_data.swapranges.last == oldfirst)
					header_data.swapranges.last = NULL;
				put_range(oldfirst);
			}
			
			header_data.swapranges.size--;
			printk("to %d.\n", header_data.swapranges.size);

			//Recalculate block chains for main pool.
			//We don't assume blocks are at start of a chain and
			//don't know how many blocks per swap entry.
			get_main_pool_phys_params();
		}
		if (!header_page_info[i-1].swap_address.val) {
			free_swap_pages_for_header();
			printk("Unable to allocate swap page for header.\n");
			return -ENOMEM;
		}
		if (get_header_params(&header_page_info[i-1]))
			return -EFAULT;
		printnolog(SUSPEND_SWAP, SUSPEND_MEDIUM, 0,
			" Got header page %d/%d. Dev is %x. Block is %lu. Blocksperpage is %d.\n",
			i+1, numpages+1,
			header_page_info[i-1].dev,
			header_page_info[i-1].blocks[0],
			header_page_info[i-1].blocks_used);
	}
	header_pages_allocated = numpages;
	printnolog(SUSPEND_SWAP, SUSPEND_LOW, 1, " Have %d swap pages in swapwriter::allocate_header_space.\n", 
			header_pages_allocated);
	PRINTFREEMEM("at end of swapwriter::allocate_header_space");
	return 0;
}

static int swapwriter_allocate_storage(unsigned long space_requested)
{
	int i, swapcount = 0, result = 0;
	int lastsize = header_data.swapranges.size;
	int numwanted = (int) (space_requested);
	int pages_to_get = numwanted - header_data.swapranges.size;
	
	if (numwanted < 1)
		return 0;

	printnolog(SUSPEND_SWAP, SUSPEND_MEDIUM, 0,
		"Started with swapranges.size == %d. Seeking to allocate %d more.\n",
		header_data.swapranges.size,
		pages_to_get);

	for(i=0; i < pages_to_get; i++) {
		swp_entry_t entry;
		printnolog(SUSPEND_SWAP, SUSPEND_VERBOSE, 1, "");
		entry = get_swap_page();
		if (!entry.val) {
			printlog(SUSPEND_SWAP, SUSPEND_MEDIUM,
				"Allocated %d/%d swap pages for main pool in allocate_swap.\n",
				swapcount, numwanted);
			printk("Unable to allocate enough swap. Got %d pages of %d wanted.\n",
				i, pages_to_get);
			result = -ENOSPC;
			goto out;
		}
		swapcount++;
		{
			int result = add_to_range_chain(&header_data.swapranges, swap_entry_to_range_val(entry));
			if (result)
				printk("add_to_range_chain returned %d.\n", result);
		}
		if (header_data.swapranges.size != (lastsize + 1))
			printk("swapranges.size == %d.\n", header_data.swapranges.size);
		lastsize = header_data.swapranges.size;
		check_shift_keys(0, NULL);
		if (TEST_RESULT_STATE(SUSPEND_ABORTED))
			break;
	}
	printlog(SUSPEND_SWAP, SUSPEND_MEDIUM,
		" Allocated %d/%d swap pages in allocate_swap.\n",
		swapcount, numwanted);
	
	printnolog(SUSPEND_SWAP, SUSPEND_MEDIUM, 0,
		"Finished with swapranges.size == %d.\n", header_data.swapranges.size);

out:
	get_main_pool_phys_params();
	
	print_chain(SUSPEND_HIGH, &header_data.swapranges, 1);
	for (i = 0; i < MAX_SWAPFILES; i++)
		if (header_data.block_chain[i].first) {
			printnolog(SUSPEND_RANGES, SUSPEND_HIGH, 0, "\nBlock chain %d:", i);
			print_chain(SUSPEND_HIGH, &header_data.block_chain[i], 1);
		}

	return result;
}

static char * swapwriter_buffer = NULL;
static int swapwriter_buffer_posn = 0;
static int current_page_number = 0;
static unsigned long * header_link = NULL;
#define BYTES_PER_HEADER_PAGE (PAGE_SIZE - sizeof(swp_entry_t))

static int swapwriter_write_header_chunk(char * buffer, int buffer_size);

static int swapwriter_write_header_init(void)
{
	int i;

	for (i = 0; i < MAX_SWAPFILES; i++)
		if (!UNUSED_SWAP_ENTRY(i)) {
			header_data.swapdevs[i] = SWAP_DEVICE_ID(i);
			header_data.blocksizes[i] = SWAP_BLOCKSIZE(SWAP_DEVICE_BDEV(i));
		}

	header_data.max_async_ios = max_async_ios;

	last_suspend_cache_page = NULL;
	swapwriter_buffer = (char *) get_zeroed_page(GFP_ATOMIC);
	header_link = (unsigned long *) (swapwriter_buffer + BYTES_PER_HEADER_PAGE);
	current_page_number = 1;
	
	/* Info needed to bootstrap goes at the start of the header.
	 * First we save the 'header_data' struct, including the number
	 * of header pages. Then we save the structs containing data needed
	 * for reading the header pages back.
	 * Note that even if header pages take more than one page, when we
	 * read back the info, we will have restored the location of the
	 * next header page by the time we go to use it.
	 */
	swapwriter_write_header_chunk((char *) &header_data, sizeof(header_data));

	return 0;
}

static int swapwriter_write_header_chunk(char * buffer, int buffer_size)
{
	int bytes_left = buffer_size;
	
	/* 
	 * We buffer the writes until a page is full and to use the last
	 * sizeof(swp_entry_t) bytes for links between pages. This is 
	 * totally transparent to the caller.
	 *
	 * Note also that buffer_size can be > PAGE_SIZE.
	 */
	printnolog(SUSPEND_IO, SUSPEND_MEDIUM, 0,
		"\nStart of write_header_chunk loop with %d bytes to store.\n",
		buffer_size);

	while (bytes_left) {
		char * source_start = buffer + buffer_size - bytes_left;
		char * dest_start = swapwriter_buffer + swapwriter_buffer_posn;
		int dest_capacity = BYTES_PER_HEADER_PAGE - swapwriter_buffer_posn;
		swp_entry_t next_header_page;
		if (bytes_left <= dest_capacity) {
			printnolog(SUSPEND_IO, SUSPEND_MEDIUM, 0,
				"Storing %d bytes from %p-%p in page %d, %p-%p.\n",
				bytes_left,
				source_start, source_start + bytes_left - 1,
				current_page_number,
				dest_start, dest_start + bytes_left - 1);
			memcpy(dest_start, source_start, bytes_left);
			swapwriter_buffer_posn += bytes_left;
			return 0;
		}
	
		/* A page is full */
		printnolog(SUSPEND_IO, SUSPEND_MEDIUM, 0,
			"Storing %d bytes from %p-%p in page %d, %p-%p.\n",
			dest_capacity,
			source_start, source_start + dest_capacity - 1,
			current_page_number,
			dest_start, dest_start + dest_capacity - 1);
		memcpy(dest_start, source_start, dest_capacity);
		bytes_left -= dest_capacity;

		next_header_page = SWP_ENTRY(SWP_TYPE(header_page_info[current_page_number].swap_address),
				header_page_info[current_page_number].blocks[0]);

		*header_link = next_header_page.val;
		
		printnolog(SUSPEND_IO, SUSPEND_MEDIUM, 0,
			"Header link is at %p. Contents set to swap device #%ld, block %ld.\n",
			header_link,
			(long) SWP_TYPE(next_header_page), SWP_OFFSET(next_header_page));

		printnolog(SUSPEND_IO, SUSPEND_MEDIUM, 0,
			"Writing header page %d/%d. Dev is %x. Block is %lu. Blocksperpage is %d.\n",
			current_page_number, header_pages_allocated,
			BDEV_TO_DEVICE_ID(header_page_info[current_page_number-1].dev),
			header_page_info[current_page_number-1].blocks[0],
			header_page_info[current_page_number-1].blocks_used);
		
		header_page_info[current_page_number-1].page = virt_to_page(swapwriter_buffer);
		do_swsusp_io(WRITE, &header_page_info[current_page_number-1], 0);

		swapwriter_buffer_posn = 0;
		current_page_number++;
	}

	return 0;
}

static int swapwriter_write_header_cleanup(void)
{
	/* Write any unsaved data */
	if (swapwriter_buffer_posn) {
		swp_entry_t next_header_page;

		next_header_page = SWP_ENTRY(SWP_TYPE(header_page_info[current_page_number].swap_address),
				header_page_info[current_page_number].blocks[0]);

		*header_link = next_header_page.val;

		printnolog(SUSPEND_IO, SUSPEND_MEDIUM, 0,
			"Header link is at %p. Contents set to swap device #%ld, block %ld.\n",
			header_link,
			(long) SWP_TYPE(next_header_page), SWP_OFFSET(next_header_page));

		printlog(SUSPEND_IO, SUSPEND_MEDIUM,
			"Writing header page %d/%d. Dev is %x. Block is %lu. Blocksperpage is %d.\n",
			current_page_number, header_pages_allocated,
			BDEV_TO_DEVICE_ID(header_page_info[current_page_number-1].dev),
			header_page_info[current_page_number-1].blocks[0],
			header_page_info[current_page_number-1].blocks_used);
		
		header_page_info[current_page_number-1].page = virt_to_page(swapwriter_buffer);
		do_swsusp_io(WRITE, &header_page_info[current_page_number-1], 0);
		
		swapwriter_buffer_posn = 0;
		current_page_number++;
	}

	/* Adjust swap header */
	bdev_page_io(READ, resume_block_device, resume_firstblock, virt_to_page(swapwriter_buffer));

	prepare_signature(&header_page_info[0],
		((union swap_header *) swapwriter_buffer)->magic.magic);
		
	bdev_page_io(WRITE, resume_block_device, resume_firstblock, virt_to_page(swapwriter_buffer));

	free_pages((unsigned long) swapwriter_buffer, 0);
	swapwriter_buffer = NULL;
	header_link = NULL;
	
	finish_all_io();

	return 0;
}

/* ------------------------- HEADER READING ------------------------- */

/*
 * read_header_init()
 * 
 * Description:
 * 1. Attempt to read the device specified with resume2=.
 * 2. Check the contents of the swap header for our signature.
 * 3. Warn, ignore, reset and/or continue as appropriate.
 * 4. If continuing, read the swapwriter configuration section
 *    of the header and set up block device info so we can read
 *    the rest of the header & image.
 *
 * Returns:
 * May not return if user choose to reboot at a warning.
 * -EINVAL if cannot resume at this time. Booting should continue
 * normally.
 */
static int swapwriter_invalidate_image(void);

static int swapwriter_read_header_init(void)
{
	int i;
	
	current_page_number = 1;

	swapwriter_buffer = (char *) get_zeroed_page(GFP_ATOMIC);

	if (!header_device) {
		printk("read_header_init called when we haven't verified there is an image!\n");
		return -EINVAL;
	}

	/* 
	 * If the header is not on the resume_device, get the resume device first.
	 */
	if (header_device != resume_device) {
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,99)
		header_block_device = open_by_devnum(header_device, FMODE_READ, BDEV_RAW);
	
		if (IS_ERR(header_block_device)) {
			if (sanity_check_failed("Failed to get access to the resume header device.\nYou could be booting with a 2.4 kernel when you suspended a 2.6 kernel or vice versa."))
				swapwriter_invalidate_image();

			return -EINVAL;
		}
#else
		int blksize = 0;
		header_block_device = header_device;

		if (!blksize_size[MAJOR(header_block_device)]) {
			printk(name_suspend "%x: Blocksize not known?\n", header_block_device);
		} else blksize = blksize_size[MAJOR(header_block_device)][MINOR(header_block_device)];
		if (!blksize) {
			printk(name_suspend "%x: Blocksize not set?\n", header_block_device);
			blksize = PAGE_SIZE;
		}
		printnolog(SUSPEND_IO, SUSPEND_MEDIUM, 0,
			"Header blocksize was %d.\n", blksize);
		if (set_blocksize(header_block_device, headerblocksize)) {
			if (sanity_check_failed("Failed to get access to the resume header device.\nYou could be booting with a 2.4 kernel when you suspended a 2.6 kernel or vice versa."))
				swapwriter_invalidate_image();

			return -EINVAL;
		}
#endif
	} else
		header_block_device = resume_block_device;

	/* Read swapwriter configuration */
	bdev_page_io(READ, header_block_device, headerblock, virt_to_page((unsigned long) swapwriter_buffer));
	//FIXME Remember location of next page to be read.
	
	printnolog(SUSPEND_IO, SUSPEND_MEDIUM, 0,
		"Retrieving %d bytes from %x:%x to page %d, %p-%p.\n",
		BDEV_TO_DEVICE_ID(header_block_device), headerblock,
		sizeof(header_data),
		current_page_number,
		swapwriter_buffer, swapwriter_buffer + sizeof(header_data) - 1);
	memcpy(&header_data, swapwriter_buffer, sizeof(header_data));
	
	/* Restore device info */
	for (i = 0; i < MAX_SWAPFILES; i++) {
		DEVICE_ID_TYPE thisdevice = header_data.swapdevs[i];
		
		printnolog(SUSPEND_IO, SUSPEND_VERBOSE, 1, "Swap device %d is %x.", i, thisdevice);
		
		if (!thisdevice)
			continue;

		if (thisdevice == resume_device) {
			printnolog(SUSPEND_IO, SUSPEND_VERBOSE, 0, "Resume root device %x", thisdevice);
			RESUME_BDEV(i) = resume_block_device;
			continue;
		}

		if (thisdevice == header_device) {
			printnolog(SUSPEND_IO, SUSPEND_VERBOSE, 0, "Resume header device %x", thisdevice);
			RESUME_BDEV(i) = header_block_device;
			continue;
		}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,99)
		RESUME_BDEV(i) = open_by_devnum(thisdevice, FMODE_READ, BDEV_RAW);
#else
		{
			int blksize = 512 * (int) header_data.blocksizes[i];
			swap_info[i].swap_device = thisdevice;
		
			printnolog(SUSPEND_IO, SUSPEND_VERBOSE, 0, "Resume secondary device %x", thisdevice);
			printnolog(SUSPEND_IO, SUSPEND_HIGH, 0, "Blocksize was %d.\n", blksize);
			set_blocksize(thisdevice, blksize);
		}
#endif
	}

	max_async_ios =	header_data.max_async_ios;

	swapwriter_buffer_posn = sizeof(header_data);

	return 0;
}

static int swapwriter_read_header_chunk(char * buffer, int buffer_size)
{
	int bytes_left = buffer_size, ret = 0;
	
	/* Read a chunk of the header */
	while ((bytes_left) && (!ret)) {
		swp_entry_t next = ((union p_diskpage) swapwriter_buffer).pointer->link.next;
		DEVICE_BLOCK_TYPE dev = RESUME_BDEV(SWP_TYPE(next));
		int pos = SWP_OFFSET(next);
		char * dest_start = buffer + buffer_size - bytes_left;
		char * source_start = swapwriter_buffer + swapwriter_buffer_posn;
		int source_capacity = BYTES_PER_HEADER_PAGE - swapwriter_buffer_posn;

		if (bytes_left <= source_capacity) {
			printnolog(SUSPEND_IO, SUSPEND_HIGH, 0,
				"Retrieving %d bytes from page %d, %p-%p into %p-%p.\n",
				bytes_left,
				current_page_number,
				source_start, source_start + bytes_left - 1,
				dest_start, dest_start + bytes_left - 1);
			memcpy(dest_start, source_start, bytes_left);
			swapwriter_buffer_posn += bytes_left;
			return buffer_size;
		}

		/* Next to read the next page */
		printnolog(SUSPEND_IO, SUSPEND_HIGH, 0,
			"Retrieving %d bytes from page %d, %p-%p to %p-%p.\n",
			source_capacity,
			current_page_number,
			source_start, source_start + source_capacity - 1,
			dest_start, dest_start + source_capacity - 1);
		memcpy(dest_start, source_start, source_capacity);
		bytes_left -= source_capacity;

		printnolog(SUSPEND_IO, SUSPEND_HIGH, 0,
			"Header link is at %p. Contents set to %lx = swap device #%x, block %d.\n",
			&((union p_diskpage) swapwriter_buffer).pointer->link.next,
			((union p_diskpage) swapwriter_buffer).pointer->link.next.val,
			BDEV_TO_DEVICE_ID(dev), pos);

		current_page_number++;

		printnolog(SUSPEND_IO, SUSPEND_MEDIUM, 0,
			"Reading header page %d. Dev is %x. Block is %lu.\n",
			current_page_number, dev, pos);
		
		bdev_page_io(READ, dev, pos, virt_to_page(swapwriter_buffer));

		swapwriter_buffer_posn = 0;
	}

	return buffer_size - bytes_left;
}

static int swapwriter_read_header_cleanup(void)
{
	free_pages((unsigned long) swapwriter_buffer, 0);
	return 0;
}

static int swapwriter_prepare_save_ranges(void)
{
	int i;

	print_chain(SUSPEND_HIGH, &header_data.swapranges, 0);
	for (i = 0; i < MAX_SWAPFILES; i++)
		print_chain(SUSPEND_HIGH, &header_data.block_chain[i], 0);
	
	relativise_chain(&header_data.swapranges);
	
	for (i = 0; i < MAX_SWAPFILES; i++)
		relativise_chain(&header_data.block_chain[i]);
	
	header_data.pd1start_block_range = RANGE_RELATIVE(header_data.pd1start_block_range);

	printlog(SUSPEND_IO, SUSPEND_MEDIUM,
		"Pagedir1 firstblockrange is %p.\n", header_data.pd1start_block_range);

	return 0;
}

static int swapwriter_post_load_ranges(void)
{
	int i;
	unsigned long * range_pages = get_rangepages_list();
	
	if (!range_pages) {
		printk("Unable to allocate rangepageslist.\n");
		return -ENOMEM;
	}
	
	absolutise_chain(&header_data.swapranges, range_pages);
	print_chain(SUSPEND_HIGH, &header_data.swapranges, 0);

	for (i = 0; i < MAX_SWAPFILES; i++) {
		absolutise_chain(&header_data.block_chain[i], range_pages);
		print_chain(SUSPEND_HIGH, &header_data.block_chain[i], 0);
	}

	header_data.pd1start_block_range = RANGE_ABSOLUTE(header_data.pd1start_block_range, range_pages);
	
	put_rangepages_list(range_pages);

	return 0;
}

static int swapwriter_write_init(int stream_number)
{
	last_suspend_cache_page = NULL;
	
	if (stream_number == 1) {
		currentblockrange = header_data.pd1start_block_range;
		currentblockoffset = header_data.pd1start_block_offset;
		currentblockchain = header_data.pd1start_chain;
		printnolog(SUSPEND_IO, SUSPEND_MEDIUM, 0,
			"Beginning from position: range %p, block %ld.\n",
			currentblockrange, currentblockoffset);
	} else {
		for (currentblockchain = 0; currentblockchain < MAX_SWAPFILES; currentblockchain++)
			if (header_data.block_chain[currentblockchain].first) {
				currentblockrange = header_data.block_chain[currentblockchain].first;
				currentblockoffset = currentblockrange->minimum;
				break;
			}
	}
	
	swapwriter_buffer_posn = 0;

	current_page_number = 1;
	current_stream = stream_number;

	reset_io_stats();

	return 0;
}

static int swapwriter_write_chunk(char * buffer)
{
	int i;
	struct submit_params submit_params;
	int blksize = 0, num_blocks = 1;

	if (currentblockchain == MAX_SWAPFILES) {
		int current_log_all;
		printk("Error! We have run out of blocks for writing data.\n");
		for (i = 0; i < MAX_SWAPFILES; i++) {
			if (UNUSED_SWAP_ENTRY(i))
				printk("Swap slot %d is unused.\n", i);
			else
				printk("Swap slot %d is device %x.\n", i, SWAP_DEVICE_ID(i));
			if (header_data.block_chain[i].size)
				printk("Chain size for device %d is %d.\n", i,
					header_data.block_chain[i].size);
		}
		printk("Swap range chain looks like this:\n");
		current_log_all = TEST_ACTION_STATE(SUSPEND_LOGALL);
		SET_ACTION_STATE(SUSPEND_LOGALL);
		print_chain(0, &header_data.swapranges, 0);
		if (!current_log_all)
			CLEAR_ACTION_STATE(SUSPEND_LOGALL);
		return -ENOSPC;
	}
	blksize = swsusp_get_block_size(SWAP_DEVICE_BDEV(currentblockchain));
	num_blocks = PAGE_SIZE / blksize;

	printlog(SUSPEND_IO, SUSPEND_MEDIUM,
		"Writing page %d. Dev is %x. Block is %lu. Blocksperpage is %d.\n",
		current_page_number,
		SWAP_DEVICE_BDEV(currentblockchain),
		currentblockoffset,
		num_blocks);
		
	submit_params.page = virt_to_page(buffer);
	submit_params.dev = SWAP_DEVICE_BDEV(currentblockchain);
		
	/* Get the blocks */
	for (i = 0; i < num_blocks; i++) {
		if (!currentblockrange) {
			do {
				currentblockchain++;
			} while ((currentblockchain < MAX_SWAPFILES) &&
				 (!header_data.block_chain[currentblockchain].first));

			/* We can validly not have a new blockrange. We
			 * might be compressing data and the user was
			 * too optimistic in setting the compression
			 * ratio or we're just copying the pageset. */

			if (currentblockchain == MAX_SWAPFILES)
				return -ENOSPC;
				
			currentblockrange = header_data.block_chain[currentblockchain].first;
			currentblockoffset = currentblockrange->minimum;
		}
		submit_params.blocks[i] = currentblockoffset;
		GET_RANGE_NEXT(currentblockrange, currentblockoffset);
	}

	submit_params.blocks_used = num_blocks;
	submit_params.readahead_index = 0;
		
	printnolog(SUSPEND_PAGESETS, SUSPEND_VERBOSE, 1,
		"page:%d. bdev:%x. blocks (%d):",
		current_page_number,
		submit_params.dev,
		submit_params.blocks_used);

	for (i = 0; i < submit_params.blocks_used; i++)
		printnolog(SUSPEND_PAGESETS, SUSPEND_VERBOSE, 0,
			"0x%lx%s",
			submit_params.blocks[i],
			((i+1) < submit_params.blocks_used) ? "," : "\n");

	check_shift_keys(0, NULL);

	do_swsusp_io(WRITE, &submit_params, 0);

	swapwriter_buffer_posn = 0;
	current_page_number++;

	return 0;
}

static int swapwriter_write_cleanup(void)
{
	//printk("Cleanup: Stream number is %d.\n", current_stream);
	if (current_stream == 2) {
		header_data.pd1start_block_range = currentblockrange;
		header_data.pd1start_block_offset = currentblockoffset;
		header_data.pd1start_chain = currentblockchain;
		//printk("Saving final position: range %p, block %ld.\n",
		//	currentblockrange, currentblockoffset);
	}
	
	finish_all_io();
	
	check_io_stats();

	return 0;
}

static int swapwriter_read_init(int stream_number)
{
	int i;
	unsigned long flags;

	if (stream_number == 1) {
		currentblockrange = header_data.pd1start_block_range;
		currentblockoffset = header_data.pd1start_block_offset;
		currentblockchain = header_data.pd1start_chain;
		//printk("Starting from final range %p, block %ld.\n",
		//	currentblockrange, currentblockoffset);
	} else {
		for (i = 0; i < MAX_SWAPFILES; i++)
			if (header_data.block_chain[i].first) {
				currentblockrange = header_data.block_chain[i].first;
				currentblockoffset = currentblockrange->minimum;
				currentblockchain = i;
				break;
			}
	}
	
	last_suspend_cache_page = NULL;
	
	current_page_number = 1;

	reset_io_stats();

	readahead_index = 0;
	spin_lock_irqsave(&readahead_flags_lock, flags);
	readahead_flags = 0;
	spin_unlock_irqrestore(&readahead_flags_lock, flags);
	for (i = 1; i <= MAX_READAHEAD; i++)
		readahead_page[i-1] = virt_to_page(get_zeroed_page(GFP_ATOMIC));

	return 0;
}

static int swapwriter_begin_read_chunk(struct page * page, int readahead_index, int sync)
{
	int i;
	struct submit_params submit_params;
	int blksize = 0, num_blocks = 1;
	struct io_info * io_info;

	blksize = swsusp_get_block_size(SWAP_DEVICE_BDEV(currentblockchain));
	num_blocks = PAGE_SIZE / blksize;

	printlog(SUSPEND_IO, SUSPEND_MEDIUM,
		"Reading page %d. Dev is %x. Block is %lu. Blocksperpage is %d. Page is %p. Readahead index is %d.",
		current_page_number,
		SWAP_DEVICE_BDEV(currentblockchain),
		currentblockoffset,
		num_blocks, page, readahead_index);
		
	submit_params.page = page;
	submit_params.dev = SWAP_DEVICE_BDEV(currentblockchain);
	submit_params.readahead_index = readahead_index;
		
	/* Get the blocks */
	for (i = 0; i < num_blocks; i++) {
		submit_params.blocks[i] = currentblockoffset;
		GET_RANGE_NEXT(currentblockrange, currentblockoffset);
		if (!currentblockrange) {
			do {
				currentblockchain++;
			} while ((!header_data.block_chain[currentblockchain].first) && 
					(i < MAX_SWAPFILES));

			/* We can validly not have a new blockrange. We
			 * might be compressing data and the user was
			 * too optimistic in setting the compression
			 * ratio or we're just copying the pageset. */

			if (i == MAX_SWAPFILES)
				return -ENOSPC;
				
			currentblockrange = header_data.block_chain[currentblockchain].first;
			currentblockoffset = currentblockrange->minimum;
		}
	}

	submit_params.blocks_used = num_blocks;
		
	printnolog(SUSPEND_PAGESETS, SUSPEND_VERBOSE, 1,
		"page:%d. bdev:%x. blocks (%d):",
		current_page_number,
		submit_params.dev,
		submit_params.blocks_used);

	for (i = 0; i < submit_params.blocks_used; i++)
		printnolog(SUSPEND_PAGESETS, SUSPEND_VERBOSE, 0,
			"0x%lx%s",
			submit_params.blocks[i],
			((i+1) < submit_params.blocks_used) ? "," : "\n");

	check_shift_keys(0, NULL);

	io_info = do_swsusp_io(READ, &submit_params, sync);

	printlog(SUSPEND_IO, SUSPEND_MEDIUM,
		"IO_info struct is at %p.\n", io_info);

	swapwriter_buffer_posn = 0;
	current_page_number++;

	return 0;
}

static int swapwriter_read_chunk(char * buffer, int sync)
{
	static int last_result;
	int i;
	unsigned long flags;

	if ((!sync) || (MAX_READAHEAD == 0))
		return swapwriter_begin_read_chunk(virt_to_page(buffer), 0, sync);

	/* Start the first one, plus read-ahead */
	if (readahead_index == 0) {
		for (i = 1; i <= MAX_READAHEAD; i++) {
			if ((last_result = swapwriter_begin_read_chunk(readahead_page[i-1], i, 0))) {
				printk("Begin read chunk for page %d returned %d.\n",
					i, last_result);
				return last_result;
			}
		}
		readahead_index++;
	}
	
	/* read_ahead_index is the one we want to return */
	while (!test_bit(readahead_index - 1, &readahead_flags)) {
		RUN_IO;
		ioinfo_cleanup_one();
	}

	memcpy(buffer, page_address(readahead_page[readahead_index - 1]), PAGE_SIZE);

	/* Start a new readahead? */
	if (!last_result) {
		spin_lock_irqsave(&readahead_flags_lock, flags);
		clear_bit(readahead_index - 1, &readahead_flags);
		spin_unlock_irqrestore(&readahead_flags_lock, flags);
		last_result = swapwriter_begin_read_chunk(readahead_page[readahead_index - 1], readahead_index, 0);
		if (last_result)
			printk("Begin read chunk for page %d returned %d.\n",
				readahead_index, last_result);
	}

	readahead_index++;
	if (readahead_index > MAX_READAHEAD)
		readahead_index = 1;

	return 0;
}

static int swapwriter_read_cleanup(void)
{
	int i;

	finish_all_io();
	check_io_stats();
	for (i = 0; i < MAX_READAHEAD; i++)
		__free_pages(readahead_page[i], 0);
	return 0;
}

extern int nr_suspends;

/* swapwriter_invalidate_image
 * 
 */
static int swapwriter_invalidate_image(void)
{
	union p_diskpage cur;
	int result = 0;
	char newsig[11];
	
	cur.address = get_zeroed_page(GFP_ATOMIC);
	if (!cur.address) {
		printk("Unable to allocate a page for restoring the swap signature.\n");
		return -ENOMEM;
	}

	/*
	 * If nr_suspends == 0, we must be booting, so no swap pages
	 * will be recorded as used yet.
	 */

	if (nr_suspends > 0)
		swapwriter_release_storage();

	/* 
	 * We don't do a sanity check here: we want to restore the swap 
	 * whatever version of kernel made the suspend image.
	 * 
	 * We need to write swap, but swap may not be enabled so
	 * we write the device directly
	 */
	
	bdev_page_io(READ, resume_block_device, resume_firstblock, virt_to_page(cur.pointer));

	result = parse_signature(cur.pointer->swh.magic.magic, 1);
		
	if (result < 4)
		goto out;

	strncpy(newsig, cur.pointer->swh.magic.magic, 10);
	newsig[10] = 0;
	printlog(SUSPEND_ANY_SECTION, SUSPEND_VERBOSE,"Swap signature will be set to %s.\n", newsig);

	bdev_page_io(WRITE, resume_block_device, resume_firstblock, virt_to_page(cur.pointer));

	if (!nr_suspends)
		printk(KERN_WARNING name_swsusp "Image invalidated.\n");
out:
	free_pages(cur.address, 0);
	return 0;
}

/*
 * workspace_size
 *
 * Description:
 * Returns the number of bytes of RAM needed for this
 * code to do its work. (Used when calculating whether
 * we have enough memory to be able to suspend & resume).
 *
 */
static unsigned long swapwriter_memory_needed(void)
{
	return (MAX(REAL_MAX_ASYNC, MAX_READAHEAD)  * (PAGE_SIZE + sizeof(struct io_info)));
}

/* Print debug info
 *
 * Description:
 */

static int swapwriter_print_debug_stats(char * buffer, int size)
{
	int len = 0, already_on = 0;
	struct sysinfo sysinfo;
	
	if (active_writer != &swapwriterops) {
		len = suspend_snprintf(buffer, size, "- Swapwriter inactive.\n");
		return len;
	}

	len = suspend_snprintf(buffer, size, "- Swapwriter active.\n");
	if (swapfilename)
		len+= suspend_snprintf(buffer+len, size-len,
			"  Attempting to automatically swapon: %s.\n", swapfilename);

	already_on = manage_swapfile(1);
	
	si_swapinfo(&sysinfo);
	
	len+= suspend_snprintf(buffer+len, size-len, "  Swap available for image: %ld.\n",
			sysinfo.freeswap);

	if (!already_on)
		manage_swapfile(0);
	
	return len;
}

/*
 * Storage needed
 *
 * Returns amount of space in the swap header required
 * for the swapwriter's data.
 *
 */
static unsigned long swapwriter_storage_needed(void)
{
	/* FIXME We need to add up space for dev info etc... */
	return sizeof(header_data);
}

/*
 * Wait on I/O
 *
 */

static int swapwriter_wait_on_io(int flush_all)
{
	if (flush_all)
		finish_all_io();
	else
		cleanup_completed_io();

	return 0;
}

/*
 * Image_exists
 *
 */

static int swapwriter_image_exists(void)
{
	int signature_found;
	union p_diskpage diskpage;
	
	if (!resume_device) {
		printk("Not even trying to read header because resume_device is not set.\n");
		return 0;
	}
	
	//PRINTFREEMEM("at start of swapwriter_image_exists.");

	last_suspend_cache_page = NULL;

	diskpage.address = get_zeroed_page(GFP_ATOMIC);

	/* FIXME: Make sure bdev_page_io handles wrong parameters */
	bdev_page_io(READ, resume_block_device, resume_firstblock, virt_to_page(diskpage.ptr));
	signature_found = parse_signature(diskpage.pointer->swh.magic.magic, 0);

	if (signature_found < 2) {
		printk(KERN_ERR name_suspend "This is normal swap space.\n" );
		return 0;	/* non fatal error */
	} else if (signature_found == -1) {
		printk(KERN_ERR name_suspend "Unable to find a signature. Could you have moved a swap file?\n");
		return 0;
	} else if (signature_found < 6) {
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
		if ((!(software_suspend_state & SOFTWARE_SUSPEND_NORESUME_SPECIFIED)) 
				&& sanity_check_failed("Detected the signature of an alternate implementation.\n"))
			software_suspend_state |= SOFTWARE_SUSPEND_NORESUME_SPECIFIED;
#else
		printk(KERN_ERR name_suspend "Detected the signature of an alternate implementation.\n");
#endif
		return 0;
	} else if ((signature_found >> 1) != SIGNATURE_VER) {
		if ((!(software_suspend_state & SOFTWARE_SUSPEND_NORESUME_SPECIFIED)) &&
				sanity_check_failed("Found a different style suspend image signature."))
			software_suspend_state |= SOFTWARE_SUSPEND_NORESUME_SPECIFIED;
	}

	return 1;
}

/*
 * Parse Image Location
 *
 * Attempt to parse a resume2= parameter.
 * Swap Writer accepts:
 * resume2=swap:DEVNAME[:FIRSTBLOCK][@BLOCKSIZE]
 *
 * Where:
 * DEVNAME is convertable to a dev_t by name_to_dev_t
 * FIRSTBLOCK is the location of the first block in the swap file
 * (specifying for a swap partition is nonsensical but not prohibited).
 * BLOCKSIZE is the logical blocksize >= 512 & <= PAGE_SIZE, 
 * mod 512 == 0 of the device.
 * Data is validated by attempting to read a swap header from the
 * location given. Failure will result in swapwriter refusing to
 * save an image, and a reboot with correct parameters will be
 * necessary.
 */

static int swapwriter_parse_image_location(char * commandline, int boot_time)
{
	char *thischar, *colon = NULL, *at_symbol = NULL;
	union p_diskpage diskpage;
	int signature_found;

	if (strncmp(commandline, "swap:", 5)) {
		printk(name_suspend "Swapwriter: Image location doesn't begin with 'swap:'");
		return 0;
	}

	commandline += 5;

	thischar = commandline;
	while ((*thischar != ':') && ((thischar - commandline) < 250) && (*thischar))
		thischar++;

	if (*thischar == ':') {
		colon = thischar;
		*colon = 0;
		thischar++;
	}

	while ((*thischar != '@') && ((thischar - commandline) < 250) && (*thischar))
		thischar++;

	if (*thischar == '@') {
		at_symbol = thischar;
		*at_symbol = 0;
	}
	
	if (colon) {
		resume_firstblock = (int) simple_strtoul(colon + 1, NULL, 0);
		printk("Looking for first block of swap header at block %x.\n", resume_firstblock);
	}

	if (at_symbol) {
		resume_firstblocksize = (int) simple_strtoul(at_symbol + 1, NULL, 0);
		if (resume_firstblocksize & 0x1FF)
			printk("Blocksizes are usually a multiple of 512. Don't expect this to work!\n");
		printk("Setting logical block size of resume device to %d.\n", resume_firstblocksize);
	}
	
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,99)
	resume_device = name_to_dev_t(commandline);
	if (!resume_device) {
		if (boot_time)
			sanity_check_failed("Failed to get the location of the device on which Software Suspend's header should be found.");
		else
			printk("Failed to get the location of the device on which Software Suspend's header should be found.");
		goto invalid;
	}
	
	resume_block_device = open_by_devnum(resume_device, FMODE_READ, BDEV_RAW);

	if (IS_ERR(resume_block_device)) {
		if (boot_time)
			sanity_check_failed("Failed to get access to the device on which Software Suspend's header should be found.");
		else
			printk("Failed to get access to the device on which Software Suspend's header should be found.");
		goto invalid;
	}
#else
	resume_block_device = resume_device = name_to_kdev_t(commandline);

	if (!resume_block_device) {
		if (boot_time)
			sanity_check_failed("Failed to get the location of the device on which Software Suspend's header should be found.");
		else
			printk("Failed to get the location of the device on which Software Suspend's header should be found.");
		goto invalid;
	}

#endif

	if (colon)
		*colon = ':';
	if (at_symbol)
		*at_symbol = '@';

	if ((swsusp_get_block_size(resume_block_device) != resume_firstblocksize) &&
	    (swsusp_set_block_size(resume_block_device, resume_firstblocksize) == -EINVAL))
		goto invalid;

	diskpage.address = get_zeroed_page(GFP_ATOMIC);
	bdev_page_io(READ, resume_block_device, resume_firstblock, virt_to_page(diskpage.ptr));
	signature_found = parse_signature(diskpage.pointer->swh.magic.magic, 0);
	free_page((unsigned long) diskpage.address);

	if (signature_found != -1) {
		printk(KERN_ERR name_suspend "Swap space signature found.\n" );
		return 1;
	}

	printk(KERN_ERR name_suspend "Sorry. No swap signature found at specified location.\n");
	return -EINVAL;

invalid:
	if (colon)
		*colon = ':';
	if (at_symbol)
		*at_symbol = '@';
	printk(KERN_ERR name_suspend "Sorry. Location looks invalid.\n");
	return -EINVAL;
}

static int swapfilename_read_proc(char * page, char ** start, off_t off, int count,
		int *eof, void *data)
{
	int len = 0;
	
	if (swapfilename)
		len = sprintf(page, "%s\n", swapfilename);
	*eof = 1;
	return len;
}

static int swapfilename_write_proc(struct file *file, const char * buffer,
		unsigned long count, void * data)
{
	if (count > 255)
		count = 255;

	if (copy_from_user(swapfilename, buffer, count))
		return -EFAULT;
	
	if ((count) && (swapfilename[count - 1] == '\n'))
		swapfilename[count - 1] = 0;
	
	swapfilename[count] = 0;

	return count;
}

int header_locations_read_proc(char * page, char ** start, off_t off, int count,
		int *eof, void *data)
{
	int i, printedpartitionsmessage = 0, len = 0, haveswap = 0, device_block_size;
	struct inode *swapf = 0;
	int zone;
	char * path_page = (char *) __get_free_page(GFP_KERNEL);
	char * path;
	int path_len;
	
	*eof = 1;
	if (!page)
		return 0;

	for (i = 0; i < MAX_SWAPFILES; i++) {
		if (UNUSED_SWAP_ENTRY(i))
			continue;
		
		if (SWAP_IS_PARTITION(i)) {
			haveswap = 1;
			if (!printedpartitionsmessage) {
				len += sprintf(page + len, 
					"For swap partitions like /dev/hd%d%d, simply use resume2=swap:/dev/hd%d%d.\n",
					((SWAP_DEVICE_ID(i) >> 4) & 15) + 96,
					SWAP_DEVICE_ID(i) & 15,
					((SWAP_DEVICE_ID(i) >> 4) & 15) + 96,
					SWAP_DEVICE_ID(i) & 15);
				printedpartitionsmessage = 1;
			}
		} else {
			path_len = 0;
			
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,99)
			path = d_path(	swap_info[i].swap_file->f_dentry,
					swap_info[i].swap_file->f_vfsmnt,
					path_page,
					PAGE_SIZE);
#else
			path = d_path(	swap_info[i].swap_file,
					swap_info[i].swap_vfsmnt,
					path_page,
					PAGE_SIZE);
#endif
			path_len = sprintf(path_page, "%-31s ", path);
			
			haveswap = 1;
			swapf = SWAP_FILE_INODE(i);
			device_block_size = SWAP_BLOCKSIZE(SWAP_DEVICE_BDEV(i));
			if (!(zone = bmap(swapf,0))) {
				len+= sprintf(page + len, 
					"Swapfile %-31s has been corrupted. Reuse mkswap on it and try again.\n",
					path_page);
			} else {
				len+= sprintf(page + len, "For swapfile `%-31s`, use resume2=swap:/dev/hda%d:0x%x",
						path_page,
						SWAP_DEVICE_ID(i) & 15,
						zone);
				if (device_block_size == PAGE_SIZE)
					len+= sprintf(page + len, " If this doesn't work, try adding ");
				len+= sprintf(page + len, "@%d.\n", device_block_size);
			}

		}
	}
	
	if (!haveswap)
		len = sprintf(page, "You need to turn on swap partitions before examining this file.\n");

	free_pages((unsigned long) path_page, 0);
	return len;
}

static void swapwriter_noresume_reset(void)
{
	int i;

 	/* 
	 * If we have read part of the image, we might have filled header_data with
	 * data that should be zeroed out.
	 */

	memset((char *) &header_data, 0, sizeof(header_data));
	for (i = 0; i < MAX_SWAPFILES; i++) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
		swap_info[i].swap_device = 0;
#endif
		RESUME_BDEV(i) = DEVICE_BLOCK_NONE;
	}
	 
}
struct swsusp_proc_data swapfilename_proc_data = {
	.filename			= "swapfilename",
	.permissions			= PROC_RW,
	.type				= SWSUSP_PROC_DATA_CUSTOM,
	.data = {
		.special = {
			.read_proc	= swapfilename_read_proc,
			.write_proc	= swapfilename_write_proc,
		}
	}
};

struct swsusp_proc_data headerlocations_proc_data = {
	.filename		= "headerlocations",
	.permissions		= PROC_READONLY,
	.type			= SWSUSP_PROC_DATA_CUSTOM,
	.data = {
		.special = {
			.read_proc = header_locations_read_proc,
		}
	}
};

struct swsusp_plugin_ops swapwriterops = {
	.type					= WRITER_PLUGIN,
	.name					= "Swap Writer",
	.memory_needed				= swapwriter_memory_needed,
	.print_debug_info			= swapwriter_print_debug_stats,
	.storage_needed				= swapwriter_storage_needed,
	.write_init				= swapwriter_write_init,
	.write_chunk				= swapwriter_write_chunk,
	.write_cleanup				= swapwriter_write_cleanup,
	.read_init				= swapwriter_read_init,
	.read_chunk				= swapwriter_read_chunk,
	.read_cleanup				= swapwriter_read_cleanup,
	.noresume_reset				= swapwriter_noresume_reset,
	.ops = {
		.writer = {
			.storage_available 	= swapwriter_storage_available,
			.storage_allocated	= swapwriter_storage_allocated,
			.release_storage	= swapwriter_release_storage,
			.allocate_header_space	= swapwriter_allocate_header_space,
			.allocate_storage	= swapwriter_allocate_storage,
			.image_exists		= swapwriter_image_exists,
			.write_header_init	= swapwriter_write_header_init,
			.write_header_chunk	= swapwriter_write_header_chunk,
			.write_header_cleanup	= swapwriter_write_header_cleanup,
			.read_header_init	= swapwriter_read_header_init,
			.read_header_chunk	= swapwriter_read_header_chunk,
			.read_header_cleanup	= swapwriter_read_header_cleanup,
			.prepare_save_ranges	= swapwriter_prepare_save_ranges,
			.post_load_ranges	= swapwriter_post_load_ranges,
			.invalidate_image	= swapwriter_invalidate_image,
			.wait_on_io		= swapwriter_wait_on_io,
			.parse_image_location	= swapwriter_parse_image_location,
		}
	}
};

/* ---- Registration ---- */
static __init int swapwriter_load(void)
{
	int result;
	printk("Software Suspend Swap Writer v1.0\n");

	if (!(result = swsusp_register_plugin(&swapwriterops))) {
		swsusp_register_procfile(&swapfilename_proc_data);
		swsusp_register_procfile(&headerlocations_proc_data);
	}
	return result;
}

__initcall(swapwriter_load);

