/*
 * linux/kernel/suspend.c
 *
 * Copyright (C) 1998-2001 Gabor Kuti <seasons@fornax.hu>
 * Copyright (C) 1998,2001,2002 Pavel Machek <pavel@suse.cz>
 * Copyright (C) 2002-2003 Florent Chabaud <fchabaud@free.fr>
 * Copyright (C) 2002-2003 Nigel Cunningham <ncunningham@clear.net.nz>
 *
 * This file is released under the GPLv2.
 *
 * This file is to realize architecture-independent
 * machine suspend feature using pretty near only high-level routines
 *
 * We'd like to thank the following people for their work:
 * 
 * Pavel Machek <pavel@ucw.cz>:
 * Modifications, defectiveness pointing, being with me at the very beginning,
 * suspend to swap space, stop all tasks. Port to 2.4.18-ac and 2.5.17.
 *
 * Steve Doddi <dirk@loth.demon.co.uk>: 
 * Support the possibility of hardware state restoring.
 *
 * Raph <grey.havens@earthling.net>:
 * Support for preserving states of network devices and virtual console
 * (including X and svgatextmode)
 *
 * Kurt Garloff <garloff@suse.de>:
 * Straightened the critical function in order to prevent compilers from
 * playing tricks with local variables.
 *
 * Andreas Mohr <a.mohr@mailto.de>
 *
 * Alex Badea <vampire@go.ro>:
 * Fixed runaway init
 *
 * Jeff Snyder <je4d@pobox.com>
 * ACPI patch
 *
 * Nathan Friess <natmanz@shaw.ca>
 * Some patches.
 *
 * Michael Frank <mhf@linuxmail.org>
 * Extensive testing and help with improving stability.
 *
 * More state savers are welcome. Especially for the scsi layer...
 *
 * For TODOs,FIXMEs also look in Documentation/swsusp.txt.
 * 
 * Variable definitions which are needed if PM is enabled but 
 * SOFTWARE_SUSPEND is disabled are found near the top of process.c.
 */

#define SWSUSP_MAIN_C

#include <linux/suspend-common.h>
#include <linux/suspend-version-specific.h>
#include <asm/i387.h> /* for kernel_fpu_end */
#ifdef	CONFIG_KDB
#include <linux/kdb.h>
#endif	/* CONFIG_KDB */

#if (SWSUSP_VERSION_SPECIFIC_REVISION != SWSUSP_CORE_REVISION)
#error The core and version specific patch revisions of Software Suspend are incompatible.
#error You can find your core version from include/linux/suspend-debug.h
#error and you version specific revision from include/linux/suspend-version-specific.h.
#endif

unsigned int nr_suspends = 0;

/* Variables to be preserved over suspend */
int pageset2_sizelow = 0;
struct pagedir __nosavedata pagedir_resume;
int __nosavedata pages_written = 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
static int pm_suspend_state = 0;
#endif

struct list_head swsusp_filters, swsusp_writers, swsusp_plugins;
int num_filters = 0, num_writers = 0;
struct swsusp_plugin_ops * first_filter = NULL;
struct swsusp_plugin_ops * active_writer = NULL;

int swsusp_act_used = 0;
int swsusp_lvl_used = 0;
int swsusp_dbg_used = 0;

char resume_file[256] = "";	/* For resume= kernel option */

unsigned long orig_mem_free = 0;

static void drivers_resume(int);
extern void do_swsusp_lowlevel(int resume);
extern unsigned long header_storage_for_plugins(void);
extern int pm_prepare_console(void);
extern void pm_restore_console(void);
void empty_swsusp_memory_pool(void);
int read_primary_suspend_image(void);

unsigned long * inusemap = NULL;
unsigned long * pageset2map = NULL;
unsigned long * nosavemap = NULL;

/* Suspend pagedir is allocated before final copy, and saved with data
 * as well as separately, so it can be used to load pageset2 at resume.
 *
 * At resume, the original pagedir is loaded as pagedir_resume and then
 * moved to a place where it doesn't collide with the data to be copied
 * back. Then the data is read (again into spots that don't collide) and
 * then copied back, giving us access to the saved pagedir again and
 * forgetting the loaded pagedir at the same time. We then used the saved
 * pagedir to load pageset2 (if necessary) before freeing that pagedir.
 */

struct pagedir pagedir1 = { 1, 0, 0}, pagedir2 = {2, 0, 0};

unsigned long swsusp_debug_state = 0;

int image_size_limit = 0;
int max_async_ios = 32;
int currentbeep = 260;

/* Pagedir.c */
extern void copy_pageset1(void);
extern int allocatemap(unsigned long ** pagemap, int setnosave);
extern void free_pagedir(struct pagedir * p);
extern int freemap(unsigned long ** pagemap);

/* Prepare_image.c */

extern int prepare_image(void);

/* proc.c */

extern int swsusp_init_proc(void);
extern int swsusp_cleanup_proc(void);

/* process.c */
extern atomic_t swsusp_cpu_counter; 

/* include/asm-i386/suspend.h */
extern void smp_swsusp_lowlevel(void * info);

static void ensure_on_processor_zero(void)
{
#ifdef CONFIG_SMP
	set_cpus_allowed(current, CPU0_MASK);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	if (unlikely(cpu_number_map(smp_processor_id()) != 0))
		BUG();
#else
	BUG_ON(smp_processor_id() != 0);
#endif
#endif
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
/* Make disk drivers accept operations, again */
void ide_unsuspend(void)
{
#ifdef CONFIG_BLK_DEV_IDE
	ide_disk_unsuspend(0);
#endif
}

void ide_suspend(void)
{
#ifdef CONFIG_BLK_DEV_IDE
	do_suspend_sync();

	ide_disk_suspend();
#endif
}
#endif

#define RESUME_PHASE1 1 /* Called from interrupts disabled */
#define RESUME_PHASE2 2 /* Called with interrupts enabled */
#define RESUME_ALL_PHASES (RESUME_PHASE1 | RESUME_PHASE2)

void drivers_resume(int flags)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
  	if(flags & RESUME_PHASE2) {
#ifdef CONFIG_BLK_DEV_HD
		do_reset_hd();			/* Kill all controller state */
#endif
	}
  	if (flags & RESUME_PHASE1) {
#ifdef CONFIG_BLK_DEV_IDE
		ide_disk_unsuspend(1);		
#endif
#ifdef CONFIG_BLK_DEV_MD
		md_autostart_arrays();
#endif
	}
  	if (flags & RESUME_PHASE2) {
		if (pm_suspend_state) {
			if (pm_send_all(PM_RESUME,(void *)0))
				printk(name_suspend "Problem while sending resume event\n");
			pm_suspend_state=0;
		} else
			printk(name_suspend "PM suspend state wasn't raised\n");

#ifdef DEFAULT_SUSPEND_CONSOLE
		update_screen(fg_console);
#endif
	}
#else
	if (flags & RESUME_PHASE1)
		device_resume();
#endif
}

/* Called from process context */
static int drivers_suspend(void)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#ifdef CONFIG_BLK_DEV_MD
	md_notify_reboot(NULL, SYS_HALT, NULL);
#endif
	IDE_SUSPEND
	if (!pm_suspend_state) {
		if (pm_send_all(PM_SUSPEND,(void *)3)) {
			printk(name_suspend "Problem while sending suspend event\n");
			drivers_resume(RESUME_ALL_PHASES);
			return(1);
		}
		pm_suspend_state=1;
	}
	  
	return(0);
#else
	return device_suspend(4);
#endif
}

/*
 * save_image
 * Result code (int): Zero on success, non zero on failure.
 * Functionality    : High level routine which performs the steps necessary
 *                    to prepare and save the image after preparatory steps
 *                    have been taken.
 * Key Assumptions  : Processes frozen, sufficient memory available, drivers
 *                    suspended.
 * Called from      : do_swsusp2_suspend_2
 */
extern struct pageset_sizes_result recalculate_stats(void);
extern void display_stats(void);
extern int write_pageset(struct pagedir * pagedir, int whichtowrite);
extern int write_image_header(void);
extern int read_secondary_pagedir(int overwrittenpagesonly);

static int save_image(void)
{
	int temp_result;

	if (RAM_TO_SUSPEND > max_mapnr) {
		printk(KERN_CRIT name_suspend "Couldn't get enough free pages, on %ld pages short\n",
			 RAM_TO_SUSPEND - max_mapnr);
		prepare_status(1, 1, "Couldn't get enough free pages, on %ld pages short\n",
			 RAM_TO_SUSPEND - max_mapnr);
		goto abort_saving;
	}
	
	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW," - Final values: %d and %d.\n", 
		pageset1_size, 
		pageset2_size);

	pagedir1.lastpageset_size = pageset1_size;
	pagedir2.lastpageset_size = pageset2_size;

	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,"-- Preparing to write pages\n");

	check_shift_keys(1, "About to write pagedir2. ");

	temp_result = write_pageset(&pagedir2, 2);
	
	if (temp_result == -1 || TEST_RESULT_STATE(SUSPEND_ABORTED)) {
		goto abort_saving;
	}

	pages_written = temp_result;

	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,"-- Written pageset2\n");
	
	check_shift_keys(1, "About to copy pageset 1.");

	drivers_suspend();

	PRINTPREEMPTCOUNT("Prior to lowlevel suspend.");
	
	do_swsusp_lowlevel(0);

	PRINTPREEMPTCOUNT("At exit from save_image.");
	return 0;
abort_saving:
	{
		int was_locked = (spin_is_locked(&suspend_irq_lock));
		if (!was_locked)
			swsusp_spin_lock_irqsave(&suspend_irq_lock, swsuspirqflags);
	}

	PRINTPREEMPTCOUNT("At exit from save_image.");
	return -1;		
}

int save_image_part1(void)
{
	int temp_result;

	prepare_status(1, 0, "Copying pageset1.");

	PRINTPREEMPTCOUNT("Prior to copying pageset 1.");

	copy_pageset1();
	
	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,"-- Done\n");
	
	/*
	 *  ---------------------   FROM HERE ON, NEED TO REREAD PAGESET2 IF ABORTING!!! -----------------
	 *  (We need to ensure saved pages are freed even if memory is still correct).
	 */
	
	swsusp_spin_unlock_irqrestore(&suspend_irq_lock, swsuspirqflags);
	
#ifdef CONFIG_PREEMPT
	preempt_enable();
#endif

	kernel_fpu_end();

	PRINTPREEMPTCOUNT("Prior to resuming drivers.");

	drivers_resume(RESUME_PHASE1);
	
	IDE_UNSUSPEND

	/* Other processors have waited for me to make the atomic copy of the kernel */
	swsusp_state &= ~FREEZE_SMP;

	while (atomic_read(&swsusp_cpu_counter) &&
		(!TEST_RESULT_STATE(SUSPEND_ABORTED))) {
		cpu_relax();
		barrier();
		check_shift_keys(0, "");
	}

	if (TEST_RESULT_STATE(SUSPEND_ABORTED))
		goto abort_reloading_pagedir_two;

	check_shift_keys(1, "About to write pageset1. ");

	/*
	 * End of critical section.
	 */
	
	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,"-- Writing pageset1\n");

	temp_result = write_pageset(&pagedir1, 1);

	if (temp_result == -1 || TEST_RESULT_STATE(SUSPEND_ABORTED))
		goto abort_reloading_pagedir_two;

	pages_written += temp_result;

	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,"-- Done\n");

	check_shift_keys(1, "About to write header. ");

	if (TEST_RESULT_STATE(SUSPEND_ABORTED))
		goto abort_reloading_pagedir_two;

	temp_result = write_image_header();

	if (temp_result || (TEST_RESULT_STATE(SUSPEND_ABORTED)))
		goto abort_reloading_pagedir_two;

	check_shift_keys(1, "About to power down or reboot.");

	PRINTPREEMPTCOUNT("At exit from save_image.");
	return 0;

abort_reloading_pagedir_two:
	temp_result = read_secondary_pagedir(1);
	if (temp_result)
		panic("Attempt to reload pagedir 2 while aborting a suspend failed.");

	PRINTPREEMPTCOUNT("At exit from save_image.");
	return -1;		

}

/*
 * suspend_power_down
 * Functionality   : Powers down or reboots the computer once the image
 *                   has been written to disk.
 * Key Assumptions : Able to reboot/power down via code called or that
 *                   the warning emitted if the calls fail will be visible
 *                   to the user (ie printk resumes devices).
 * Called From     : do_swsusp2_suspend_2
 */

extern asmlinkage long sys_reboot(int magic1, int magic2, unsigned int cmd, void * arg);
extern void apm_power_off(void);

void suspend_power_down(void)
{
	C_A_D = 0;
	/* No delay should be needed - ide_disk_suspend purges cache now */
	beepOK;			/* last beep */
	prepare_status(1, 0, "Ready to power down.");
#if defined(CONFIG_VT) && defined(CONFIG_SOFTWARE_SUSPEND_DEBUG)
	if (TEST_ACTION_STATE(SUSPEND_REBOOT))
		sys_reboot(LINUX_REBOOT_MAGIC1, LINUX_REBOOT_MAGIC2, LINUX_REBOOT_CMD_RESTART, NULL);
#endif
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
	drivers_suspend();
#endif
	sys_reboot(LINUX_REBOOT_MAGIC1, LINUX_REBOOT_MAGIC2, LINUX_REBOOT_CMD_POWER_OFF, NULL);

	printk(KERN_EMERG name_suspend "Probably not capable for powerdown.\n");
	machine_halt();
	printk(KERN_EMERG name_suspend "System is now halted.\n");
	while (1)
		cpu_relax();
	/* NOTREACHED */
}

/*
 * do_swsusp2_resume_1
 * Functionality   : Preparatory steps for copying the original kernel back.
 * Called From     : include/asm/suspend.h:do_swsusp_lowlevel
 */

void do_swsusp2_resume_1(void)
{
	/* Get other cpus ready to restore their original contexts */

	swsusp_state |= FREEZE_SMP;

	printk("Putting other CPUs in swsusp_lowevel in preparation for restoring contexts:\n");

	smp_call_function(smp_swsusp_lowlevel, NULL, 0, 0);

	while ((atomic_read(&swsusp_cpu_counter) < (NUM_CPUS - 1)) &&
		(!TEST_RESULT_STATE(SUSPEND_ABORTED))) {
		cpu_relax();
		barrier();
		check_shift_keys(0, "");
	}

	printk("\nDone.\n");

	drivers_suspend();
	barrier();
	mb();
	swsusp_spin_lock_irqsave(&suspend_irq_lock, swsuspirqflags);	/* Done to disable interrupts */ 
	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW, name_suspend "Waiting for DMAs to settle down...\n");
	mdelay(1000);	/* We do not want some readahead with DMA to corrupt our memory, right?
			   Do it with disabled interrupts for best effect. That way, if some
			   driver scheduled DMA, we have good chance for DMA to finish ;-). */
	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW, name_suspend "About to copy pageset1 back...\n");
}

/*
 * do_swsusp2_resume_2
 * Functionality   : Steps taken after copying back the original kernel at
 *                   resume.
 * Key Assumptions : Will be able to read back secondary pagedir (if 
 *                   applicable).
 * Called From     : include/asm/suspend.h:do_swsusp_lowlevel
 */

extern void post_resume_console_redraw(void);

void do_swsusp2_resume_2(void)
{

	flush_tlb_all();

	tainted |= 4; /* Taint the kernel because we have suspended */
	now_resuming = 1;

#ifdef DEFAULT_SUSPEND_CONSOLE
	post_resume_console_redraw();
#endif

	PRINTPREEMPTCOUNT("In resume_2 after copy back.");

	kernel_fpu_end();

	PRINTPREEMPTCOUNT("In resume_2 after fpu end.");
	
#ifdef CONFIG_PREEMPT
	preempt_enable();
#endif

	PRINTPREEMPTCOUNT("After preempt enable.");
	swsusp_spin_unlock_irqrestore(&suspend_irq_lock, swsuspirqflags);
	PRINTPREEMPTCOUNT("After unlocking irq_lock.");

	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW, name_suspend "Resume drivers.\n");
	drivers_resume(RESUME_PHASE1);

	IDE_UNSUSPEND
	
	check_shift_keys(1, "About to reload secondary pagedir.\n");

	read_secondary_pagedir(0);
	
	prepare_status(0, 0, "Cleaning up...");
#ifdef CONFIG_SOFTWARE_SUSPEND_KEEP_IMAGE
	if (TEST_ACTION_STATE(SUSPEND_KEEP_IMAGE)) {
		printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW, name_suspend "Not invalidating the image due to Keep Image being enabled.\n");
		SET_RESULT_STATE(SUSPEND_KEPT_IMAGE);
	} else
#endif
	{
		printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW, name_suspend "Removing image... ");
		active_writer->ops.writer.invalidate_image();
	}
	swsusp_state &= ~USE_MEMORY_POOL;
	drivers_resume(RESUME_ALL_PHASES);
	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW, "ok\n");
}

/*
 * do_swsusp2_suspend_1
 * Functionality   : Steps taken prior to saving CPU state and the image
 *                   itself.
 * Called From     : include/asm/suspend.h:do_swsusp_lowlevel
 */

void do_swsusp2_suspend_1(void)
{
	/* Save other cpu contexts */

	swsusp_state |= FREEZE_SMP;

	printnolog(SUSPEND_FREEZER, SUSPEND_MEDIUM, 0,
		"Saving contexts:\n");

	smp_call_function(smp_swsusp_lowlevel, NULL, 0, 0);

	while ((atomic_read(&swsusp_cpu_counter) < (NUM_CPUS - 1)) &&
		(!TEST_RESULT_STATE(SUSPEND_ABORTED))) {
		cpu_relax();
		barrier();
		check_shift_keys(0, "");
	}

	printnolog(SUSPEND_FREEZER, SUSPEND_MEDIUM, 0,
		"All other processors have saved their contexts. CPU counter ->%d\n", 
		atomic_read(&swsusp_cpu_counter));

	printnolog(SUSPEND_FREEZER, SUSPEND_MEDIUM, 0,
		"\nDone.\n");
	
	mb();
	barrier();

	PRINTPREEMPTCOUNT("Before preempt disable.");
#ifdef CONFIG_PREEMPT
	preempt_disable();
#endif
	PRINTPREEMPTCOUNT("Before irq lock.");
	swsusp_spin_lock_irqsave(&suspend_irq_lock, swsuspirqflags);
	PRINTPREEMPTCOUNT("After irq lock.");
}

/*
 * do_swsusp2_suspend_2
 * Functionality   : Steps taken after saving CPU state to save the
 *                   image and powerdown/reboot or recover on failure.
 * Key Assumptions : save_image returns zero on success; otherwise we need to
 *                   clean up and exit. The state on exiting this routine 
 *                   should be essentially the same as if we have suspended,
 *                   resumed and reached the end of do_swsusp2_resume_2.
 * Called From     : include/asm/suspend.h:do_swsusp_lowlevel
 */
void do_swsusp2_suspend_2(void)
{
	if (!save_image_part1())
		suspend_power_down();

	if (!TEST_RESULT_STATE(SUSPEND_ABORT_REQUESTED))
		printk(KERN_EMERG name_suspend "Suspend failed, trying to recover...\n");
	MDELAY(1000); /* So user can wait and report us messages if armageddon comes :-) */

	barrier();
	mb();

	active_writer->ops.writer.invalidate_image();
}

static inline void lru_check_page(struct page * page)
{
	if (!PageLRU(page))
		printk("Page %p/%p in inactivelist but not marked LRU.\n",
			page, page_address(page));
}

static void lru_check(void)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
	struct zone * zone;

	for_each_zone(zone) {
		spin_lock(&zone->lru_lock);
		INACTIVE_LIST_FOR_EACH(lru_check_page);
		ACTIVE_LIST_FOR_EACH(lru_check_page);
		spin_unlock(&zone->lru_lock);
	}
#else
	spin_lock(&pagemap_lru_lock);
	INACTIVE_LIST_FOR_EACH(lru_check_page);
	ACTIVE_LIST_FOR_EACH(lru_check_page);
	spin_unlock(&pagemap_lru_lock);
#endif
}

int suspend_snprintf(char * buffer, int buffer_size, const char *fmt, ...)
{
	int result;
	va_list args;

	if (!buffer_size) {
		return 0;
	}

	va_start(args, fmt);
	result = vsnprintf(buffer, buffer_size, fmt, args);
	va_end(args);

	if (result > buffer_size) {
		return buffer_size;
	}

	return result;
}

static int print_plugin_debug_info(char * buffer, int buffer_size)
{
	struct list_head *plugin;
	struct swsusp_plugin_ops *this_plugin = NULL;
	int len = 0;

	list_for_each(plugin, &swsusp_plugins) {
		this_plugin = list_entry(plugin, struct swsusp_plugin_ops, plugin_list);
		if (this_plugin->print_debug_info) {
			int result;
			result = this_plugin->print_debug_info(buffer + len, buffer_size - len);
			len += result;
		}
	}

	return len;
}

/* get_debug_info
 * Functionality:	Store debug info in a buffer.
 * Called from:		Worker thread or via do_software_suspend.
 */

#define SNPRINTF(a...) 	len += suspend_snprintf(buffer + len, buffer_size - len - 1, ## a);

static int get_suspend_debug_info(char * buffer, int buffer_size)
{
	int len = 0;

	SNPRINTF("Please include the following information in bug reports:\n");
	SNPRINTF("- SWSUSP core    : %s\n", swsusp_version);
	SNPRINTF("- Kernel Version : %s\n", UTS_RELEASE);
	SNPRINTF("- Version spec.  : %s\n", SWSUSP_VERSION_SPECIFIC_REVISION_STRING);
	SNPRINTF("- Compiler vers. : %d.%d\n", __GNUC__, __GNUC_MINOR__);
#ifdef CONFIG_MODULES
	{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
		struct module *this_mod;
		SNPRINTF("- Modules loaded : ");
		this_mod = module_list;
		while (this_mod) {
			if (this_mod->name)
				SNPRINTF("%s ", this_mod->name);
			this_mod = this_mod->next;
		}
#else
		extern int print_module_list_to_buffer(char * buffer, int size);
		len+= print_module_list_to_buffer(buffer + len, buffer_size - len - 1);
#endif
	}
	SNPRINTF("\n");
#else
	SNPRINTF("- No module support.\n");
#endif
	SNPRINTF("- Attempt number : %d\n", nr_suspends);
	SNPRINTF("- Pageset sizes  : %d and %d (%d low).\n", 
			pagedir1.lastpageset_size, 
			pagedir2.lastpageset_size, 
			pageset2_sizelow);
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
	SNPRINTF("- Parameters     : %ld %ld %ld %d %d %d\n",
			swsusp_result,
			swsusp_action,
			swsusp_debug_state,
			swsusp_default_console_level,
			image_size_limit,
			max_async_ios);
#else
	SNPRINTF("- Parameters     : %ld %ld %d %d\n",
			swsusp_result,
			swsusp_action,
			image_size_limit,
			max_async_ios);
#endif
	SNPRINTF("- Calculations   : Image size: %lu. Ram to suspend: %ld.\n",
		STORAGE_NEEDED, RAM_TO_SUSPEND);
	SNPRINTF("- Limits         : %lu pages RAM. Initial boot: %lu.\n",
		max_mapnr, orig_mem_free);
	SNPRINTF("- Overall expected compression percentage: %d.\n", 100 - expected_compression_ratio());
	len+= print_plugin_debug_info(buffer + len, buffer_size - len - 1);
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
	SNPRINTF("- Debugging compiled in.\n");
#endif
#ifdef CONFIG_PREEMPT
	SNPRINTF("- Preemptive kernel.\n");
#endif
#ifdef CONFIG_SMP
	SNPRINTF("- SMP kernel.\n");
#endif
#ifdef CONFIG_HIGHMEM
	SNPRINTF("- Highmem Support.\n");
#endif
	SNPRINTF("- Max ranges used: %d ranges in %d pages.\n",
		max_ranges_used, num_range_pages);

	return len;
}

/*
 * display_debug_info
 * Functionality   : At the end of resuming, displays information that may be
 *                   helpful in debugging software suspend.
 * Called From     : do_software_suspend
 */
static void display_debug_info(void)
{
	int len;
	char * buffer = (char *) get_zeroed_page(GFP_ATOMIC);

	if (!buffer)
		printk("Unable to allocate a buffer for debug info!.\n");

	len = get_suspend_debug_info(buffer, PAGE_SIZE);
	printk("%s", buffer);
	free_pages((unsigned long) buffer, 0);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
static unsigned long kstat_store = 0;
/* kstat_save
 * Save the contents of the kstat array so that
 * our labours are hidden from vmstat
 */
static int kstat_save(void)
{
	const int bytes_per_page = PAGE_SIZE - sizeof(unsigned long);
	int kstat_pages = (sizeof(kstat) + bytes_per_page - 1) / bytes_per_page;
	int i, source_offset = 0, bytes_to_copy;
	unsigned long this_page, last_page = 0;

	kstat_store = this_page = get_zeroed_page(GFP_ATOMIC);
	for (i = 1; i <= kstat_pages; i++) {
		if (!this_page) {
			int k;
			this_page = kstat_store;
			for (k = 1; k < i; k++) {
				unsigned long next_page = *((unsigned long *) this_page + ((PAGE_SIZE / sizeof(unsigned long)) - 1));
				free_pages(this_page, 0);
				this_page = next_page;
			}
			kstat_store = 0;
			return -ENOMEM;
		}
		bytes_to_copy = ((sizeof(kstat) - source_offset) >= bytes_per_page) ?
			bytes_per_page : (sizeof(kstat) - source_offset);
		memcpy((char *) this_page, ((char *) &kstat) + source_offset, bytes_to_copy);
		if (i < kstat_pages) {
			last_page = this_page;
			this_page = get_zeroed_page(GFP_ATOMIC);
			*((unsigned long *) last_page + ((PAGE_SIZE / sizeof(unsigned long)) - 1)) = this_page;
			source_offset += bytes_per_page;
		}
	}
	return 0;
}

/* kstat_restore
 * Restore a previously saved kstat array
 */
static void kstat_restore(void)
{
	const int bytes_per_page = PAGE_SIZE - sizeof(unsigned long);
	int kstat_pages = (sizeof(kstat) + bytes_per_page - 1) / bytes_per_page;
	int i, source_offset = 0, bytes_to_copy;
	unsigned long this_page = kstat_store, next_page;

	if (!kstat_store)
		return;

	for (i = 1; i <= kstat_pages; i++) {
		bytes_to_copy = ((sizeof(kstat) - source_offset) >= bytes_per_page) ?
			bytes_per_page : (sizeof(kstat) - source_offset);
		memcpy(((char *) &kstat) + source_offset, (char *) this_page, bytes_to_copy);
		next_page = *((unsigned long *) this_page + ((PAGE_SIZE / sizeof(unsigned long)) - 1));
		source_offset += bytes_per_page;
		free_pages(this_page, 0);
		this_page = next_page;
	}
	kstat_store = 0;
}
#else
#define kstat_save()
#define kstat_restore()
#endif

#define SUSPEND_C
#include <asm/suspend.h>

/*
 * do_software_suspend
 * Functionality   : First level of code for software suspend invocations.
 *                   Stores and restores load averages (to avoid a spike),
 *                   allocates bitmaps, freezes processes and eats memory
 *                   as required before suspending drivers and invoking
 *                   the 'low level' code to save the state to disk.
 *                   By the time we return from do_swsusp_lowlevel, we
 *                   have either failed to save the image or successfully
 *                   suspended and reloaded the image. The difference can
 *                   be discerned by checking SUSPEND_ABORTED.
 * Called From     : 
 */
extern inline int notify_resume(void);
static void display_debug_info(void);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#ifdef CONFIG_MAGIC_SYSRQ
extern int sysrq_pressed;
#endif
#endif

static void do_software_suspend(void)
{
	unsigned long avenrun_save[3];
	int i;

	if (software_suspend_state & 3) {
		printk(name_swsusp "Software suspend is disabled or already running.\n");
		printk("This may be because you haven't put something along the lines of\n");
		printk("\nresume2=swap:/dev/hda1\n\n");
		printk("In lilo.conf or equivalent. (Where /dev/hda1 is your swap partition).\n");

		SET_RESULT_STATE(SUSPEND_ABORTED);
		return;
	}

        /* Suspend always runs on processor 0 */ 
	ensure_on_processor_zero();

#ifdef CONFIG_SOFTWARE_SUSPEND_KEEP_IMAGE
	if (TEST_RESULT_STATE(SUSPEND_KEPT_IMAGE)) {
		if (TEST_ACTION_STATE(SUSPEND_KEEP_IMAGE)) {
			printk("Image already stored: powering down immediately.");
			suspend_power_down();
			return;	/* It shouldn't, but just in case */
		} else {
			printk("Invalidating previous image.\n");
			active_writer->ops.writer.invalidate_image();
		}
	}
#endif

	printk(name_swsusp "Initiating a software_suspend cycle.\n");
	BUG_ON(in_interrupt());
	software_suspend_state |= SOFTWARE_SUSPEND_RUNNING;
	swsusp_result = 0;
	max_ranges_used = 0;
	nr_suspends++;
	now_resuming = 0;
	
	//printk("LRU check at start of cycle...\n");
	lru_check();
	//printk("Done.\n");
	
	PRINTFREEMEM("at start of do_software_suspend");

	/*
	 * Running swsusp makes for a very high load average. I'm told that
	 * sendmail and crond check the load average, so in order for them
	 * not to be unnecessarily affected by the operation of swsusp, we
	 * store the avenrun values prior to suspending and restore them
	 * at the end of the resume cycle. Thus, the operation of swsusp
	 * should be invisible to them. Thanks to Marcus Gaugusch and Bernard
	 * Blackham for noticing the problem and suggesting the solution.
	 */
	
	for (i = 0; i < 3; i++)
		avenrun_save[i] = avenrun[i];

	/* Suspend console switch (if necessary) */
	if (pm_prepare_console())
		printk(name_suspend "Can't allocate a console... proceeding\n");

	PRINTFREEMEM("after preparing suspend_console");

	beepOK;			/* first beep */

	prepare_status(1, 1, "Attempting to allocate memory for page flags.");

	printlog(SUSPEND_MEMORY, SUSPEND_VERBOSE, "Allocating inusemap\n");
	if (allocatemap(&inusemap, 1))
		goto out;
	PRINTFREEMEM("after allocating inusemap");
	
	if (allocatemap(&pageset2map, 1))
		goto out;
	PRINTFREEMEM("after allocating pageset2 map");
	
	set_chain_names(&pagedir1);
	set_chain_names(&pagedir2);

	/* Free up memory if necessary */
	printlog(SUSPEND_ANY_SECTION, SUSPEND_VERBOSE, "Preparing image.\n");
	PRINTPREEMPTCOUNT("Before preparing image.");
	if (prepare_image() || TEST_RESULT_STATE(SUSPEND_ABORTED))
		goto out;

	PRINTPREEMPTCOUNT("After prepare_image.");
	PRINTFREEMEM("after preparing image");

	if (TEST_ACTION_STATE(SUSPEND_FREEZER_TEST))
		goto out;

	beepOK;		/* second beep */

	/* We don't want suspend to show in the kernel statistics -
	 * it should be transparent to userspace */
	kstat_save();

	if (!TEST_RESULT_STATE(SUSPEND_ABORTED)) {
		prepare_status(1, 0, "Starting to save the image..");
		save_image();
		PRINTPREEMPTCOUNT("After saving the image.");
		beepOK;
	}
	
	display_debug_info();
	
	check_shift_keys(1, NULL);

out:
	PRINTFREEMEM("at 'out'");

	swsusp_state &= ~USE_MEMORY_POOL;
	IDE_UNSUSPEND

	free_pagedir(&pagedir2);
	free_pagedir(&pagedir1);
	
#ifdef CONFIG_SOFTWARE_SUSPEND_KEEP_IMAGE
	if (!TEST_ACTION_STATE(SUSPEND_KEEP_IMAGE))
#endif
		active_writer->ops.writer.invalidate_image();

	empty_swsusp_memory_pool();
	PRINTFREEMEM("after freeing memory pool");

	free_ranges();
	PRINTFREEMEM("after freeing ranges");

	freemap(&pageset2map);
	PRINTFREEMEM("after freeing pageset2 map");

	freemap(&inusemap);
	PRINTFREEMEM("after freeing inuse map");

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#ifdef CONFIG_MAGIC_SYSRQ
	sysrq_pressed = 0;
#endif
#endif
	/* Restore stats before we restart processes */
	for (i = 0; i < 3; i++)
		avenrun[i] = avenrun_save[i];
	
	kstat_restore();

	thaw_processes();

	if (notify_resume())
		printk(KERN_EMERG "Failed to notify resume chain.\n");

	MDELAY(1000);
	check_shift_keys(1, "About to restore original console. ");
	pm_restore_console();
#ifdef DEFAULT_SUSPEND_CONSOLE
	update_screen(fg_console);
#endif
	PRINTFREEMEM("at end of do_software_suspend");

	software_suspend_state &= ~SOFTWARE_SUSPEND_RUNNING;
#ifdef CONFIG_PREEMPT
	PRINTPREEMPTCOUNT("Exiting with preempt count");
#endif
	
	//printk("LRU check at end of cycle...\n");
	lru_check();
	//printk("Done.\n");
}

static DECLARE_MUTEX(kswsuspd_access);
static DECLARE_WAIT_QUEUE_HEAD(swsusp_wait);
static DECLARE_COMPLETION(work_complete);
static void * work_data = NULL;

void start_kswsuspd(void * data)
{
	down(&kswsuspd_access);
	init_completion(&work_complete);
	work_data = data;
	wake_up_interruptible(&swsusp_wait);
	wait_for_completion(&work_complete);
	up(&kswsuspd_access);
}

/*
 * This is main interface to the outside world. It needs to be
 * called from process context. It sets the active flag, 
 * notifies kswsuspd thread through the wake up and waits 
 * for ACTIVE flag to be cleared.
 */

void software_suspend_pending(void)
{
	int action = 1;

	current->flags |= PF_NOFREEZE;
	start_kswsuspd(&action);
	current->flags &= ~PF_NOFREEZE;
}

/*
 * This is kswsuspd, the kernel daemon which waits to be woken by a userspace
 * daemon (via software_suspend_pending). It signals completion (failure or
 * the end of a full suspend/resume cycle) by resetting cycle_unfinished. In
 * this way, the userspace process blocks until the cycle is complete and a
 * script doesn't need to poll the proc entry (there is now no way to tell
 * from /proc/sys/kernel/swsusp whether the cycle has completed anyway).
 */

static int get_suspend_debug_info(char * buffer, int buffer_size);

static int swsusp_mainloop(void *unused)
{
  	printk(name_swsusp "kswsuspd starting\n"); /* This will print the current version on boot */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	daemonize();
	strcpy(current->comm, "kswsuspd");
#else
        daemonize("kwsuspd");
#endif
	current->flags |= PF_NOFREEZE;

	sigfillset(&current->blocked);
	software_suspend_state &= ~SOFTWARE_SUSPEND_RUNNING;

	for(;;) {
		interruptible_sleep_on(&swsusp_wait);
		switch (*((int *) work_data)) {
			case 1:	/* Suspend */
				swsusp_result = 0;
				do_software_suspend();
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
				currentbeep = 100; /* not the original one, so that we know if this is first or subsequent suspend */
#endif
				complete(&work_complete);
				break;
			case 2:	/* Temporarily turn on swap (where applicable) for getting debugging info */
				{
					struct debug_info_data * did = (struct debug_info_data *) work_data;
					did->bytes_used = get_suspend_debug_info(did->buffer, did->buffer_size);
					complete(&work_complete);
				}
		}
	}

	swsusp_cleanup_proc();
	
	return(0);
}

int swsusp_register_plugin(struct swsusp_plugin_ops * plugin)
{
	if (!(num_filters + num_writers)) {
		INIT_LIST_HEAD(&swsusp_filters);
		INIT_LIST_HEAD(&swsusp_writers);
		INIT_LIST_HEAD(&swsusp_plugins);
	}

	switch (plugin->type) {
		case FILTER_PLUGIN:
			{
				list_add_tail(&plugin->ops.filter.filter_list, &swsusp_filters);
				if (!num_filters)
					first_filter = plugin;
				num_filters++;
			}
			break;
		case WRITER_PLUGIN:
			{
				list_add_tail(&plugin->ops.writer.writer_list, &swsusp_writers);
				num_writers++;
			}
			break;
		default:
			printk("Hmmm. Plugin '%s' has an invalid type. It has been ignored.\n", plugin->name);
			return -EINVAL;
	}
	list_add(&plugin->plugin_list, &swsusp_plugins);

	return 0;	
}

struct swsusp_plugin_ops * get_next_filter(struct swsusp_plugin_ops * filter_sought)
{
	struct list_head *filter;
	struct swsusp_plugin_ops * last_filter = NULL, *this_filter = NULL;

	list_for_each(filter, &swsusp_filters) {
		this_filter = list_entry(filter, struct swsusp_plugin_ops, ops.filter.filter_list);
		if (last_filter == filter_sought)
			return this_filter;
		last_filter = this_filter;
	}

	return active_writer;
}

int attempt_to_parse_resume_device(int boot_time)
{
	struct list_head *writer;
	struct swsusp_plugin_ops * this_writer;
	int result = 0;

	if (!num_writers) {
		printk(name_suspend "No writers have been registered.\n");
		return 0;
	}
	
	software_suspend_state &= ~SOFTWARE_SUSPEND_RESUME_DEVICE_OK;
	software_suspend_state |= SOFTWARE_SUSPEND_DISABLED;

	list_for_each(writer, &swsusp_writers) {
		this_writer = list_entry(writer, struct swsusp_plugin_ops, ops.writer.writer_list);
		result = this_writer->ops.writer.parse_image_location(resume_file, boot_time);
		switch (result) {
			case -EINVAL:
				printk(name_suspend "Not able to successfully parse this resume device.\n");
				return result;
			case 0:	
				break;
			case 1:
				active_writer = this_writer;

				/* We may not have any filters compiled in */
				if (!first_filter)
					first_filter = (struct swsusp_plugin_ops *) this_writer;
				software_suspend_state |= SOFTWARE_SUSPEND_RESUME_DEVICE_OK;
				software_suspend_state &= ~3;
		}
	}
	return 0;
}

/*
 * Called from init kernel_thread.
 * We check if we have an image and if so we try to resume.
 * We also start kswsuspd if configuration looks right.
 */

extern int freeze_processes(int no_progress);
extern int orig_loglevel;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
void software_resume2(void)
{
#define RESUME2_RET
#else
static int __init software_resume2(void)
{
#define RESUME2_RET ret
	int ret = 0;
#endif
	int read_image_result = 0;
	orig_loglevel = console_loglevel;

	if (sizeof(swp_entry_t) != sizeof(long)) {
		printk(KERN_WARNING name_swsusp "The size of swp_entry_t != size of long. Please report this!\n");
		return RESUME2_RET;
	}
	
	if (!resume_file)
		printk(KERN_WARNING name_swsusp "You need to use a resume2= command line parameter to tell Software Suspend 2 where to look for an image.\n");
	else
		attempt_to_parse_resume_device(1);
	
	swsusp_init_proc();

	if (!(software_suspend_state & SOFTWARE_SUSPEND_RESUME_DEVICE_OK)) {
		/* Without a usable storage device we can do nothing - even if noresume is given */
		if (!num_writers)
			printk(KERN_ALERT name_swsusp "No writers have been registered. You must select a method of storing the image when compiling the kernel.\n");
		else
			printk(KERN_ALERT name_swsusp "Missing or invalid storage location (resume2= parameter). Please correct and rerun lilo (or equivalent) before suspending.\n");
		kernel_thread(swsusp_mainloop, NULL, CLONE_FS | CLONE_FILES | CLONE_SIGHAND | SIGCHLD);
		return RESUME2_RET;
	}

	/* We enable the possibility of machine suspend */
	orig_mem_free = nr_free_pages();

	suspend_task = current->pid;

	printk(name_swsusp "Checking for image...\n");

	read_image_result = read_primary_suspend_image(); /* non fatal error ignored */

	if (software_suspend_state & SOFTWARE_SUSPEND_NORESUME_SPECIFIED)
		printk(KERN_WARNING name_swsusp "Resuming disabled as requested.\n");

	if (read_image_result) {
		suspend_task = 0;
		software_suspend_state &= ~3;
		kernel_thread(swsusp_mainloop, NULL, CLONE_FS | CLONE_FILES | CLONE_SIGHAND | SIGCHLD);
		console_loglevel = orig_loglevel;
		return RESUME2_RET;
	}

	freeze_processes(1);

	prepare_status(0, 0, "Copying original kernel back (no status - sensitive!)...");
	
	software_suspend_state |= SOFTWARE_SUSPEND_RUNNING;

	PRINTPREEMPTCOUNT("Prior to calling do_swsusp_lowlevel.");
        
	/* Suspend always runs on processor 0 */
	ensure_on_processor_zero();

	do_swsusp_lowlevel(1);
	BUG();

	return RESUME2_RET;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
late_initcall(software_resume2);
#endif

/*
 * Resume setup: obtain the storage device.
 */

static int __init resume_setup(char *str)
{
	if (str == NULL)
		return 1;
	
	strncpy(resume_file, str, 255);
	return 0;
}

/*
 * Allow the user to set the action parameter from lilo, prior to resuming.
 */
static int __init swsusp_act_setup(char *str)
{
	if(str)
		swsusp_action=simple_strtol(str,NULL,0);
	swsusp_act_used = 1;
	return 0;
}

/*
 * Allow the user to set the debug parameter from lilo, prior to resuming.
 */
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
static int __init swsusp_dbg_setup(char *str)
{
	if(str)
		swsusp_debug_state=simple_strtol(str,NULL,0);
	swsusp_dbg_used = 1;
	return 0;
}

/*
 * Allow the user to set the debug level parameter from lilo, prior to
 * resuming.
 */
static int __init swsusp_lvl_setup(char *str)
{
	if(str)
		console_loglevel = swsusp_default_console_level= simple_strtol(str,NULL,0);
	swsusp_lvl_used = 1;
	return 0;
}
#endif

/*
 * Allow the user to specify that we should ignore any image found and invalidate
 * the image if necesssary. This is equivalent to running the
 * task queue and a sync and then turning off the power. The same
 * precautions should be taken: fsck if you're not journalled.
 */
static int __init noresume_setup(char *str)
{
	software_suspend_state |= SOFTWARE_SUSPEND_NORESUME_SPECIFIED;
	/* Message printed later */
	return 0;
}

__setup("resume2=", resume_setup);
__setup("swsusp_act=", swsusp_act_setup);
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
__setup("swsusp_dbg=", swsusp_dbg_setup);
__setup("swsusp_lvl=", swsusp_lvl_setup);
#endif
__setup("noresume2", noresume_setup);

EXPORT_SYMBOL(software_suspend_pending);
EXPORT_SYMBOL(refrigerator);
EXPORT_SYMBOL(suspend_task);
