/*
 * kernel/power/ui.c
 *
 * Copyright (C) 1998-2001 Gabor Kuti <seasons@fornax.hu>
 * Copyright (C) 1998,2001,2002 Pavel Machek <pavel@suse.cz>
 * Copyright (C) 2002-2003 Florent Chabaud <fchabaud@free.fr>
 * Copyright (C) 2002-2003 Nigel Cunningham <ncunningham@clear.net.nz>
 *
 * This file is released under the GPLv2.
 *
 * Routines for Software Suspend's user interface.
 * 
 * The user interface includes support for both a 'nice display' and
 * a bootsplash screen (bootsplash.org), and for run-time debugging.
 *
 * The 'nice display' is text based and implements a progress bar and
 * (optional) textual progress, as well as an overall description of
 * the current action and the display of a header and the code version.
 *
 * The bootsplash support uses calls to bootsplash's proc interface
 * to set the progress bar value. These calls replace the display of
 * the text based progress bar only; if the bootsplash is in verbose
 * mode, the header, version and description still show. In silent
 * mode, the calls are still made but the text is not seen.
 *
 * Note that this implies a bootsplash picture version of 3 or later.
 * Earlier ones require a separate bootsplash option patch.
 *
 * As well as displaying status information, the user has some control
 * over the software while suspending. Hooks in drivers/char/console.c
 * and drivers/char/serial.c allow a user on the keyboard or serial
 * console to...
 *
 * 					Keyboard Key	Serial Key
 * Toggle rebooting			    R		    R
 * Toggle logging all output		    L		    L
 * Toggle pausing between major steps (1) Pause		    P
 * Toggle pausing at minor steps	Shift Lock
 * Switch log levels (2)		   0-7		   0-7
 * Cancel the suspend (3)		  Escape	  Escape
 * Continue when paused			  Shift		  Space
 *
 * (1) Pausing only occurs when the log level is > 1.
 * (2) When debugging is not compiled in, only levels 0 & 1 work.
 * (3) Can be disabled using /proc/swsusp/enable_escape.
 *
 * (The following assumes debugging is compiled in)...
 * 
 * Fundamental to the debugging code is the idea that each debug 
 * message which suspend can display has both a section of code to
 * which it belongs and a level of verbosity. When suspend is running,
 * it only displays a message if debugging for the section has been
 * enabled prior to beginning the cycle and the current loglevel is
 * greater than or equal to the level of the message.
 * 
 * Prior to starting a suspend cycle, a user can select which sections
 * should display debugging information by setting
 * /proc/swsusp/debug_sections. This is a bit vector (values in
 * include/linux/suspend-debug.h). The initial log level can be also
 * be set using /proc/swsusp/default_console_level.
 * The debug sections and level can be changed prior to resuming using
 * the kernel command line parameters swsusp_dbg and swsusp_lvl.
 * 
 * In addition to the above ability to control whether messages are
 * displayed, messages can be displayed in two ways. The printlog call
 * displays a message using printk, with the result that it is also
 * logged to syslog in the normal way.
 *
 * A call to printnolog usually gets the text to the screen using
 * vt_console_print and is thus not logged. This is the preferred
 * means of displaying highlevel debugging information, because it
 * reduces clutter in the logs. (You can easily end up with 1/5645.^M
 * 2/5645.^M 3/5645.^M... otherwise).
 * If loggging of this output is wanted, the log all output toggle
 * can be activated and printk will be used instead of
 * vt_console_print.
 */
#define SWSUSP_CONSOLE_C

#include <linux/suspend-common.h>

/* Keyboard table, used for determining current state of the shift
 * lock key. */
extern struct kbd_struct kbd_table[MAX_NR_CONSOLES];
extern void reset_terminal(int currcons, int do_clear); /* Calls for nice display */
extern void hide_cursor(int currcons);

/* All output is suppressed if the user switches consoles after
 * suspend starts or if the NO_OUTPUT bit is on. (A vendor might
 * want this for a static display?). */
#define NO_OUTPUT_OR_PAUSING ((fg_console != suspend_console) || \
		(TEST_ACTION_STATE(SUSPEND_NO_OUTPUT)))

/* Conditions which determine whether a message is displayed. */
#define CHECKMASK(mask) (((!mask) || TEST_DEBUG_STATE(mask)) && (!NO_OUTPUT_OR_PAUSING))

/* Pausing is implmented using a completion handler, rather than
 * spinning and chewing battery */
DECLARE_COMPLETION(swsusp_wait_for_shift);

#ifdef DEFAULT_SUSPEND_CONSOLE
static int barwidth = 100, barposn = -1, newbarposn = 0;
static int orig_kmsg;
extern int swsusp_default_console_level;
#endif

#ifdef CONFIG_VT
static int orig_fgconsole;
int orig_loglevel = 0;
static char print_buf[1024];	/* Same as printk - should be safe */
#endif

#define cond_console_print(chars, count) \
	if (suspend_console == fg_console) vt_console_print((struct console *) NULL, chars, count)

/* We want the number of lines & columns for the suspend console, not the current console */
#undef video_num_columns
#define video_num_columns	(vc_cons[suspend_console].d->vc_cols)
#undef video_num_lines
#define video_num_lines		(vc_cons[suspend_console].d->vc_rows)

/* Forward declaration */
void force_console_redraw(void);

static int suspend_console = 0;	

/* Your bootsplash progress bar may have a width of (eg) 1024 pixels. That
 * doesn't necessarily mean you want the bar updated 1024 times when writing
 * the image */
static int bar_granularity_limit = 0;

/* We remember the last header that was (or could have been) displayed for
 * use during log level switches */
static char lastheader[512];
static int lastheader_message_len = 0;

/* Is the 'nice display' on right now? */
static int nice_display = 0;

/* Remember the last loglevel so we can tell when it changes */
static int lastloglevel = -1;
static int orig_default_message_loglevel;

/* ------------------  Splash screen defines  -------------------------- */

#if defined(CONFIG_FBCON_SPLASHSCREEN) && defined(CONFIG_PROC_FS)
extern struct display fb_display[MAX_NR_CONSOLES];
extern int splash_write_proc(struct file *file, const char *buffer, unsigned long count, void *data);

/* splash_is_on
 * 
 * Description: Determine whether a VT has a splash screen on.
 * Arguments:	int consolenr. The VT number of a console to check.
 * Returns:	Boolean indicating whether the splash screen for
 *		that console is on right now.
 */
static int splash_is_on(int consolenr)
{
	struct display *p = &fb_display[consolenr];

	BUG_ON(consolenr >= MAX_NR_CONSOLES);
	
	if (p && p->splash_data)
		return ((p->splash_data->splash_state & 1) == 1);
	return 0;
}

/* fb_splash-set_progress
 *
 * Description:	Set the progress bar position for a splash screen.
 * Arguments:	int consolenr. The VT number of a console to use.
 * 		unsigned long value, unsigned long maximum:
 * 		The proportion (value/maximum) of the bar to fill.
 */

static int fb_splash_set_progress(int consolenr, unsigned long value, unsigned long maximum)
{
	char procstring[15];
	int length, bitshift = generic_fls(maximum) - 16;
	static unsigned long lastvalue = 0;
	unsigned long thisvalue;

	BUG_ON(consolenr >= MAX_NR_CONSOLES);
	
	if (value > maximum)
		value = maximum;

	/* Avoid math problems - we can't do 64 bit math here
	 * (and don't need it - anyone got screen resolution
	 * of 65536 pixels or more?) */
	if (bitshift > 0) {
		maximum = maximum >> bitshift;
		value = value >> bitshift;
	}

	thisvalue = value * 65534 / maximum;

	length = sprintf(procstring, "show %lu", thisvalue);
	
	splash_write_proc(NULL, procstring, length, NULL);
	
	/* Ensure redraw when the progress bar goes to a lower value */
	if (thisvalue < lastvalue)
		force_console_redraw();

	lastvalue = thisvalue;
	
	return 0;
}
#else
#define splash_is_on(consolenr) (0)
#define fb_splash_set_progress(...) do { } while(0)
#define splash_write_proc(...) do { } while(0)
#endif

/* abort_suspend
 *
 * Description: Begin to abort a cycle. If this wasn't at the user's request
 * 		(and we're displaying output), tell the user why and wait for
 * 		them to acknowledge the message.
 * Arguments:	A parameterised string (imagine this is printk) to display,
 *	 	telling the user why we're aborting.
 */

void abort_suspend(const char *fmt, ...)
{
	va_list args;
	int printed_len = 0;

	if (!TEST_RESULT_STATE(SUSPEND_ABORTED)) {
		if ((!TEST_RESULT_STATE(SUSPEND_ABORT_REQUESTED)) && (!NO_OUTPUT_OR_PAUSING)) {
			int locked = (spin_is_locked(&suspend_irq_lock));
			if (locked)
				swsusp_spin_unlock_irqrestore(&suspend_irq_lock, swsuspirqflags);
			va_start(args, fmt);
			printed_len = vsnprintf(print_buf, sizeof(print_buf), fmt, args);
			va_end(args);
			printed_len = sprintf(print_buf + printed_len,
					" (Press SHIFT to continue)");
			prepare_status(1, 1, print_buf);
			beepERR;
			/* 
			 * Make sure message seen - wait for shift to be
			 * released if being pressed 
			 */
			init_completion(&swsusp_wait_for_shift);
			wait_for_completion(&swsusp_wait_for_shift);
			if (locked)
				swsusp_spin_lock_irqsave(&suspend_irq_lock, swsuspirqflags);
		}
		/* Turn on aborting flag */
		SET_RESULT_STATE(SUSPEND_ABORTED);
	}
}

/* handle_loglevel_change
 *
 * Description:	Update the display when the user changes the log level.
 * Returns:	Boolean indicating whether the level was changed.
 */

static int handle_loglevel_change(void)
{
	static int recursive = 0;
	
	if ((console_loglevel == lastloglevel) || recursive)
		return 0;
	
	recursive = 1;
	
	/* Calculate progress bar width. Note that whether the
	 * splash screen is on might have changed (this might be
	 * the first call in a new cycle), so we can't take it
	 * for granted that the width should be the same as
	 * last time we came in here */
	if (splash_is_on(suspend_console)) {
		/* proc interface ensures bar_granularity_limit >= 0 */
		if (bar_granularity_limit)
			barwidth = bar_granularity_limit;
		else
			barwidth = 100;
	} else 
		barwidth = (video_num_columns - 2 * (video_num_columns / 4) - 2);

	/* Only reset the display if we're switching between nice display
	 * and displaying debugging output */
	if (console_loglevel > 1) {
		if (nice_display || (lastloglevel == -1)) {
			nice_display = 0;
			reset_terminal(suspend_console, 1);
			hide_cursor(suspend_console);
			printnolog(0, console_loglevel, 0, "Switched to console loglevel %d.\n", console_loglevel);
			if (splash_is_on(suspend_console))
				splash_write_proc(NULL, "verbose\n", 9, NULL);
		}
	} else if (!nice_display) {
		nice_display = 1;
		reset_terminal(suspend_console, 1);
		hide_cursor(suspend_console);
		if (splash_is_on(suspend_console))
			splash_write_proc(NULL, "silent\n", 8, NULL);
	}
	
	lastloglevel = console_loglevel;

	/* Get the nice display or last action [re]drawn */
	prepare_status(0, 0, NULL);
	
	recursive = 0;
	
	return 1;
}

#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
/* printnolog.
 *
 * Description:	This function is intended to do the same job as printk, but
 * 		without normally logging what is printed. The point is to be
 * 		able to get debugging info on screen without filling the logs
 * 		with "1/534. ^M 2/534^M. 3/534^M"
 *
 * 		It may be called from an interrupt context - can't sleep!
 *
 * Arguments:	int mask: The debugging section(s) this message belongs to.
 * 		int level: The level of verbosity of this message.
 * 		int restartline: Whether to output a \r or \n with this line
 * 			(\n if we're logging all output).
 * 		const char *fmt, ...: Message to be displayed a la printk.
 */
void printnolog(int mask, int level, int restartline, const char *fmt, ...)
{
#if defined(DEFAULT_SUSPEND_CONSOLE)
	va_list args;
	int printed_len = 0;

	if (console_loglevel < level)
		return;

	if (!CHECKMASK(mask))
		return;

	/* Don't do this if not printing anything - print[no]log get called before
	 * we prepare the console at resume time */
	handle_loglevel_change();

	if ((restartline) && (!TEST_ACTION_STATE(SUSPEND_LOGALL)))
		printed_len = vsnprintf(print_buf, sizeof(print_buf), "\r", NULL);

	va_start(args, fmt);
	printed_len += vsnprintf(print_buf + printed_len, sizeof(print_buf) - printed_len, fmt, args);
	va_end(args);

	if TEST_ACTION_STATE(SUSPEND_LOGALL) {
		/* If we didn't print anything, don't do the \n anyway! */
		if (!printed_len)
			return;
		if (restartline)
			printk("\n");
		printk(print_buf);
	} else
		cond_console_print(print_buf, printed_len);
	hide_cursor(suspend_console);
#endif
}

/* printlog
 *
 * Description: This function is a wrapper around printk. It makes the display
 * 		conditional and adds the log level of the message to the front,
 * 		so we can change the detail displayed at runtime.
 * Arguments:	int mask: The debugging section(s) this message belongs to.
 * 		int level: The level of verbosity of this message.
 * 		const char *fmt, ...: Message to be displayed via printk.
 */
void printlog(int mask, int level, const char *fmt, ...)
{
#if defined(DEFAULT_SUSPEND_CONSOLE)
	va_list args;
	int printed_len = 0;

	if (console_loglevel < level)
		return;

	if (!CHECKMASK(mask))
		return;

	/* Don't do this if not printing anything - print[no]log get called before
	 * we prepare the console at resume time */
	handle_loglevel_change();

	va_start(args, fmt);
	printed_len += vsnprintf(print_buf + printed_len, sizeof(print_buf) - printed_len, fmt, args);
	va_end(args);

	printk(print_buf);
	hide_cursor(suspend_console);
#endif
}
#endif

/* prepare_status
 * Description:	Prepare the 'nice display', drawing the header and version,
 * 		along with the current action and perhaps also resetting the
 * 		progress bar.
 * Arguments:	int printalways: Whether to print the action when debugging
 * 		is on.
 * 		int clearbar: Whether to reset the progress bar.
 * 		const char *fmt, ...: The action to be displayed.
 */
void prepare_status(int printalways, int clearbar, const char *fmt, ...)
{
#if defined(DEFAULT_SUSPEND_CONSOLE)
	unsigned char posn[2];
	va_list args;

	if (NO_OUTPUT_OR_PAUSING)
		return;

	if (fmt) {
		va_start(args, fmt);
		lastheader_message_len = vsnprintf(lastheader, 512, fmt, args);
		va_end(args);
	}

	handle_loglevel_change();

	if (console_loglevel >= SUSPEND_ERROR) {

		if (printalways)
			printk("\n** %s\n", lastheader);
	
		/* If nice display is off, there is no progress bar to clear */
		if ((clearbar) && (splash_is_on(suspend_console)))
			fb_splash_set_progress(suspend_console, 0, 1);
		return;
	}
	
	/* Print version */
	posn[0] = (unsigned char) (0);
	posn[1] = (unsigned char) (video_num_lines);
	putconsxy(suspend_console, posn);
	cond_console_print(swsusp_version, strlen(swsusp_version));

	/* Print header */
	posn[0] = (unsigned char) ((video_num_columns - 29) / 2);
	posn[1] = (unsigned char) ((video_num_lines / 3) -4);
	putconsxy(suspend_console, posn);

	if (now_resuming) {
		cond_console_print(console_resume, strlen(console_resume));
	} else
		cond_console_print(console_suspend, strlen(console_suspend));
		
	/* Print action */
	posn[1] = (unsigned char) (video_num_lines / 3);
	posn[0] = (unsigned char) 0;
	putconsxy(suspend_console, posn);
	
	/* Clear old message */
	for (barposn = 0; barposn < video_num_columns; barposn++) 
		cond_console_print(" ", 1);

	posn[0] = (unsigned char) (((video_num_columns - lastheader_message_len) / 2));
	putconsxy(suspend_console, posn);
	cond_console_print(lastheader, lastheader_message_len);
	
	if (!splash_is_on(suspend_console)) {
		/* Draw left bracket of progress bar. */
		posn[0] = (unsigned char) (video_num_columns / 4);
		posn[1]++;
		putconsxy(suspend_console, posn);
		cond_console_print("[", 1);

		/* Draw right bracket of progress bar. */
		posn[0] = (unsigned char) (video_num_columns - (video_num_columns / 4) - 1);
		putconsxy(suspend_console, posn);
		cond_console_print("]", 1);
	}

	if (clearbar) {
		if (splash_is_on(suspend_console))
			fb_splash_set_progress(suspend_console, 0, 1);
		else {
			/* Position at start of progress */
			posn[0] = (unsigned char) (video_num_columns / 4 + 1);
			putconsxy(suspend_console, posn);

			/* Clear bar */
			for (barposn = 0; barposn < barwidth; barposn++)
				cond_console_print(" ", 1);
			putconsxy(suspend_console, posn);
		}
	}

	barposn = 0;
	hide_cursor(suspend_console);
#endif
}

/* update_status
 *
 * Description: Update the progress bar and (if on) in-bar message.
 * Arguments:	UL value, maximum: Current progress percentage (value/max).
 * 		const char *fmt, ...: Message to be displayed in the middle
 * 		of the progress bar.
 * 		Note that a NULL message does not mean that any previous
 * 		message is erased! For that, you need prepare_status with
 * 		clearbar on.
 * Returns:	Unsigned long: The next value where status needs to be updated.
 * 		This is to reduce unnecessary calls to update_status.
 */
unsigned long update_status(unsigned long value, unsigned long maximum, const char *fmt, ...)
{
	unsigned long next_update = 0;
#if defined(DEFAULT_SUSPEND_CONSOLE)
	int bitshift = generic_fls(maximum) - 16;
	unsigned char posn[2];
	va_list args;
	int message_len = 0;

	handle_loglevel_change();

	if (NO_OUTPUT_OR_PAUSING || (!maximum) || (!barwidth))
		return maximum;

	if (value < 0)
		value = 0;

	if (value > maximum)
		value = maximum;

	/* Try to avoid math problems - we can't do 64 bit math here
	 * (and shouldn't need it - anyone got screen resolution
	 * of 65536 pixels or more?) */
	if (bitshift > 0) {
		unsigned long temp_maximum = maximum >> bitshift;
		unsigned long temp_value = value >> bitshift;
		newbarposn = (int) (temp_value * barwidth / temp_maximum);
	} else
		newbarposn = (int) (value * barwidth / maximum);
	
	if (newbarposn < barposn)
		barposn = 0;

	next_update = ((newbarposn + 1) * maximum / barwidth) + 1;

	if (console_loglevel >= SUSPEND_ERROR) {
		if ((splash_is_on(suspend_console)) && (newbarposn != barposn)) {
			fb_splash_set_progress(suspend_console, value, maximum);
			barposn = newbarposn;
		}
		return next_update;
	}

	/* Update bar */
	if (splash_is_on(suspend_console)) {
		if (newbarposn != barposn)
			fb_splash_set_progress(suspend_console, value, maximum);
	} else {
		posn[1] = (unsigned char) ((video_num_lines / 3) + 1);

		/* Clear bar if at start */
		if (!barposn) {
			posn[0] = (unsigned char) (video_num_columns / 4 + 1);
			putconsxy(suspend_console, posn);
			for (; barposn < barwidth; barposn++)
				cond_console_print(" ", 1);
			barposn = 0;
		}
		posn[0] = (unsigned char) (video_num_columns / 4 + 1 + barposn);
		putconsxy(suspend_console, posn);

		for (; barposn < newbarposn; barposn++)
			cond_console_print("-", 1);
	}

	/* Print string in progress bar on loglevel 1 */
	if ((fmt) && (console_loglevel)) {
		va_start(args, fmt);
		message_len = vsnprintf(print_buf, sizeof(print_buf), " ", NULL);
		message_len += vsnprintf(print_buf + message_len, sizeof(print_buf) - message_len, fmt, args);
		message_len += vsnprintf(print_buf + message_len, sizeof(print_buf) - message_len, " ", NULL);
		va_end(args);

		if (message_len) {
			posn[0] = (unsigned char) ((video_num_columns - message_len) / 2);
			posn[1] = (unsigned char) ((video_num_lines / 3) + 1);
			putconsxy(suspend_console, posn);
			cond_console_print(print_buf, message_len);
		}
	}

	barposn = newbarposn;
	hide_cursor(suspend_console);
#else
	next_update = maximum;
#endif
	return next_update;
}

/* request_abort_suspend
 *
 * Description:	Handle the user requesting the cancellation of a suspend by
 * 		pressing escape. Note that on a second press, we try a little
 * 		harder, attempting to forcefully thaw processes. This shouldn't
 * 		been needed, and may result in an oops (if we've overwritten
 * 		memory), but has been useful on ocassion.
 * Callers:	Called from drivers/char/keyboard.c or drivers/char/serial.c
 * 		when the user presses escape.
 */
void request_abort_suspend(void)
{
	if ((now_resuming) || (TEST_RESULT_STATE(SUSPEND_ABORT_REQUESTED)))
		return;

	if (TEST_RESULT_STATE(SUSPEND_ABORTED)) {
		prepare_status(1, 1, "--- ESCAPE PRESSED AGAIN : TRYING HARDER TO ABORT ---");
		show_state();
		thaw_processes();
	} else {
		prepare_status(1, 1, "--- ESCAPE PRESSED : ABORTING PROCESS ---");
		SET_RESULT_STATE(SUSPEND_ABORTED);
		SET_RESULT_STATE(SUSPEND_ABORT_REQUESTED);
		abort_suspend("Escape pressed.");
	}
}

/* check_shift_keys
 * 
 * Description:	Potentially pause and wait for the user to tell us to continue.
 * 		We normally only pause when @pause is set.
 * Arguments:	int pause: Whether we normally pause.
 * 		char * message: The message to display. Not parameterised
 * 		 because it's normally a constant.
 */

void check_shift_keys(int pause, char * message)
{
#ifdef DEFAULT_SUSPEND_CONSOLE 
	static struct kbd_struct * kbd;
	
	kbd = kbd_table + fg_console;

	//set_task_state(current, TASK_UNINTERRUPTIBLE);
	//schedule_timeout(5);

	if (NO_OUTPUT_OR_PAUSING)
		return;

	if (((TEST_ACTION_STATE(SUSPEND_PAUSE) && pause) || (vc_kbd_led(kbd, VC_CAPSLOCK))) 
			&& (console_loglevel > 1)) { 
		prepare_status(1, 0, "%s%sPress SHIFT to continue.\n",
			(vc_kbd_led(kbd, VC_CAPSLOCK)) ? "(Caps lock on) " : "",
			message ? message : "");
		init_completion(&swsusp_wait_for_shift);
		wait_for_completion(&swsusp_wait_for_shift);
	}
#endif
}

/* pm_prepare_console
 * Description:	Prepare a console for use, save current settings.
 * Returns:	Boolean: Whether an error occured. Errors aren't
 * 		treated as fatal, but a warning is printed.
 */
int pm_prepare_console(void)
{
	if (TEST_ACTION_STATE(SUSPEND_NO_OUTPUT))
		/* No output should be produced. */ 
		return 0;

	orig_loglevel = console_loglevel;
	orig_default_message_loglevel = default_message_loglevel;
	if (!now_resuming)
		console_loglevel = swsusp_default_console_level;

#ifdef CONFIG_VT_CONSOLE
	orig_fgconsole = fg_console;

	if ((orig_fgconsole != DEFAULT_SUSPEND_CONSOLE) && (!splash_is_on(fg_console))) {
		suspend_console = DEFAULT_SUSPEND_CONSOLE;
		if (vc_allocate(DEFAULT_SUSPEND_CONSOLE))
		  /* we can't have a free VC for now. Too bad,
		   * we don't want to mess the screen for now. */
			return 1;

		set_console(DEFAULT_SUSPEND_CONSOLE);
		if (vt_waitactive(DEFAULT_SUSPEND_CONSOLE)) {
			printk("Can't switch virtual consoles.");
			return 1;
		}
	} else 
		suspend_console = fg_console;
	reset_terminal(suspend_console, 1);
	orig_kmsg = kmsg_redirect;
	kmsg_redirect = suspend_console;
	prepare_status(0, 0, "");

	default_message_loglevel = 1;
#endif
	if (splash_is_on(suspend_console)) {
		if (console_loglevel > 1)
			splash_write_proc(NULL, "verbose\n", 9, NULL);
		else
			splash_write_proc(NULL, "silent\n", 8, NULL);
	}
	return 0;
}

/* pm_restore_console
 *
 * Description: Restore the settings we saved above.
 */

void pm_restore_console(void)
{

	if (TEST_ACTION_STATE(SUSPEND_NO_OUTPUT))
		return;

	reset_terminal(suspend_console, 1);
	
#ifdef CONFIG_VT_CONSOLE
	if (orig_fgconsole != suspend_console) {
		set_console(orig_fgconsole);
		if (vt_waitactive(orig_fgconsole)) {
			printk("Can't switch virtual consoles.");
			return;
		}
	}
#endif
	
	kmsg_redirect = orig_kmsg;
	swsusp_default_console_level = console_loglevel;
	console_loglevel = orig_loglevel;
	default_message_loglevel = orig_default_message_loglevel;
	nice_display = 0;
	lastloglevel = -1;
	return;
}

extern void unblank_screen(void);
extern int console_blanked;
extern int handle_loglevel_change(void);

/* force_console_redraw
 *
 * Description:	Force a redraw of the console. Necessary after copying the
 * 		original kernel back and when the progress bar is moved
 * 		backwards.
 */
void force_console_redraw(void)
{
	console_blanked = fg_console + 1;
	handle_loglevel_change();
	unblank_screen();
	update_screen(fg_console);
}

/* post_resume_console_redraw(void)
 *
 * Description:	Redraw the console after copying the original kernel back.
 */

void post_resume_console_redraw(void)
{
	unblank_screen();
	fb_splash_set_progress(suspend_console, pageset1_size, pageset1_size + pageset2_size);
}

#if defined(CONFIG_SOFTWARE_SUSPEND2) && defined(CONFIG_PROC_FS)
/*
 * User interface specific /proc/swsusp entries.
 */

static struct swsusp_proc_data proc_params[] = {
	{ .filename			= "beeping",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_BIT,
	  .data = {
		  .bit = {
			  .bit_vector	= &swsusp_action,
			  .bit		= SUSPEND_BEEP,
		  }
	  }
	},

	{ .filename			= "default_console_level",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_INTEGER,
	  .data = {
		  .integer = {
			  .variable	= &swsusp_default_console_level,
			  .minimum	= 0,
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
			  .maximum	= 7,
#else
			  .maximum	= 1,
#endif

		  }
	  }
	},

	{ .filename			= "enable_escape",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_BIT,
	  .data = {
		  .bit = {
			  .bit_vector	= &swsusp_action,
			  .bit		= SUSPEND_CAN_CANCEL,
		  }
	  }
	},

	{ .filename			= "no_output",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_BIT,
	  .data = {
		  .bit = {
			  .bit_vector	= &swsusp_action,
			  .bit		= SUSPEND_NO_OUTPUT,
		  }
	  }
	},
	
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
	{ .filename			= "debug_sections",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_UL,
	  .data = {
		  .ul = {
			  .variable	= &swsusp_debug_state,
			  .minimum	= 0,
			  .maximum	= 2 << 30,
		  }
	  }
	},

	{ .filename			= "log_everything",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_BIT,
	  .data = {
		  .bit = {
			  .bit_vector	= &swsusp_action,
			  .bit		= SUSPEND_LOGALL,
		  }
	  }
	},
	  
	{ .filename			= "pause_between_steps",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_BIT,
	  .data = {
		  .bit = {
			  .bit_vector	= &swsusp_action,
			  .bit		= SUSPEND_PAUSE,
		  }
	  }
	},
#endif

#ifdef CONFIG_FBCON_SPLASHSCREEN
	{ .filename			= "progressbar_granularity_limit",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_INTEGER,
	  .data = {
		  .integer = {
			  .variable	= &bar_granularity_limit,
			  .minimum	= 1,
			  .maximum	= 2000,
		  }
	  }
	}
#endif
};

/* swsusp_console_proc_init
 * Description: Boot time initialisation for user interface.
 */
__init void swsusp_console_proc_init(void)
{
	int i, numfiles = sizeof(proc_params) / sizeof(struct swsusp_proc_data);

	for (i=0; i< numfiles; i++)
		swsusp_register_procfile(&proc_params[i]);
}

#endif

#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
EXPORT_SYMBOL(printlog);
#endif
